@0xf66f85018d81e41f;

using Cxx = import "/capnp/c++.capnp";
$Cxx.namespace("serial");

struct LogVariable {
    # A variable in a process event log (e.g. department, case ID) that would typically be indicated
    # in the column header of the table

    name @0 :Text; # Unique name of the variable, used to uniquely identify the log variable

    union {
        categorical :group {
            # The variable is nominal, it can only take a value out of a finite and unordered set

            domain @1 :List(Text); # Which values are possible for the variable
        }
        numerical @2 :Void; # The variable can (in principle) take the value of any real number
    }
}

struct LogColumn {
    # A column in a process event log that collects all entries for the same variable

    variableName @2 :Text; # Name of the variable that the entries belong to (reference)

    union {
        categorical :group {
            # The column and the variable are nominal

            entries @0 :List(Text); # categorical entries are stored as texts
        }
        numerical :group {
            # The column and the variable can (in principle) take the value of any real number

            entries @1 :List(Float64); # numerical entries are stored as floating point numbers
        }
    }
}

struct Trace {
    # A trace of a process event log that collects all event data for one case (e.g. patient, order)
    # handled by the process

    caseId @1 :Text; # Unique identifier of the case
    columns @0 :List(LogColumn); # All event data organized per log variable/per column
}

struct EventLog {
    # Dataset containing the digital footprint of a business process

    traces @0 :List(Trace); # Process event data organized per case
    variables @1 :List(LogVariable); # Data attributes that were recorded for the process
}

struct UpdateRule {
    # Specifies how a variable is changed in the process event log

    variable @0 :Text; # Name of the variable to change (reference)

    union {
        valueSetRule :group {
            # Specifies that a categorical variable is a member of a subset of its domain

            set @1 :List(Text);
        }
        valueRangeRule :group {
            # Specifies that a numerical variable is in the interval [min, max]

            min @2 :Float64;
            max @3 :Float64;
        }
        changeRangeRule :group {
            # Specifies that the change (difference) of a numerical variable with respect to its
            # value at the previous sample/event is in the interval [min, max]

            min @4 :Float64;
            max @5 :Float64;
        }
        valueRule :group {
            # Specifies that a numerical variable is assigned a specific value

            constant @6 :Float64;
        }
        factorRule :group {
            # Specifies that the relative change (factor) of a numerical variable with respect to
            # its value at the previous sample/event equals the constant of the rule

            constant @7 :Float64;
        }
        differenceRule :group {
            # Specifies that the change (difference) of a numerical variable with respect to its
            # value at the previous sample/event equals the constant of the rule

            constant @8 :Float64;
        }
    }
}

struct AtomicCondition {
    # Contains a single check to be performed on a value of a log variable

    variable @0 :Text; # Name of the variable to check (reference)

    union {
        lowerBoundCondition :group {
            # Checks value <= variable

            value @1 :Float64;
        }
        upperBoundCondition :group {
            # Checks variable <= value

            value @2 :Float64;
        }
        numericalValueCondition :group {
            # Checks variable == value

            value @3 :Float64;
        }
        categoricalValueCondition :group {
            # Checks variable == value

            value @4 :Text;
        }
        numericalTransitionCondition :group {
            # Checks that the variable had the value <before> in the last sample/event and
            # the value <after> in the current sample/event

            before @5 :Float64;
            after @6 :Float64;
        }
        categoricalTransitionCondition :group {
            # Checks that the variable had the value <before> in the last sample/event and
            # the value <after> in the current sample/event

            before @7 :Text;
            after @8 :Text;
        }
    }
}

struct ConditionElement {
    # Part of a condition, specifies how to check values or which logical operator to use
    union {
        notOperator @3 :Void; # Logical NOT operator
        andOperator @0 :Void; # Logical AND operator
        orOperator @1 :Void; # Logical OR operator
        atomicCondition @2 :AtomicCondition; # Check on a value of a log variable
    }
}

struct Condition {
    # Specifies when a rule should be applied, is a compound of elements

    elements @0 :List(ConditionElement);
    # Elements of the condition specifying the checks on the variables and the logical operators.
    # Checks and operators are organized in postfix notation.
}

struct ModificationRule {
    # A single data modification rule that specifies how and under which conditions data is changed
    # in a business process

    condition @0 :Condition; # Specifies when the rule should be applied
    updateRule @1 :UpdateRule; # Specifies how the data is changed
}

struct RuleModel {
    # Contains all data modification rules for a business process

    rules @0 :List(ModificationRule); # The data modification rules
    variables @1 :List(LogVariable); # Data attributes that were recorded for the process
}

struct RuleUsageStatistic {
    # Indicates how many times a data modification rule was used while generating an artificial
    # event log

    rule @0 :ModificationRule; # The data modification rule used
    frequency @1 :Int64; # The number of samples/events where the rule was used to generate a value
}

struct SearchMode {
    # Specifies how candidate data modification rules are generated during the search

    union {
        conditionFirst @0 :Void;
        # First generate frequent conditions and then find fitting update rules for them

        updateRuleFirst @1 :Void;
        # First generate common update rules and then find conditions for them
    }
}

struct DiscretizationMode {
    # Specifies how numerical values should be discretized

    union {
        nml @0 :Void;
        # Use variable-width histograms that are optimized using Normalized Maximum Likelihood
        # (NML). Can be very slow but gives very good discretizations.

        quantiles @1 :Void;
        # Use histograms based on quantiles of the data. Very fast but results in lower quality
        # discretizations.
    }
}

struct SetPrediction {
    # Specifies which values are predicted for a categorical variable by a rule model in
    # an event log. Each prediction is a set of possible values to rule out stochastic behavior.

    variable @0 :Text; # Name of the categorical variable that is predicted
    predictions @1 :List(List(Text)); # Sets of possible values for each event in the log
}

struct EventLogPosition {
    # Identifies a single event in an event log.

    caseId @0 :Text; # The case ID identifying the trace that the event is in
    index @1 :Int64; # The event's position inside the trace
}

struct ApplicabilityMapping {
    # Stores which positions in an event log have a prediction for the target variable

    targetVariable @0 :Text; # The name of the target variable
    positions @1 :List(EventLogPosition); # Positions that have at least one applicable rule
}

interface ProcessDataUpdates {
    # Provides functionality to find data modification rules in process event logs

    calculateMdlScore @0 (log :EventLog, model :RuleModel,
        discretizationMode :DiscretizationMode = (quantiles = void)) -> (result :Float64);
    # Computes the central quality criterion used to evaluate how well a model fits an event log.
    # Provides configuration for how numerical variables are discretized.

    generateEventLog @1 (model :RuleModel, numberOfTraces :Int64,
        traceLength :Int64) -> (log :EventLog, ruleUsage :List(RuleUsageStatistic));
    # Generates an artificial event log using the data modification rules provided in the model.
    # The length of the event log can be controlled through the numer of traces and the numer of
    # samples/events per trace. Also returns information about rule usage besides the artificial
    # event log.
    
    findModel @2 (log :EventLog,
        targetVariables :List(Text) = [],
        maxLength :Int64 = 0x7fffffffffffffff,
        searchMode :SearchMode = (conditionFirst = void),
        discretizationMode :DiscretizationMode = (quantiles = void),
        conditionCount :UInt64 = 50) -> (model :RuleModel);
    # Main function to find a model of data modification rules from a given event log. If needed,
    # one can specify which target variables should be considered for the model. By default, with
    # an empty list of target variables, all variables will be considered. The length
    # of the model (i.e. its number of rules) can be limited to reduce execution time. Provides
    # further configuration capabilities for how candidate rules are generated during the search and
    # for how numerical variables are discretized for computing the MDL score. Finally, one can
    # configure how many conditions per non-target variable and condition type should be generated
    # during search. This significantly influences the search speed.
    
    modelToString @3 (model :RuleModel) -> (string :Text);
    # Converts a model of data modification rules into a human-readable text

    encodeRealNumber @4 (realNumber :Float64, precision :Int64) -> (codeLength :Float64);
    # Returns the MDL code length (in bits) of the real number up to the given number of significant
    # digits (precision)

    isModelCyclic @5 (model :RuleModel) -> (cyclic :Bool);
    # Returns whether a model has at least one cyclic dependency between its variables

    predictEventLog @6 (sourceLog :EventLog, model :RuleModel, predictMeans :Bool = false,
        modelIsOrderedList :Bool = false) -> (log :EventLog);
    # Generates an event log of predictions for each colum. These predictions are generated using
    # the model of modification rules. Should only be used to assess the prediction quality of the
    # model. The resulting log is not intrinsically consistent with the model. By default all values
    # are randomly sampled from the values that are possible according to the rule (or global data
    # generation). If predictMeans is set, numerical values will not be random but the middle value
    # of an interval. If modelIsOrderedList is set, the rules will be interpreted as an ordered list
    # instead of an unordered set.

    predictValueSets @7 (sourceLog :EventLog, model :RuleModel, modelIsOrderedList :Bool = false)
        -> (setPredictions :List(SetPrediction));
    # Generates predictions of value sets for each categorical variable/column. These predictions
    # are generated using the model of modification rules. Should be used to assess the prediction
    # quality without the stochastic behavior of data generation. If modelIsOrderedList is set,
    # the rules will be interpreted as an ordered list instead of an unordered set.

    estimateBitsSaved @8 (log :EventLog, model :RuleModel, newRule :ModificationRule,
        discretizationMode :DiscretizationMode = (quantiles = void)) -> (result :Float64);
    # Estimates how many bits can be saved by adding a new rule to the model.
    # Provides configuration for how numerical variables are discretized.

    predictRealEventLog @9 (trainingLog :EventLog, testLog :EventLog, model :RuleModel,
        predictMeans :Bool = false, modelIsOrderedList :Bool = false) -> (log :EventLog);
    # Generates an event log of predictions for each colum. These predictions are generated using
    # the test log and the model of modification rules. The training log is used to estimate each
    # variable's distribution. These distributions are used for prediction if no rule applies.
    # If predictMeans is set, numerical values will not be random but the middle value
    # of an interval. If modelIsOrderedList is set, the rules will be interpreted as an ordered list
    # instead of an unordered set.

    getPredictingPositions @10 (eventLog :EventLog, model :RuleModel)
    -> (positions :List(ApplicabilityMapping));
    # Finds all events (their positions) in the event log that the model makes actual predictions
    # for, i.e. the model has at least one applicable rule for the event.
}

interface SimpleProcessDataUpdates {
    # Provides functionality to find data modification rules in process event logs

    findModel @0 (log :EventLog, targetVariables :List(Text) = []) -> (model :RuleModel);
    # Main function to find a model of data modification rules from a given event log. If needed,
    # one can specify which target variables should be considered for the model. By default, with
    # an empty list of target variables, all variables will be considered. 
}
