#pragma once
#include <memory>
#include "../datamodel/rulemodel.hpp"
#include "../datamodel/eventlogs/eventlog.hpp"
#include "rulegenerator.hpp"
#include "../logger.hpp"
#include <queue>

/// @brief Discovers rule models from event logs by searching over the MDL score
class ModelFinder
{
    private:
    using EstimateRulePair = std::pair<double_t, std::unique_ptr<ModificationRule>>;

    Logger logger;
    /// @brief Merges the conditions of two modification rules using a logical OR operator
    /// @param oldRule first rule that will contain the merged condition after calling this method
    /// @param newRule second rule whose condition will be merged into the first one, all condition
    /// elements of this rule will be null after calling this method
    void mergeRuleConditions (ModificationRule* oldRule, ModificationRule* newRule);

    public:
    
    /// @brief Searches for good rule models for the event log using the MDL score and a greedy
    /// search strategy 
    /// @param log event log to find rule model for
    /// @param ruleGenerator generator object used to create candidate modification rules
    /// @param maxLength how many rules should be found at most (limits search time)
    /// @param targetVariables specifies which target variables should be considered for the model
    /// (all variables by default)
    /// @return the best rule model that the search could find
    std::unique_ptr<RuleModel> findRuleModel (
        const EventLog* const log, RuleGenerator* ruleGenerator,
        const int64_t maxLength = 0x7fffffffffffffff,
        const std::vector<std::shared_ptr<LogVariable>> targetVariables = {});

    /// @brief Adds a new rule to the model. If a modification rule with the same prediction
    /// (UpdateRule) already exists in the model, the new rule's condition is merged with the
    /// existing modification rule. Merging two modification rules means that their conditions are
    /// connected through a logical OR operator.
    /// @param model current model of modification rules that the new rule should be added to
    /// @param newRule new modification rule to be added to the model
    void addRuleWhileMergingConditions (
        RuleModel* model, std::unique_ptr<ModificationRule>&& newRule);
};
