#include <gtest/gtest.h>
#include <cmath>
#include "generators.hpp"
#include "../classes/applicabilityfinder.hpp"
#include "../classes/datamodel/conditions/valuecondition.hpp"
#include "../classes/datamodel/updaterules/valuerule.hpp"
#include "../classes/datamodel/updaterules/valuesetrule.hpp"

TEST(ApplicabilityFinderTest, EmptyLogTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<EventLog> log = std::make_unique<EventLog>();

    RuleModel::RulesType rules;

    Condition::ElementsType elements1;
    elements1.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "a"));
    std::unique_ptr<Condition> condition1 = std::make_unique<Condition>(elements1);
    std::unique_ptr<UpdateRule> updateRule1 = std::make_unique<ValueRule>(numVariable, 42);
    rules.push_back(
        std::make_unique<ModificationRule>(std::move(condition1), std::move(updateRule1)));

    Condition::ElementsType elements2;
    elements2.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "i"));
    std::unique_ptr<Condition> condition2 = std::make_unique<Condition>(elements2);
    std::unique_ptr<UpdateRule> updateRule2 = std::make_unique<ValueRule>(numVariable, 42);
    rules.push_back(
        std::make_unique<ModificationRule>(std::move(condition2), std::move(updateRule2)));

    auto model = std::make_unique<RuleModel>(rules);

    ApplicabilityFinder finder;
    ApplicabilityFinder::EventLogPositions expectedResult = {};
    EXPECT_EQ(expectedResult, finder.findEventsWithApplicableRules(log.get(), model.get()));
}

TEST(ApplicabilityFinderTest, EmptyModelTest)
{
    std::unique_ptr<EventLog> log = Generators::generateEventLog();
    std::unique_ptr<RuleModel> model = std::make_unique<RuleModel>();

    ApplicabilityFinder finder;
    ApplicabilityFinder::EventLogPositions expectedResult = {};
    EXPECT_EQ(expectedResult, finder.findEventsWithApplicableRules(log.get(), model.get()));
}

TEST(ApplicabilityFinderTest, EmptyLogAndModelTest)
{
    std::unique_ptr<EventLog> log = std::make_unique<EventLog>();
    std::unique_ptr<RuleModel> model = std::make_unique<RuleModel>();

    ApplicabilityFinder finder;
    ApplicabilityFinder::EventLogPositions expectedResult = {};
    EXPECT_EQ(expectedResult, finder.findEventsWithApplicableRules(log.get(), model.get()));
}

TEST(ApplicabilityFinderTest, NonEmptyTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<EventLog> log = Generators::generateEventLog(catVariable, numVariable);

    RuleModel::RulesType rules;
    Condition::ElementsType elements1;
    elements1.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "a"));
    std::unique_ptr<Condition> condition1 = std::make_unique<Condition>(elements1);
    std::unique_ptr<UpdateRule> updateRule1 = std::make_unique<ValueRule>(numVariable, 42);
    rules.push_back(
        std::make_unique<ModificationRule>(std::move(condition1), std::move(updateRule1)));

    Condition::ElementsType elements2;
    elements2.push_back(std::make_unique<ValueCondition<double_t>>(numVariable, 3));
    std::unique_ptr<Condition> condition2 = std::make_unique<Condition>(elements2);
    std::unique_ptr<UpdateRule> updateRule2 = std::make_unique<ValueSetRule>(catVariable,
        ValueSetRule::SetType{"u"});
    rules.push_back(
        std::make_unique<ModificationRule>(std::move(condition2), std::move(updateRule2)));

    std::unique_ptr<RuleModel> model = std::make_unique<RuleModel>(rules);

    ApplicabilityFinder finder;
    ApplicabilityFinder::EventLogPositions expectedResult = {
        {numVariable, {{"0", 1}, {"1", 1}, {"1", 4}}},
        {catVariable, {{"1", 1}, {"1", 2}, {"1", 3}}}
    };
    EXPECT_EQ(expectedResult, finder.findEventsWithApplicableRules(log.get(), model.get()));
}
