#include <gtest/gtest.h>
#include <cmath>
#include "generators.hpp"
#include "../classes/columncollector.hpp"
#include "../classes/datamodel/conditions/valuecondition.hpp"
#include "../classes/datamodel/updaterules/valuerule.hpp"
#include "../classes/datamodel/updaterules/valuesetrule.hpp"
#include "../classes/datamodel/updaterules/factorrule.hpp"
#include "../classes/datamodel/updaterules/differencerule.hpp"
#include "../classes/datamodel/updaterules/valuerangerule.hpp"
#include "../classes/datamodel/updaterules/changerangerule.hpp"
#include "../classes/datamodel/eventlogs/logcolumn.hpp"

TEST(ColumnCollectorTest, CollectCategoricalTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<EventLog> log = Generators::generateEventLog(catVariable, numVariable);

    ColumnCollector collector;
    ColumnCollector::MultiColumnsType<std::string> columns = 
        collector.collectAllColumns<std::string>(log.get());

    std::multiset<std::string> collectedValues(columns[catVariable].begin(), columns[catVariable].end());
    std::multiset<std::string> expectedValues{"u", "a", "u", "e", "i", "a", "u", "e", "a", "u"};

    EXPECT_EQ(expectedValues, collectedValues);
}

TEST(ColumnCollectorTest, CollectNumericalTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<EventLog> log = Generators::generateEventLog(catVariable, numVariable);

    ColumnCollector collector;
    ColumnCollector::MultiColumnsType<double_t> columns = 
        collector.collectAllColumns<double_t>(log.get());

    std::multiset<double_t> collectedValues(columns[numVariable].begin(), columns[numVariable].end());
    std::multiset<double_t> expectedValues{9, 9, 7, 7, 1, 3, 3, 3, 5, 6};

    EXPECT_EQ(expectedValues, collectedValues);
}

TEST(ColumnCollectorTest, CollectNumericalMissingValuesTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<EventLog> log = Generators::generateEventLogWithMissingValues(catVariable, numVariable);

    ColumnCollector collector;
    ColumnCollector::MultiColumnsType<double_t> columns = 
        collector.collectAllColumns<double_t>(log.get());

    std::multiset<double_t> collectedValues(columns[numVariable].begin(), columns[numVariable].end());
    std::multiset<double_t> expectedValues{9, 7, 7, 1, 3, 3, 5};

    EXPECT_EQ(expectedValues, collectedValues);
}

TEST(ColumnCollectorTest, CollectResidualValueTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<EventLog> log = Generators::generateEventLog(catVariable, numVariable);

    Condition::ElementsType elements;
    elements.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "u"));
    std::unique_ptr<Condition> condition = std::make_unique<Condition>(elements);
    std::unique_ptr<AbstractValueRule> updateRule = std::make_unique<ValueRule>(numVariable, 5);

    ColumnCollector collector;
    std::vector<double_t> result = 
        collector.collectResiduals(log.get(), condition.get(), updateRule.get());

    std::multiset<double_t> collectedValues(result.begin(), result.end());
    std::multiset<double_t> expectedValues{4, 2, -2, 1};

    EXPECT_EQ(expectedValues, collectedValues);
}

TEST(ColumnCollectorTest, CollectResidualFactorTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<EventLog> log = Generators::generateEventLog(catVariable, numVariable);

    Condition::ElementsType elements;
    elements.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "u"));
    std::unique_ptr<Condition> condition = std::make_unique<Condition>(elements);
    std::unique_ptr<AbstractValueRule> updateRule = std::make_unique<FactorRule>(numVariable, 2.0);

    ColumnCollector collector;
    std::vector<double_t> result = 
        collector.collectResiduals(log.get(), condition.get(), updateRule.get());

    std::multiset<double_t> collectedValues(result.begin(), result.end());
    std::multiset<double_t> expectedValues{-11, -3, -4};

    EXPECT_EQ(expectedValues, collectedValues);
}

TEST(ColumnCollectorTest, CollectResidualDifferenceTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<EventLog> log = Generators::generateEventLog(catVariable, numVariable);

    Condition::ElementsType elements;
    elements.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "u"));
    std::unique_ptr<Condition> condition = std::make_unique<Condition>(elements);
    std::unique_ptr<AbstractValueRule> updateRule = std::make_unique<DifferenceRule>(numVariable, -2.0);

    ColumnCollector collector;
    std::vector<double_t> result = 
        collector.collectResiduals(log.get(), condition.get(), updateRule.get());

    std::multiset<double_t> collectedValues(result.begin(), result.end());
    std::multiset<double_t> expectedValues{0, 2, 3};

    EXPECT_EQ(expectedValues, collectedValues);
}

TEST(ColumnCollectorTest, CollectTransitionsTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<EventLog> log = Generators::generateEventLog(catVariable, numVariable);

    ColumnCollector collector;
    auto result = collector.collectTransitions<double_t>(log.get(), numVariable);

    std::multiset<std::pair<double_t, double_t>> collectedValues(result.begin(), result.end());
    std::multiset<std::pair<double_t, double_t>> expectedValues{
        {9,9}, {9,7}, {7,7}, {1,3}, {3,3}, {3,3}, {3,5}, {5,6}};

    EXPECT_EQ(expectedValues, collectedValues);
}

TEST(ColumnCollectorTest, CollectTransitionsMissingValuesTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<EventLog> log = Generators::generateEventLogWithMissingValues(
        catVariable, numVariable);

    ColumnCollector collector;
    auto result = collector.collectTransitions<std::string>(log.get(), catVariable);

    std::multiset<std::pair<std::string, std::string>> collectedValues(result.begin(), result.end());
    std::multiset<std::pair<std::string, std::string>> expectedValues{
        {"u", "e"}, {"i", "a"}, {"e", "a"}};

    EXPECT_EQ(expectedValues, collectedValues);
}

TEST(ColumnCollectorTest, CollectTransitionsNoTransitionTest)
{
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    EventLog::TracesType traces;
    Trace::ColumnsType columns;
    columns[numVariable] = std::make_unique<LogColumn<double_t>>(std::vector<double_t>{42.0});
    traces["0"] = std::make_unique<Trace>(columns);
    std::unique_ptr<EventLog> log = std::make_unique<EventLog>(traces);

    ColumnCollector collector;
    auto result = collector.collectTransitions<double_t>(log.get(), numVariable);

    EXPECT_EQ(0, result.size());
}

TEST(ColumnCollectorTest, CollectAllTransitionsTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<CategoricalVariable> otherVariable = std::make_shared<CategoricalVariable>("other",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    EventLog::TracesType traces;
    traces["0"] = std::move(Generators::generateComplexTrace(catVariable, otherVariable));
    std::unique_ptr<EventLog> log = std::make_unique<EventLog>(traces);

    ColumnCollector collector;
    auto result = collector.collectAllTransitions<std::string>(log.get());

    std::multiset<std::pair<std::string, std::string>> collectedCatValues(
        result[catVariable].begin(), result[catVariable].end());
    std::multiset<std::pair<std::string, std::string>> expectedCatValues{
        {"u","a"}, {"a","u"}, {"u","e"}, {"e","a"}, {"a","e"}, {"e","u"}};

    EXPECT_EQ(expectedCatValues, collectedCatValues);

    std::multiset<std::pair<std::string, std::string>> collectedOtherValues(
        result[otherVariable].begin(), result[otherVariable].end());
    std::multiset<std::pair<std::string, std::string>> expectedOtherValues{
        {"u","i"}, {"i","i"}, {"i","a"}, {"a","u"}, {"u","e"}, {"e","u"}};

    EXPECT_EQ(expectedOtherValues, collectedOtherValues);
}

TEST(ColumnCollectorTest, CollectForModificationRuleWithValueSetRuleTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<EventLog> log = Generators::generateEventLog(catVariable, numVariable);

    Condition::ElementsType elements;
    elements.push_back(std::make_unique<ValueCondition<double_t>>(numVariable, 7));
    std::unique_ptr<Condition> condition = std::make_unique<Condition>(elements);
    std::unique_ptr<UpdateRule> updateRule = std::make_unique<ValueSetRule>(
        catVariable, std::set<std::string>{});

    ColumnCollector collector;
    std::vector<std::string> result = collector.collectValuesForModificationRule<std::string>(
        log.get(), condition.get(), updateRule.get());

    std::multiset<std::string> collectedValues(result.begin(), result.end());
    std::multiset<std::string> expectedValues{"u", "e"};

    EXPECT_EQ(expectedValues, collectedValues);
}

TEST(ColumnCollectorTest, CollectForModificationRuleWithValueRuleTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<EventLog> log = Generators::generateEventLog(catVariable, numVariable);

    Condition::ElementsType elements;
    elements.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "u"));
    std::unique_ptr<Condition> condition = std::make_unique<Condition>(elements);
    std::unique_ptr<UpdateRule> updateRule = std::make_unique<ValueRule>(numVariable, 0.0);

    ColumnCollector collector;
    std::vector<double_t> result = collector.collectValuesForModificationRule<double_t>(
        log.get(), condition.get(), updateRule.get());

    std::multiset<double_t> collectedValues(result.begin(), result.end());
    std::multiset<double_t> expectedValues{9, 7, 3, 6};

    EXPECT_EQ(expectedValues, collectedValues);
}

TEST(ColumnCollectorTest, CollectForModificationRuleWithValueRangeRuleTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<EventLog> log = Generators::generateEventLog(catVariable, numVariable);

    Condition::ElementsType elements;
    elements.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "u"));
    std::unique_ptr<Condition> condition = std::make_unique<Condition>(elements);
    std::unique_ptr<UpdateRule> updateRule = std::make_unique<ValueRangeRule>(numVariable, 0.0, 1.0);

    ColumnCollector collector;
    std::vector<double_t> result = collector.collectValuesForModificationRule<double_t>(
        log.get(), condition.get(), updateRule.get());

    std::multiset<double_t> collectedValues(result.begin(), result.end());
    std::multiset<double_t> expectedValues{9, 7, 3, 6};

    EXPECT_EQ(expectedValues, collectedValues);
}

TEST(ColumnCollectorTest, CollectForModificationRuleWithFactorRuleTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<EventLog> log = Generators::generateEventLog(catVariable, numVariable);

    Condition::ElementsType elements;
    elements.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "u"));
    std::unique_ptr<Condition> condition = std::make_unique<Condition>(elements);
    std::unique_ptr<UpdateRule> updateRule = std::make_unique<FactorRule>(numVariable, 1.0);

    ColumnCollector collector;
    std::vector<double_t> result = collector.collectValuesForModificationRule<double_t>(
        log.get(), condition.get(), updateRule.get());

    std::multiset<double_t> collectedValues(result.begin(), result.end());
    std::multiset<double_t> expectedValues{7.0/9.0, 3.0/3.0, 6.0/5.0};

    EXPECT_EQ(expectedValues, collectedValues);
}

TEST(ColumnCollectorTest, CollectForModificationRuleWithFactorRuleAndZerosTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<CategoricalVariable> otherVariable = std::make_shared<CategoricalVariable>("other",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<EventLog> log = Generators::generateEventLogWithNumericalUpdateRule(
        catVariable, otherVariable, numVariable);

    std::unique_ptr<Condition> condition = std::make_unique<Condition>();
    std::unique_ptr<UpdateRule> updateRule = std::make_unique<FactorRule>(numVariable, 1.0);

    ColumnCollector collector;
    std::vector<double_t> result = collector.collectValuesForModificationRule<double_t>(
        log.get(), condition.get(), updateRule.get());

    std::multiset<double_t> collectedValues(result.begin(), result.end());
    std::multiset<double_t> expectedValues;
    for (size_t i = 0; i < 6 * 40; i++)
    {
        expectedValues.insert(0.0);
    }

    EXPECT_EQ(expectedValues, collectedValues);
}

TEST(ColumnCollectorTest, CollectForModificationRuleWithDifferenceRuleTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<EventLog> log = Generators::generateEventLog(catVariable, numVariable);

    Condition::ElementsType elements;
    elements.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "u"));
    std::unique_ptr<Condition> condition = std::make_unique<Condition>(elements);
    std::unique_ptr<UpdateRule> updateRule = std::make_unique<DifferenceRule>(numVariable, 1.0);

    ColumnCollector collector;
    std::vector<double_t> result = collector.collectValuesForModificationRule<double_t>(
        log.get(), condition.get(), updateRule.get());

    std::multiset<double_t> collectedValues(result.begin(), result.end());
    std::multiset<double_t> expectedValues{7.0 - 9.0, 3.0 - 3.0, 6.0 - 5.0};

    EXPECT_EQ(expectedValues, collectedValues);
}

TEST(ColumnCollectorTest, CollectForModificationRuleWithChangeRangeRuleTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<EventLog> log = Generators::generateEventLog(catVariable, numVariable);

    Condition::ElementsType elements;
    elements.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "u"));
    std::unique_ptr<Condition> condition = std::make_unique<Condition>(elements);
    std::unique_ptr<UpdateRule> updateRule = std::make_unique<ChangeRangeRule>(numVariable, 0.0, 1.0);

    ColumnCollector collector;
    std::vector<double_t> result = collector.collectValuesForModificationRule<double_t>(
        log.get(), condition.get(), updateRule.get());

    std::multiset<double_t> collectedValues(result.begin(), result.end());
    std::multiset<double_t> expectedValues{7.0 - 9.0, 3.0 - 3.0, 6.0 - 5.0};

    EXPECT_EQ(expectedValues, collectedValues);
}
