#include <gtest/gtest.h>
#include "generators.hpp"
#include "../classes/scoring/conditionmatcher.hpp"
#include "../classes/datamodel/conditions/valuecondition.hpp"
#include "../classes/datamodel/conditions/upperboundcondition.hpp"
#include "../classes/datamodel/conditions/lowerboundcondition.hpp"
#include "../classes/datamodel/conditions/transitioncondition.hpp"
#include "../classes/datamodel/eventlogs/logcolumn.hpp"
#include "../classes/datamodel/conditions/andoperator.hpp"
#include "../classes/datamodel/conditions/oroperator.hpp"
#include "../classes/datamodel/conditions/notoperator.hpp"

TEST(ConditionMatcherTest, ValueEqualityTest)
{
    std::set<std::string> vowelDomain = {"a", "e", "i", "o", "u"};
    std::shared_ptr<CategoricalVariable> catVariable =
        std::make_shared<CategoricalVariable>("vowel", vowelDomain);
    std::unique_ptr<ValueCondition<std::string>> cond =
        std::make_unique<ValueCondition<std::string>>(catVariable, "o");

    std::vector<std::string> entries {"o"};
    Trace::ColumnsType columns;
    columns.insert({catVariable, std::make_unique<LogColumn<std::string>>(entries)});
    std::unique_ptr<Trace> trace = std::make_unique<Trace>(columns);

    ConditionMatcher matcher;
    EXPECT_TRUE(matcher.matches(cond.get(), trace.get(), 0));
}

TEST(ConditionMatcherTest, ValueNonEqualityTest)
{
    std::set<std::string> vowelDomain = {"a", "e", "i", "o", "u"};
    std::shared_ptr<CategoricalVariable> catVariable =
        std::make_shared<CategoricalVariable>("vowel", vowelDomain);
    std::unique_ptr<ValueCondition<std::string>> cond =
        std::make_unique<ValueCondition<std::string>>(catVariable, "o");

    std::vector<std::string> entries {"e"};
    Trace::ColumnsType columns;
    columns.insert({catVariable, std::make_unique<LogColumn<std::string>>(entries)});
    std::unique_ptr<Trace> trace = std::make_unique<Trace>(columns);

    ConditionMatcher matcher;
    EXPECT_FALSE(matcher.matches(cond.get(), trace.get(), 0));
}

TEST(ConditionMatcherTest, LowerBoundBelowTest)
{
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<LowerBoundCondition> cond =
        std::make_unique<LowerBoundCondition>(numVariable, 4.0);
    std::unique_ptr<Trace> trace = Generators::generateBelowTrace(numVariable);

    ConditionMatcher matcher;
    EXPECT_FALSE(matcher.matches(cond.get(), trace.get(), 0));
}

TEST(ConditionMatcherTest, UpperBoundBelowTest)
{
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<UpperBoundCondition> cond =
        std::make_unique<UpperBoundCondition>(numVariable, 4.0);
    std::unique_ptr<Trace> trace = Generators::generateBelowTrace(numVariable);

    ConditionMatcher matcher;
    EXPECT_TRUE(matcher.matches(cond.get(), trace.get(), 0));
}

TEST(ConditionMatcherTest, LowerBoundEqualTest)
{
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<LowerBoundCondition> cond =
        std::make_unique<LowerBoundCondition>(numVariable, 4.0);
    std::unique_ptr<Trace> trace = Generators::generateEqualTrace(numVariable);

    ConditionMatcher matcher;
    EXPECT_TRUE(matcher.matches(cond.get(), trace.get(), 0));
}

TEST(ConditionMatcherTest, UpperBoundEqualTest)
{
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<UpperBoundCondition> cond =
        std::make_unique<UpperBoundCondition>(numVariable, 4.0);
    std::unique_ptr<Trace> trace = Generators::generateEqualTrace(numVariable);

    ConditionMatcher matcher;
    EXPECT_TRUE(matcher.matches(cond.get(), trace.get(), 0));
}

TEST(ConditionMatcherTest, LowerBoundAboveTest)
{
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<LowerBoundCondition> cond =
        std::make_unique<LowerBoundCondition>(numVariable, 4.0);
    std::unique_ptr<Trace> trace = Generators::generateAboveTrace(numVariable);

    ConditionMatcher matcher;
    EXPECT_TRUE(matcher.matches(cond.get(), trace.get(), 0));
}

TEST(ConditionMatcherTest, UpperBoundAboveTest)
{
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<UpperBoundCondition> cond =
        std::make_unique<UpperBoundCondition>(numVariable, 4.0);
    std::unique_ptr<Trace> trace = Generators::generateAboveTrace(numVariable);

    ConditionMatcher matcher;
    EXPECT_FALSE(matcher.matches(cond.get(), trace.get(), 0));
}

TEST(ConditionMatcherTest, TransitionMatchTest)
{
    std::pair<std::unique_ptr<AtomicCondition>, std::unique_ptr<Trace>> data =
        Generators::generateTransitionData();

    ConditionMatcher matcher;
    EXPECT_TRUE(matcher.matches(data.first.get(), data.second.get(), 2));
}

TEST(ConditionMatcherTest, EarlyTransitionTest)
{
    std::pair<std::unique_ptr<AtomicCondition>, std::unique_ptr<Trace>> data =
        Generators::generateTransitionData();

    ConditionMatcher matcher;
    EXPECT_FALSE(matcher.matches(data.first.get(), data.second.get(), 3));
}

TEST(ConditionMatcherTest, LateTransitionTest)
{
    std::pair<std::unique_ptr<AtomicCondition>, std::unique_ptr<Trace>> data =
        Generators::generateTransitionData();

    ConditionMatcher matcher;
    EXPECT_FALSE(matcher.matches(data.first.get(), data.second.get(), 1));
}

TEST(ConditionMatcherTest, DifferentTransitionTest)
{
    std::pair<std::unique_ptr<AtomicCondition>, std::unique_ptr<Trace>> data =
        Generators::generateTransitionData();

    ConditionMatcher matcher;
    EXPECT_FALSE(matcher.matches(data.first.get(), data.second.get(), 6));
}

TEST(ConditionMatcherTest, FirstValueTransitionTest)
{
    std::pair<std::unique_ptr<AtomicCondition>, std::unique_ptr<Trace>> data =
        Generators::generateTransitionData();

    ConditionMatcher matcher;
    EXPECT_FALSE(matcher.matches(data.first.get(), data.second.get(), 0));
}

TEST(ConditionMatcherTest, TransitionInvalidIndexTest)
{
    std::pair<std::unique_ptr<AtomicCondition>, std::unique_ptr<Trace>> data =
        Generators::generateTransitionData();

    ConditionMatcher matcher;
    EXPECT_THROW(matcher.matches(data.first.get(), data.second.get(), 7), std::out_of_range);
}

TEST(ConditionMatcherTest, TransitionWrongVariableTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::unique_ptr<Trace> trace = Generators::generateComplexTrace(catVariable);
    std::unique_ptr<TransitionCondition<std::string>> cond =
        std::make_unique<TransitionCondition<std::string>>(catVariable, "a", "u");

    ConditionMatcher matcher;
    EXPECT_FALSE(matcher.matches(cond.get(), trace.get(), 4));
}

TEST(ConditionMatcherTest, AndConditionTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<CategoricalVariable> otherVariable = std::make_shared<CategoricalVariable>("other",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::unique_ptr<Trace> trace = Generators::generateComplexTrace(catVariable, otherVariable);

    Condition::ElementsType elements;
    elements.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "a"));
    elements.push_back(std::make_unique<ValueCondition<std::string>>(otherVariable, "i"));
    elements.push_back(std::make_unique<AndOperator>());

    std::unique_ptr<Condition> cond = std::make_unique<Condition>(elements);

    ConditionMatcher matcher;
    EXPECT_FALSE(matcher.matches(cond.get(), trace.get(), 0));
    EXPECT_TRUE(matcher.matches(cond.get(), trace.get(), 1));
    EXPECT_FALSE(matcher.matches(cond.get(), trace.get(), 2));
    EXPECT_FALSE(matcher.matches(cond.get(), trace.get(), 4));
}

TEST(ConditionMatcherTest, OrConditionTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<CategoricalVariable> otherVariable = std::make_shared<CategoricalVariable>("other",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::unique_ptr<Trace> trace = Generators::generateComplexTrace(catVariable, otherVariable);

    Condition::ElementsType elements;
    elements.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "a"));
    elements.push_back(std::make_unique<ValueCondition<std::string>>(otherVariable, "i"));
    elements.push_back(std::make_unique<OrOperator>());

    std::unique_ptr<Condition> cond = std::make_unique<Condition>(elements);

    ConditionMatcher matcher;
    EXPECT_FALSE(matcher.matches(cond.get(), trace.get(), 0));
    EXPECT_TRUE(matcher.matches(cond.get(), trace.get(), 1));
    EXPECT_TRUE(matcher.matches(cond.get(), trace.get(), 2));
    EXPECT_TRUE(matcher.matches(cond.get(), trace.get(), 4));
}

TEST(ConditionMatcherTest, ComplexConditionTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<CategoricalVariable> otherVariable = std::make_shared<CategoricalVariable>("other",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<Trace> trace = Generators::generateComplexTrace(catVariable, otherVariable, numVariable);

    Condition::ElementsType elements;
    elements.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "a"));
    elements.push_back(std::make_unique<ValueCondition<std::string>>(otherVariable, "i"));
    elements.push_back(std::make_unique<AndOperator>());
    elements.push_back(std::make_unique<ValueCondition<double_t>>(numVariable, 7.0));
    elements.push_back(std::make_unique<OrOperator>());

    std::unique_ptr<Condition> cond = std::make_unique<Condition>(elements);

    ConditionMatcher matcher;
    EXPECT_FALSE(matcher.matches(cond.get(), trace.get(), 0));
    EXPECT_TRUE(matcher.matches(cond.get(), trace.get(), 1));
    EXPECT_FALSE(matcher.matches(cond.get(), trace.get(), 2));
    EXPECT_FALSE(matcher.matches(cond.get(), trace.get(), 4));
    EXPECT_TRUE(matcher.matches(cond.get(), trace.get(), 6));
}

TEST(ConditionMatcherTest, TooFewConditionOperatorsTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<CategoricalVariable> otherVariable = std::make_shared<CategoricalVariable>("other",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::unique_ptr<Trace> trace = Generators::generateComplexTrace(catVariable, otherVariable);

    Condition::ElementsType elements;
    elements.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "a"));
    elements.push_back(std::make_unique<ValueCondition<std::string>>(otherVariable, "i"));

    std::unique_ptr<Condition> cond = std::make_unique<Condition>(elements);

    ConditionMatcher matcher;
    EXPECT_THROW(matcher.matches(cond.get(), trace.get(), 0), std::invalid_argument);
}

TEST(ConditionMatcherTest, TooManyConditionOperatorsTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<CategoricalVariable> otherVariable = std::make_shared<CategoricalVariable>("other",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::unique_ptr<Trace> trace = Generators::generateComplexTrace(catVariable, otherVariable);

    Condition::ElementsType elements;
    elements.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "a"));
    elements.push_back(std::make_unique<ValueCondition<std::string>>(otherVariable, "i"));
    elements.push_back(std::make_unique<AndOperator>());
    elements.push_back(std::make_unique<AndOperator>());

    std::unique_ptr<Condition> cond = std::make_unique<Condition>(elements);

    ConditionMatcher matcher;
    EXPECT_THROW(matcher.matches(cond.get(), trace.get(), 0), std::invalid_argument);
}

TEST(ConditionMatcherTest, EmptyConditionTest)
{
    std::unique_ptr<Trace> trace = Generators::generateComplexTrace();
    std::unique_ptr<Condition> cond = std::make_unique<Condition>();

    ConditionMatcher matcher;
    for (size_t i = 0; i < 7; i++)
    {
        EXPECT_TRUE(matcher.matches(cond.get(), trace.get(), i));
    }
}

TEST(ConditionMatcherTest, ConditionInvalidIndexTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::unique_ptr<Trace> trace = Generators::generateComplexTrace(catVariable);

    Condition::ElementsType elements;
    elements.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "a"));
    std::unique_ptr<Condition> cond = std::make_unique<Condition>(elements);

    ConditionMatcher matcher;
    EXPECT_THROW(matcher.matches(cond.get(), trace.get(), 7), std::out_of_range);
}

TEST(ConditionMatcherTest, NotConditionTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::unique_ptr<Trace> trace = Generators::generateComplexTrace(catVariable);

    Condition::ElementsType elements;
    elements.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "a"));
    elements.push_back(std::make_unique<NotOperator>());

    std::unique_ptr<Condition> cond = std::make_unique<Condition>(elements);

    ConditionMatcher matcher;
    EXPECT_TRUE(matcher.matches(cond.get(), trace.get(), 0));
    EXPECT_FALSE(matcher.matches(cond.get(), trace.get(), 1));
}

TEST(ConditionMatcherTest, NotTooFewOperandsTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::unique_ptr<Trace> trace = Generators::generateComplexTrace(catVariable);

    Condition::ElementsType elements;
    elements.push_back(std::make_unique<NotOperator>());

    std::unique_ptr<Condition> cond = std::make_unique<Condition>(elements);

    ConditionMatcher matcher;
    EXPECT_THROW(matcher.matches(cond.get(), trace.get(), 0), std::invalid_argument);
}

TEST(ConditionMatcherTest, MissingValuesTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    auto log = Generators::generateEventLogWithMissingValues(catVariable);
    const auto& trace = log->getTraces().at("1");

    Condition::ElementsType elements;
    elements.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "a"));
    elements.push_back(std::make_unique<NotOperator>());

    std::unique_ptr<Condition> cond = std::make_unique<Condition>(elements);

    ConditionMatcher matcher;
    EXPECT_TRUE(matcher.matches(cond.get(), trace.get(), 0));
    EXPECT_FALSE(matcher.matches(cond.get(), trace.get(), 1));
    EXPECT_FALSE(matcher.matches(cond.get(), trace.get(), 2));
    EXPECT_TRUE(matcher.matches(cond.get(), trace.get(), 3));
    EXPECT_FALSE(matcher.matches(cond.get(), trace.get(), 4));
    EXPECT_FALSE(matcher.matches(cond.get(), trace.get(), 5));
}

TEST(ConditionMatcherTest, TransitionMissingValuesTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    auto log = Generators::generateEventLogWithMissingValues(catVariable);
    const auto& trace = log->getTraces().at("1");
    
    Condition::ElementsType elements;
    elements.push_back(std::make_unique<TransitionCondition<std::string>>(catVariable, "e", "a"));
    elements.push_back(std::make_unique<NotOperator>());

    std::unique_ptr<Condition> cond = std::make_unique<Condition>(elements);

    ConditionMatcher matcher;
    EXPECT_FALSE(matcher.matches(cond.get(), trace.get(), 0));
    EXPECT_TRUE(matcher.matches(cond.get(), trace.get(), 1));
    EXPECT_FALSE(matcher.matches(cond.get(), trace.get(), 2));
    EXPECT_FALSE(matcher.matches(cond.get(), trace.get(), 3));
    EXPECT_FALSE(matcher.matches(cond.get(), trace.get(), 4));
    EXPECT_FALSE(matcher.matches(cond.get(), trace.get(), 5));
}
