#include <gtest/gtest.h>
#include <memory>
#include "generators.hpp"
#include "../classes/preprocessing/cycledetector.hpp"
#include "../classes/datamodel/conditions/valuecondition.hpp"
#include "../classes/datamodel/updaterules/valuerule.hpp"

TEST(CycleDetectorTest, EmptyModelAdjacencyListTest)
{
    CycleDetector detector;
    auto model = std::make_unique<RuleModel>();
    CycleDetector::AdjacencyList list = detector.computeAdjacencyList(model.get());
    EXPECT_EQ(0, list.size());
}

TEST(CycleDetectorTest, SingletonModelAdjacencyListTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    
    RuleModel::RulesType rules;

    Condition::ElementsType elements;
    std::unique_ptr<Condition> condition;
    std::unique_ptr<UpdateRule> updateRule;

    elements = Condition::ElementsType();
    elements.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "o"));
    condition = std::make_unique<Condition>(elements);
    updateRule = std::make_unique<ValueRule>(numVariable, 42.0);
    rules.push_back(std::make_unique<ModificationRule>(condition, updateRule));

    CycleDetector detector;
    auto model = std::make_unique<RuleModel>(rules);
    CycleDetector::AdjacencyList list = detector.computeAdjacencyList(model.get());

    CycleDetector::AdjacencyList expectedList {
        {"vowel", {"num"}}
    };

    EXPECT_EQ(expectedList, list);
}

TEST(CycleDetectorTest, EmptyConditionModelAdjacencyListTest)
{
    CycleDetector detector;
    auto model = Generators::generateRuleModel();
    CycleDetector::AdjacencyList list = detector.computeAdjacencyList(model.get());

    EXPECT_EQ(0, list.size());
}

TEST(CycleDetectorTest, ComplexModelAdjacencyListTest)
{
    CycleDetector detector;
    auto model = Generators::generateCompleteRuleModel();
    CycleDetector::AdjacencyList list = detector.computeAdjacencyList(model.get());

    CycleDetector::AdjacencyList expectedList {
        {"num", {"num"}},
        {"vowel", {"vowel", "num"}}
    };

    EXPECT_EQ(expectedList, list);
}

TEST(CycleDetectorTest, EmptyCyclicTest)
{
    CycleDetector detector;
    CycleDetector::AdjacencyList list;

    EXPECT_FALSE(detector.hasCycle(list));
}

TEST(CycleDetectorTest, DiamondCyclicTest)
{
    CycleDetector detector;
    CycleDetector::AdjacencyList list {
        {"a", {"b", "c"}},
        {"b", {"d"}},
        {"c", {"d"}}
    };

    EXPECT_FALSE(detector.hasCycle(list));
}

TEST(CycleDetectorTest, SelfCycleCyclicTest)
{
    CycleDetector detector;
    CycleDetector::AdjacencyList list {
        {"a", {"a", "b"}}
    };

    EXPECT_TRUE(detector.hasCycle(list));
}

TEST(CycleDetectorTest, SimpleCycleCyclicTest)
{
    CycleDetector detector;
    CycleDetector::AdjacencyList list {
        {"a", {"b"}},
        {"b", {"a"}}
    };

    EXPECT_TRUE(detector.hasCycle(list));
}

TEST(CycleDetectorTest, ComplexCycleCyclicTest)
{
    CycleDetector detector;
    CycleDetector::AdjacencyList list {
        {"a", {"b", "c"}},
        {"b", {"d"}},
        {"c", {"d"}},
        {"d", {"e"}},
        {"e", {"a"}}
    };

    EXPECT_TRUE(detector.hasCycle(list));
}

TEST(CycleDetectorTest, CyclicRuleModelTest)
{
    CycleDetector detector;
    auto model = Generators::generateCompleteRuleModel();

    EXPECT_TRUE(detector.hasCycle(model.get()));
}

TEST(CycleDetectorTest, AddEdgesTest)
{
    std::shared_ptr<LogVariable> varA = std::make_shared<NumericalVariable>("a");
    std::shared_ptr<LogVariable> varB = std::make_shared<NumericalVariable>("b");
    std::shared_ptr<LogVariable> varC = std::make_shared<NumericalVariable>("c");
    std::shared_ptr<LogVariable> varD = std::make_shared<NumericalVariable>("d");
    std::shared_ptr<LogVariable> varE = std::make_shared<NumericalVariable>("e");
    CycleDetector::AdjacencyList list {
        {"a", {"b", "c"}},
        {"d", {"e"}}
    };
    CycleDetector detector;
    std::set<std::shared_ptr<LogVariable>> conditionVariables {varB, varD};
    detector.addEdgesToAdjacencyList(list, conditionVariables, varA);

    EXPECT_EQ(CycleDetector::AdjacencyList({
        {"a", {"b", "c"}},
        {"b", {"a"}},
        {"d", {"a", "e"}}
    }), list);
}
