#include <gtest/gtest.h>
#include "generators.hpp"
#include "../classes/scoring/dataencoder.hpp"
#include "mock_distributionestimator.hpp"
#include "mock_rulefinder.hpp"
#include "mock_updaterulematcher.hpp"

using ::testing::Return;
using ::testing::_;

TEST(DataEncoderTest, EmptyModelTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<EventLog> log = Generators::generateEventLog(catVariable, numVariable);

    std::shared_ptr<MdlCalculator> calc = std::make_shared<MdlCalculator>();
    std::shared_ptr<DistributionEstimator> distributionEstimator =
        std::make_shared<MockDistributionEstimator>();
    std::unique_ptr<RuleFinder> ruleFinder = std::make_unique<RuleFinder>();
    std::unique_ptr<UpdateRuleMatcher> ruleMatcher = std::make_unique<UpdateRuleMatcher>();

    MockDistributionEstimator* distributionEstimatorRef =
        static_cast<MockDistributionEstimator*>(distributionEstimator.get());

    std::unique_ptr<RuleModel> model = std::make_unique<RuleModel>();
    DataEncoder dataEnc(calc, distributionEstimator, ruleFinder, ruleMatcher);

    EXPECT_CALL(*distributionEstimatorRef, getGlobalProbability(catVariable, _)).Times(10)
        .WillOnce(Return(1.0/2)).WillOnce(Return(1.0/4)).WillOnce(Return(1.0/8))
        .WillOnce(Return(1.0/16)).WillOnce(Return(1.0/32)).WillOnce(Return(1.0/64))
        .WillOnce(Return(1.0/128)).WillOnce(Return(1.0/256)).WillOnce(Return(1.0/512))
        .WillOnce(Return(1.0/1024));
    EXPECT_CALL(*distributionEstimatorRef, getGlobalProbability(numVariable, _)).Times(10)
        .WillOnce(Return(1.0/2)).WillOnce(Return(1.0/4)).WillOnce(Return(1.0/8))
        .WillOnce(Return(1.0/16)).WillOnce(Return(1.0/32)).WillOnce(Return(1.0/64))
        .WillOnce(Return(1.0/128)).WillOnce(Return(1.0/256)).WillOnce(Return(1.0/512))
        .WillOnce(Return(1.0/1024));

    EXPECT_DOUBLE_EQ(2*(1+2+3+4+5+6+7+8+9+10), dataEnc.encodeEventLog(log.get(), model.get()));
}

TEST(DataEncoderTest, MissingValuesTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<EventLog> log = Generators::generateEventLogWithMissingValues(
        catVariable, numVariable);

    std::shared_ptr<MdlCalculator> calc = std::make_shared<MdlCalculator>();
    std::shared_ptr<DistributionEstimator> distributionEstimator =
        std::make_shared<MockDistributionEstimator>();
    std::unique_ptr<RuleFinder> ruleFinder = std::make_unique<MockRuleFinder>();
    std::unique_ptr<UpdateRuleMatcher> ruleMatcher = std::make_unique<MockUpdateRuleMatcher>();

    MockDistributionEstimator* distributionEstimatorRef =
        static_cast<MockDistributionEstimator*>(distributionEstimator.get());
    MockRuleFinder* ruleFinderRef =
        static_cast<MockRuleFinder*>(ruleFinder.get());
    MockUpdateRuleMatcher* ruleMatcherRef =
        static_cast<MockUpdateRuleMatcher*>(ruleMatcher.get());

    std::unique_ptr<RuleModel> model = Generators::generateRuleModel(catVariable, numVariable);
    DataEncoder dataEnc(calc, distributionEstimator, ruleFinder, ruleMatcher);

    EXPECT_CALL(*ruleFinderRef, findRulesThatApply(model.get(), catVariable.get(), _, _)).Times(7)
        .WillRepeatedly(Return(std::vector<const ModificationRule*>{model->getRules()[1]}));
    EXPECT_CALL(*ruleFinderRef, findRulesThatApply(model.get(), numVariable.get(), _, _)).Times(7)
        .WillRepeatedly(Return(std::vector<const ModificationRule*>{model->getRules()[2]}));
    EXPECT_CALL(*ruleMatcherRef, matches(static_cast<const ValueSetRule*>(
        model->getRules()[1]->getUpdateRule()), _, _)).Times(7).WillRepeatedly(Return(true));
    EXPECT_CALL(*ruleMatcherRef, matches(static_cast<const AbstractRangeRule*>(
        model->getRules()[2]->getUpdateRule()), _, _)).Times(7).WillRepeatedly(Return(true));
   EXPECT_CALL(*distributionEstimatorRef, getRuleProbability(static_cast<const ValueSetRule*>(
        model->getRules()[1]->getUpdateRule()), _)).Times(7)
        .WillOnce(Return(1.0/2)).WillOnce(Return(1.0/4)).WillOnce(Return(1.0/8))
        .WillOnce(Return(1.0/16)).WillOnce(Return(1.0/32)).WillOnce(Return(1.0/64))
        .WillOnce(Return(1.0/128));
    EXPECT_CALL(*distributionEstimatorRef, getRuleProbability(static_cast<const AbstractRangeRule*>(
        model->getRules()[2]->getUpdateRule()))).Times(7)
        .WillRepeatedly(Return(1.0/7.0));

    const std::unordered_map<bool, int64_t> expectedMatches{{true, 7}, {false, 0}};
    const double_t expectedModelStreamLength = 2*calc->codePrequential(expectedMatches);

    EXPECT_DOUBLE_EQ(1+2+3+4+5+6+7 + 7*std::log2(7) + expectedModelStreamLength,
        dataEnc.encodeEventLog(log.get(), model.get()));
}

TEST(DataEncoderTest, AlwaysMatchingModelTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<EventLog> log = Generators::generateEventLog(catVariable, numVariable);
    numVariable->setAccuracy(0.1);

    std::shared_ptr<MdlCalculator> calc = std::make_shared<MdlCalculator>();
    std::shared_ptr<DistributionEstimator> distributionEstimator =
        std::make_shared<MockDistributionEstimator>();
    std::unique_ptr<RuleFinder> ruleFinder = std::make_unique<MockRuleFinder>();
    std::unique_ptr<UpdateRuleMatcher> ruleMatcher = std::make_unique<MockUpdateRuleMatcher>();

    MockDistributionEstimator* distributionEstimatorRef =
        static_cast<MockDistributionEstimator*>(distributionEstimator.get());
    MockRuleFinder* ruleFinderRef =
        static_cast<MockRuleFinder*>(ruleFinder.get());
    MockUpdateRuleMatcher* ruleMatcherRef =
        static_cast<MockUpdateRuleMatcher*>(ruleMatcher.get());

    std::unique_ptr<RuleModel> model = Generators::generateRuleModel(catVariable, numVariable);
    DataEncoder dataEnc(calc, distributionEstimator, ruleFinder, ruleMatcher);

    EXPECT_CALL(*ruleFinderRef, findRulesThatApply(model.get(), catVariable.get(), _, _)).Times(10)
        .WillRepeatedly(Return(std::vector<const ModificationRule*>{model->getRules()[1]}));
    EXPECT_CALL(*ruleFinderRef, findRulesThatApply(model.get(), numVariable.get(), _, _)).Times(10)
        .WillRepeatedly(Return(std::vector<const ModificationRule*>{model->getRules()[2]}));
    EXPECT_CALL(*ruleMatcherRef, matches(static_cast<const ValueSetRule*>(
        model->getRules()[1]->getUpdateRule()), _, _)).Times(10).WillRepeatedly(Return(true));
    EXPECT_CALL(*ruleMatcherRef, matches(static_cast<const AbstractRangeRule*>(
        model->getRules()[2]->getUpdateRule()), _, _)).Times(10).WillRepeatedly(Return(true));
    EXPECT_CALL(*distributionEstimatorRef, getRuleProbability(static_cast<const ValueSetRule*>(
        model->getRules()[1]->getUpdateRule()), _)).Times(10)
        .WillOnce(Return(1.0/2)).WillOnce(Return(1.0/4)).WillOnce(Return(1.0/8))
        .WillOnce(Return(1.0/16)).WillOnce(Return(1.0/32)).WillOnce(Return(1.0/64))
        .WillOnce(Return(1.0/128)).WillOnce(Return(1.0/256)).WillOnce(Return(1.0/512))
        .WillOnce(Return(1.0/1024));
    EXPECT_CALL(*distributionEstimatorRef, getRuleProbability(static_cast<const AbstractRangeRule*>(
        model->getRules()[2]->getUpdateRule()))).Times(10)
        .WillRepeatedly(Return(1.0/7.0));

    const std::unordered_map<bool, int64_t> expectedMatches{{true, 10}, {false, 0}};
    const double_t expectedModelStreamLength = 2*calc->codePrequential(expectedMatches);

    EXPECT_DOUBLE_EQ(1+2+3+4+5+6+7+8+9+10+10*std::log2(7)+expectedModelStreamLength,
        dataEnc.encodeEventLog(log.get(), model.get()));
}

TEST(DataEncoderTest, PointPredictingRuleTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<EventLog> log = Generators::generateEventLog(catVariable, numVariable);
    numVariable->setAccuracy(0.1);

    std::shared_ptr<MdlCalculator> calc = std::make_shared<MdlCalculator>();
    std::shared_ptr<DistributionEstimator> distributionEstimator =
        std::make_shared<MockDistributionEstimator>();
    std::unique_ptr<RuleFinder> ruleFinder = std::make_unique<MockRuleFinder>();
    std::unique_ptr<UpdateRuleMatcher> ruleMatcher = std::make_unique<MockUpdateRuleMatcher>();

    MockDistributionEstimator* distributionEstimatorRef =
        static_cast<MockDistributionEstimator*>(distributionEstimator.get());
    MockRuleFinder* ruleFinderRef =
        static_cast<MockRuleFinder*>(ruleFinder.get());
    MockUpdateRuleMatcher* ruleMatcherRef =
        static_cast<MockUpdateRuleMatcher*>(ruleMatcher.get());

    std::unique_ptr<RuleModel> model = Generators::generateRuleModel(catVariable, numVariable);
    DataEncoder dataEnc(calc, distributionEstimator, ruleFinder, ruleMatcher);

    EXPECT_CALL(*ruleFinderRef, findRulesThatApply(model.get(), catVariable.get(), _, _)).Times(10)
        .WillRepeatedly(Return(std::vector<const ModificationRule*>{model->getRules()[1]}));
    EXPECT_CALL(*ruleFinderRef, findRulesThatApply(model.get(), numVariable.get(), _, _)).Times(10)
        .WillRepeatedly(Return(std::vector<const ModificationRule*>{model->getRules()[0]}));
    EXPECT_CALL(*ruleMatcherRef, matches(static_cast<const ValueSetRule*>(
        model->getRules()[1]->getUpdateRule()), _, _)).Times(10).WillRepeatedly(Return(true));
    EXPECT_CALL(*ruleMatcherRef, matches(static_cast<const AbstractValueRule*>(
        model->getRules()[0]->getUpdateRule()), _, _)).Times(10).WillRepeatedly(Return(true));
    EXPECT_CALL(*distributionEstimatorRef, getRuleProbability(static_cast<const ValueSetRule*>(
        model->getRules()[1]->getUpdateRule()), _)).Times(10)
        .WillOnce(Return(1.0/2)).WillOnce(Return(1.0/4)).WillOnce(Return(1.0/8))
        .WillOnce(Return(1.0/16)).WillOnce(Return(1.0/32)).WillOnce(Return(1.0/64))
        .WillOnce(Return(1.0/128)).WillOnce(Return(1.0/256)).WillOnce(Return(1.0/512))
        .WillOnce(Return(1.0/1024));
    EXPECT_CALL(*distributionEstimatorRef, getRuleProbability(static_cast<const AbstractValueRule*>(
        model->getRules()[0]->getUpdateRule()))).Times(10)
        .WillRepeatedly(Return(1.0));

    const std::unordered_map<bool, int64_t> expectedMatches{{true, 10}, {false, 0}};
    const double_t expectedModelStreamLength = 2*calc->codePrequential(expectedMatches);

    EXPECT_DOUBLE_EQ(1+2+3+4+5+6+7+8+9+10+expectedModelStreamLength,
        dataEnc.encodeEventLog(log.get(), model.get()));
}

TEST(DataEncoderTest, SometimesMatchingModelWithRepetitionTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<EventLog> log = Generators::generateEventLog(catVariable, numVariable);
    numVariable->setAccuracy(0.1);

    std::shared_ptr<MdlCalculator> calc = std::make_shared<MdlCalculator>();
    std::shared_ptr<DistributionEstimator> distributionEstimator =
        std::make_shared<MockDistributionEstimator>();
    std::unique_ptr<RuleFinder> ruleFinder = std::make_unique<MockRuleFinder>();
    std::unique_ptr<UpdateRuleMatcher> ruleMatcher = std::make_unique<MockUpdateRuleMatcher>();

    MockDistributionEstimator* distributionEstimatorRef =
        static_cast<MockDistributionEstimator*>(distributionEstimator.get());
    MockRuleFinder* ruleFinderRef =
        static_cast<MockRuleFinder*>(ruleFinder.get());
    MockUpdateRuleMatcher* ruleMatcherRef =
        static_cast<MockUpdateRuleMatcher*>(ruleMatcher.get());

    std::unique_ptr<RuleModel> model = Generators::generateRuleModel(catVariable, numVariable);
    DataEncoder dataEnc(calc, distributionEstimator, ruleFinder, ruleMatcher);

    EXPECT_CALL(*ruleFinderRef, findRulesThatApply(model.get(), catVariable.get(), _, _)).Times(10)
        .WillRepeatedly(Return(std::vector<const ModificationRule*>{model->getRules()[1]}));
    EXPECT_CALL(*ruleFinderRef, findRulesThatApply(model.get(), numVariable.get(), _, _)).Times(10)
        .WillRepeatedly(Return(std::vector<const ModificationRule*>{model->getRules()[2]}));
    EXPECT_CALL(*ruleMatcherRef, matches(static_cast<const ValueSetRule*>(
        model->getRules()[1]->getUpdateRule()), _, _)).Times(10)
        .WillOnce(Return(false)).WillOnce(Return(false)).WillOnce(Return(false))
        .WillOnce(Return(true)).WillOnce(Return(true)).WillOnce(Return(true)).WillOnce(Return(true))
        .WillOnce(Return(true)).WillOnce(Return(true)).WillOnce(Return(true));
    EXPECT_CALL(*ruleMatcherRef, matches(static_cast<const AbstractRangeRule*>(
        model->getRules()[2]->getUpdateRule()), _, _)).Times(10)
        .WillOnce(Return(false)).WillOnce(Return(false))
        .WillOnce(Return(true)).WillOnce(Return(true)).WillOnce(Return(true)).WillOnce(Return(true))
        .WillOnce(Return(true)).WillOnce(Return(true)).WillOnce(Return(true)).WillOnce(Return(true));
    EXPECT_CALL(*distributionEstimatorRef, getRuleProbability(static_cast<const ValueSetRule*>(
        model->getRules()[1]->getUpdateRule()), _)).Times(7)
        .WillOnce(Return(1.0/2)).WillOnce(Return(1.0/4)).WillOnce(Return(1.0/8))
        .WillOnce(Return(1.0/16)).WillOnce(Return(1.0/32)).WillOnce(Return(1.0/64))
        .WillOnce(Return(1.0/128));
    EXPECT_CALL(*distributionEstimatorRef, getRuleProbability(static_cast<const AbstractRangeRule*>(
        model->getRules()[2]->getUpdateRule()))).Times(8)
        .WillRepeatedly(Return(1.0/7.0));
    EXPECT_CALL(*distributionEstimatorRef, getGlobalProbability(catVariable, _)).Times(3)
        .WillOnce(Return(1.0/2)).WillOnce(Return(1.0/4)).WillOnce(Return(1.0/8));
    EXPECT_CALL(*distributionEstimatorRef, getGlobalProbability(numVariable, _)).Times(2)
        .WillOnce(Return(1.0/2)).WillOnce(Return(1.0/4));

    const std::unordered_map<bool, int64_t> expectedMatchesCat{{true, 7}, {false, 3}};
    const std::unordered_map<bool, int64_t> expectedMatchesNum{{true, 8}, {false, 2}};
    const double_t expectedModelStreamLength =
        calc->codePrequential(expectedMatchesCat) + calc->codePrequential(expectedMatchesNum);

    EXPECT_DOUBLE_EQ(1+2+3+4+5+6+7+1+2+3+8*std::log2(7)+1+2+expectedModelStreamLength,
        dataEnc.encodeEventLog(log.get(), model.get()));
}

TEST(DataEncoderTest, OverlappingRulesTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<EventLog> log = Generators::generateEventLog(catVariable, numVariable);
    numVariable->setAccuracy(0.1);

    std::shared_ptr<MdlCalculator> calc = std::make_shared<MdlCalculator>();
    std::shared_ptr<DistributionEstimator> distributionEstimator =
        std::make_shared<MockDistributionEstimator>();
    std::unique_ptr<RuleFinder> ruleFinder = std::make_unique<MockRuleFinder>();
    std::unique_ptr<UpdateRuleMatcher> ruleMatcher = std::make_unique<MockUpdateRuleMatcher>();

    MockDistributionEstimator* distributionEstimatorRef =
        static_cast<MockDistributionEstimator*>(distributionEstimator.get());
    MockRuleFinder* ruleFinderRef =
        static_cast<MockRuleFinder*>(ruleFinder.get());
    MockUpdateRuleMatcher* ruleMatcherRef =
        static_cast<MockUpdateRuleMatcher*>(ruleMatcher.get());

    std::unique_ptr<RuleModel> model = Generators::generateRuleModel(catVariable, numVariable);
    DataEncoder dataEnc(calc, distributionEstimator, ruleFinder, ruleMatcher);

    EXPECT_CALL(*ruleFinderRef, findRulesThatApply(model.get(), catVariable.get(), _, _)).Times(10)
        .WillRepeatedly(Return(std::vector<const ModificationRule*>{}));
    EXPECT_CALL(*ruleFinderRef, findRulesThatApply(model.get(), numVariable.get(), _, _)).Times(10)
        .WillRepeatedly(Return(
            std::vector<const ModificationRule*>{model->getRules()[0], model->getRules()[2]}));
    EXPECT_CALL(*ruleMatcherRef, matches(static_cast<const AbstractRangeRule*>(
        model->getRules()[2]->getUpdateRule()), _, _)).Times(10).WillRepeatedly(Return(true));
    EXPECT_CALL(*ruleMatcherRef, matches(static_cast<const AbstractValueRule*>(
        model->getRules()[0]->getUpdateRule()), _, _)).Times(10).WillRepeatedly(Return(true));
    EXPECT_CALL(*distributionEstimatorRef, getGlobalProbability(catVariable, _)).Times(10)
        .WillOnce(Return(1.0/2)).WillOnce(Return(1.0/4)).WillOnce(Return(1.0/8))
        .WillOnce(Return(1.0/16)).WillOnce(Return(1.0/32)).WillOnce(Return(1.0/64))
        .WillOnce(Return(1.0/128)).WillOnce(Return(1.0/256)).WillOnce(Return(1.0/512))
        .WillOnce(Return(1.0/1024));
    EXPECT_CALL(*distributionEstimatorRef, getRuleProbability(static_cast<const AbstractValueRule*>(
        model->getRules()[0]->getUpdateRule()))).Times(10)
        .WillRepeatedly(Return(1.0));
    EXPECT_CALL(*distributionEstimatorRef, getRuleProbability(static_cast<const AbstractRangeRule*>(
        model->getRules()[2]->getUpdateRule()))).Times(10)
        .WillRepeatedly(Return(1.0/7.0));

    const std::unordered_map<bool, int64_t> expectedMatches{{true, 10}, {false, 0}};
    const double_t expectedModelStreamLength = calc->codePrequential(expectedMatches);
    const double_t expectedRuleSelectionStreamLength = 10 * calc->codeUniformPossibilities(2);

    EXPECT_DOUBLE_EQ(1+2+3+4+5+6+7+8+9+10
        +expectedModelStreamLength+expectedRuleSelectionStreamLength,
        dataEnc.encodeEventLog(log.get(), model.get()));
}
