#include <gtest/gtest.h>
#include "../classes/scoring/dataencodingestimator.hpp"
#include "mock_distributionestimator.hpp"
#include "../classes/datamodel/updaterules/valuerule.hpp"
#include "../classes/datamodel/updaterules/valuerangerule.hpp"
#include "../classes/datamodel/updaterules/differencerule.hpp"
#include "../classes/datamodel/updaterules/factorrule.hpp"
#include "../classes/datamodel/updaterules/changerangerule.hpp"

using ::testing::Return;
using ::testing::ReturnRef;
using ::testing::_;

TEST(DataEncodingEstimatorTest, ValueSetRuleTest)
{
    auto catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    auto valueSetRule = std::make_unique<ValueSetRule>(catVariable, ValueSetRule::SetType{"e", "i"});
    DistributionEstimator::CatDistributionType distribution {
        {"a", 1}, {"e", 3}, {"i", 4}, {"o", 5}, {"u", 5}};
    
    std::shared_ptr<MdlCalculator> calc = std::make_shared<MdlCalculator>();
    std::shared_ptr<DistributionEstimator> distributionEstimator =
        std::make_shared<MockDistributionEstimator>();

    MockDistributionEstimator* distributionEstimatorRef =
        static_cast<MockDistributionEstimator*>(distributionEstimator.get());

    EXPECT_CALL(*distributionEstimatorRef, getCategoricalDistribution(catVariable)).Times(1)
        .WillOnce(ReturnRef(distribution));
    EXPECT_CALL(*distributionEstimatorRef, getGlobalProbability(catVariable, "e")).Times(1)
        .WillOnce(Return(1.0/32.0));
    EXPECT_CALL(*distributionEstimatorRef, getGlobalProbability(catVariable, "i")).Times(1)
        .WillOnce(Return(1.0/8.0));
    EXPECT_CALL(*distributionEstimatorRef, getRuleProbability(valueSetRule.get(), "e")).Times(1)
        .WillOnce(Return(1.0/4.0));
    EXPECT_CALL(*distributionEstimatorRef, getRuleProbability(valueSetRule.get(), "i")).Times(1)
        .WillOnce(Return(1.0/2.0));

    DataEncodingEstimator encodingEstimator(calc, distributionEstimator);

    EXPECT_DOUBLE_EQ(3*(5-2) + 2*(3-1), encodingEstimator.estimateSavedBits(valueSetRule.get(), 5));
}

TEST(DataEncodingEstimatorTest, TooManySamplesTest)
{
    auto catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    auto valueSetRule = std::make_unique<ValueSetRule>(catVariable, ValueSetRule::SetType{"e", "i"});
    DistributionEstimator::CatDistributionType distribution {
        {"a", 1}, {"e", 3}, {"i", 4}, {"o", 5}, {"u", 5}};
    
    std::shared_ptr<MdlCalculator> calc = std::make_shared<MdlCalculator>();
    std::shared_ptr<DistributionEstimator> distributionEstimator =
        std::make_shared<MockDistributionEstimator>();

    MockDistributionEstimator* distributionEstimatorRef =
        static_cast<MockDistributionEstimator*>(distributionEstimator.get());

    EXPECT_CALL(*distributionEstimatorRef, getCategoricalDistribution(catVariable)).Times(1)
        .WillOnce(ReturnRef(distribution));
    EXPECT_CALL(*distributionEstimatorRef, getGlobalProbability(catVariable, "e")).Times(1)
        .WillOnce(Return(1.0/32.0));
    EXPECT_CALL(*distributionEstimatorRef, getGlobalProbability(catVariable, "i")).Times(1)
        .WillOnce(Return(1.0/8.0));
    EXPECT_CALL(*distributionEstimatorRef, getRuleProbability(valueSetRule.get(), "e")).Times(1)
        .WillOnce(Return(1.0/4.0));
    EXPECT_CALL(*distributionEstimatorRef, getRuleProbability(valueSetRule.get(), "i")).Times(1)
        .WillOnce(Return(1.0/2.0));

    DataEncodingEstimator encodingEstimator(calc, distributionEstimator);

    EXPECT_DOUBLE_EQ(3*(5-2) + 4*(3-1), encodingEstimator.estimateSavedBits(valueSetRule.get(), 8));
}

TEST(DataEncodingEstimatorTest, ValueRuleTest)
{
    auto numVariable = std::make_shared<NumericalVariable>("num");
    numVariable->setAccuracy(1.0);
    auto valueRule = std::make_unique<ValueRule>(numVariable, 3.0);
    Histogram histogram (std::vector<double_t>({0.5, 1.5, 2.5, 3.5}),
        std::vector<int64_t>({7, 11, 13}));
    
    std::shared_ptr<MdlCalculator> calc = std::make_shared<MdlCalculator>();
    std::shared_ptr<DistributionEstimator> distributionEstimator =
        std::make_shared<MockDistributionEstimator>();

    MockDistributionEstimator* distributionEstimatorRef =
        static_cast<MockDistributionEstimator*>(distributionEstimator.get());

    EXPECT_CALL(*distributionEstimatorRef, getNumericalDistribution(numVariable))
        .WillRepeatedly(ReturnRef(histogram));
    EXPECT_CALL(*distributionEstimatorRef, getGlobalProbability(numVariable, 3.0)).Times(1)
        .WillOnce(Return(1.0/64.0));
    EXPECT_CALL(*distributionEstimatorRef, getRuleProbability(valueRule.get())).Times(1)
        .WillOnce(Return(1.0/4.0));

    DataEncodingEstimator encodingEstimator(calc, distributionEstimator);

    EXPECT_DOUBLE_EQ(12*(6-2), encodingEstimator.estimateSavedBits(valueRule.get(), 12));
}

TEST(DataEncodingEstimatorTest, ValueRangeRuleTest)
{
    auto numVariable = std::make_shared<NumericalVariable>("num");
    numVariable->setAccuracy(1.0);
    auto valueRangeRule = std::make_unique<ValueRangeRule>(numVariable, 1.0, 3.0);
    Histogram histogram (std::vector<double_t>({-1.5, 0.5, 1.5, 2.5, 3.5, 5.5}),
        std::vector<int64_t>({1, 4, 5, 6, 2}));
    
    std::shared_ptr<MdlCalculator> calc = std::make_shared<MdlCalculator>();
    std::shared_ptr<DistributionEstimator> distributionEstimator =
        std::make_shared<MockDistributionEstimator>();

    MockDistributionEstimator* distributionEstimatorRef =
        static_cast<MockDistributionEstimator*>(distributionEstimator.get());

    EXPECT_CALL(*distributionEstimatorRef, getNumericalDistribution(numVariable))
        .WillRepeatedly(ReturnRef(histogram));
    EXPECT_CALL(*distributionEstimatorRef, getGlobalProbability(numVariable, 1.0)).Times(1)
        .WillOnce(Return(1.0/128.0));
    EXPECT_CALL(*distributionEstimatorRef, getGlobalProbability(numVariable, 2.0)).Times(1)
        .WillOnce(Return(1.0/64.0));
    EXPECT_CALL(*distributionEstimatorRef, getRuleProbability(valueRangeRule.get())).Times(2)
        .WillRepeatedly(Return(1.0/2.0));

    DataEncodingEstimator encodingEstimator(calc, distributionEstimator);

    EXPECT_DOUBLE_EQ(4*(7-1) + 5*(6-1), encodingEstimator.estimateSavedBits(valueRangeRule.get(), 9));
}

TEST(DataEncodingEstimatorTest, DifferenceRuleTest)
{
    auto numVariable = std::make_shared<NumericalVariable>("num");
    numVariable->setAccuracy(1.0);
    auto differenceRule = std::make_unique<DifferenceRule>(numVariable, -2.0);
    Histogram histogram (std::vector<double_t>({0.5, 1.5, 2.5, 4.5, 5.5}),
        std::vector<int64_t>({2, 8, 3, 1}));
    
    std::shared_ptr<MdlCalculator> calc = std::make_shared<MdlCalculator>();
    std::shared_ptr<DistributionEstimator> distributionEstimator =
        std::make_shared<MockDistributionEstimator>();

    MockDistributionEstimator* distributionEstimatorRef =
        static_cast<MockDistributionEstimator*>(distributionEstimator.get());

    EXPECT_CALL(*distributionEstimatorRef, getNumericalDistribution(numVariable))
        .WillRepeatedly(ReturnRef(histogram));
    EXPECT_CALL(*distributionEstimatorRef, getGlobalProbability(numVariable, 1.0)).Times(1)
        .WillOnce(Return(1.0/128.0));
    EXPECT_CALL(*distributionEstimatorRef, getGlobalProbability(numVariable, 3.0)).Times(1)
        .WillOnce(Return(1.0/64.0));
    EXPECT_CALL(*distributionEstimatorRef, getRuleProbability(differenceRule.get())).Times(2)
        .WillRepeatedly(Return(1.0));

    DataEncodingEstimator encodingEstimator(calc, distributionEstimator);

    EXPECT_DOUBLE_EQ(2*7 + 3*6, encodingEstimator.estimateSavedBits(differenceRule.get(), 5));
}

TEST(DataEncodingEstimatorTest, FactorRuleShiftTest)
{
    auto numVariable = std::make_shared<NumericalVariable>("num");
    numVariable->setAccuracy(1.0);
    auto factorRule = std::make_unique<FactorRule>(numVariable, 4.0);
    Histogram histogram (std::vector<double_t>({0.5, 1.5, 2.5, 3.5, 4.5, 5.5}),
        std::vector<int64_t>({1, 2, 3, 4, 5}));
    
    std::shared_ptr<MdlCalculator> calc = std::make_shared<MdlCalculator>();
    std::shared_ptr<DistributionEstimator> distributionEstimator =
        std::make_shared<MockDistributionEstimator>();

    MockDistributionEstimator* distributionEstimatorRef =
        static_cast<MockDistributionEstimator*>(distributionEstimator.get());

    EXPECT_CALL(*distributionEstimatorRef, getNumericalDistribution(numVariable))
        .WillRepeatedly(ReturnRef(histogram));
    EXPECT_CALL(*distributionEstimatorRef, getGlobalProbability(numVariable, 2.0)).Times(1)
        .WillOnce(Return(1.0/128.0));
    EXPECT_CALL(*distributionEstimatorRef, getGlobalProbability(numVariable, 3.0)).Times(1)
        .WillOnce(Return(1.0/64.0));
    EXPECT_CALL(*distributionEstimatorRef, getRuleProbability(factorRule.get())).Times(2)
        .WillRepeatedly(Return(1.0));

    DataEncodingEstimator encodingEstimator(calc, distributionEstimator);

    EXPECT_DOUBLE_EQ(2*7 + 1*6, encodingEstimator.estimateSavedBits(factorRule.get(), 3));
}

TEST(DataEncodingEstimatorTest, FactorRuleScaleTest)
{
    auto numVariable = std::make_shared<NumericalVariable>("num");
    numVariable->setAccuracy(1.0);
    auto factorRule = std::make_unique<FactorRule>(numVariable, 0.25);
    Histogram histogram (std::vector<double_t>({-2.5, -1.5, 1.5, 2.5, 3.5}),
        std::vector<int64_t>({9, 2, 3, 4}));
    
    std::shared_ptr<MdlCalculator> calc = std::make_shared<MdlCalculator>();
    std::shared_ptr<DistributionEstimator> distributionEstimator =
        std::make_shared<MockDistributionEstimator>();

    MockDistributionEstimator* distributionEstimatorRef =
        static_cast<MockDistributionEstimator*>(distributionEstimator.get());

    EXPECT_CALL(*distributionEstimatorRef, getNumericalDistribution(numVariable))
        .WillRepeatedly(ReturnRef(histogram));
    EXPECT_CALL(*distributionEstimatorRef, getGlobalProbability(numVariable, -1.0)).Times(1)
        .WillOnce(Return(1.0/128.0));
    EXPECT_CALL(*distributionEstimatorRef, getRuleProbability(factorRule.get())).Times(1)
        .WillRepeatedly(Return(1.0));

    DataEncodingEstimator encodingEstimator(calc, distributionEstimator);

    EXPECT_DOUBLE_EQ(2*7, encodingEstimator.estimateSavedBits(factorRule.get(), 2));
}

TEST(DataEncodingEstimatorTest, FactorRuleScaleFlipTest)
{
    auto numVariable = std::make_shared<NumericalVariable>("num");
    numVariable->setAccuracy(1.0);
    auto factorRule = std::make_unique<FactorRule>(numVariable, -0.25);
    Histogram histogram (std::vector<double_t>({-2.5, -1.5, 1.5, 2.5, 3.5}),
        std::vector<int64_t>({9, 2, 3, 4}));
    
    std::shared_ptr<MdlCalculator> calc = std::make_shared<MdlCalculator>();
    std::shared_ptr<DistributionEstimator> distributionEstimator =
        std::make_shared<MockDistributionEstimator>();

    MockDistributionEstimator* distributionEstimatorRef =
        static_cast<MockDistributionEstimator*>(distributionEstimator.get());

    EXPECT_CALL(*distributionEstimatorRef, getNumericalDistribution(numVariable))
        .WillRepeatedly(ReturnRef(histogram));
    EXPECT_CALL(*distributionEstimatorRef, getGlobalProbability(numVariable, -1.0)).Times(1)
        .WillOnce(Return(1.0/128.0));
    EXPECT_CALL(*distributionEstimatorRef, getRuleProbability(factorRule.get())).Times(1)
        .WillRepeatedly(Return(1.0));

    DataEncodingEstimator encodingEstimator(calc, distributionEstimator);

    EXPECT_DOUBLE_EQ(2*7, encodingEstimator.estimateSavedBits(factorRule.get(), 2));
}

TEST(DataEncodingEstimatorTest, ChangeRangeRuleTest)
{
    auto numVariable = std::make_shared<NumericalVariable>("num");
    numVariable->setAccuracy(1.0);
    auto changeRangeRule = std::make_unique<ChangeRangeRule>(numVariable, -4.0, -2.0);
    Histogram histogram (std::vector<double_t>({0.5, 1.5, 2.5, 4.5, 5.5}),
        std::vector<int64_t>({5, 8, 4, 1}));
    
    std::shared_ptr<MdlCalculator> calc = std::make_shared<MdlCalculator>();
    std::shared_ptr<DistributionEstimator> distributionEstimator =
        std::make_shared<MockDistributionEstimator>();

    MockDistributionEstimator* distributionEstimatorRef =
        static_cast<MockDistributionEstimator*>(distributionEstimator.get());

    EXPECT_CALL(*distributionEstimatorRef, getNumericalDistribution(numVariable))
        .WillRepeatedly(ReturnRef(histogram));
    EXPECT_CALL(*distributionEstimatorRef, getGlobalProbability(numVariable, 1.0)).Times(1)
        .WillOnce(Return(1.0/128.0));
    EXPECT_CALL(*distributionEstimatorRef, getGlobalProbability(numVariable, 3.0)).Times(1)
        .WillOnce(Return(1.0/64.0));
    EXPECT_CALL(*distributionEstimatorRef, getRuleProbability(changeRangeRule.get())).Times(2)
        .WillRepeatedly(Return(0.5));

    DataEncodingEstimator encodingEstimator(calc, distributionEstimator);

    EXPECT_DOUBLE_EQ(1*(7-1) + 4*(6-1), encodingEstimator.estimateSavedBits(changeRangeRule.get(), 5));
}

TEST(DataEncodingEstimatorTest, ChangeRangeRuleFullRangeTest)
{
    auto numVariable = std::make_shared<NumericalVariable>("num");
    numVariable->setAccuracy(1.0);
    auto changeRangeRule = std::make_unique<ChangeRangeRule>(numVariable, -5.0, 5.0);
    Histogram histogram (std::vector<double_t>({0.5, 1.5, 2.5, 4.5, 5.5}),
        std::vector<int64_t>({5, 8, 4, 1}));
    
    std::shared_ptr<MdlCalculator> calc = std::make_shared<MdlCalculator>();
    std::shared_ptr<DistributionEstimator> distributionEstimator =
        std::make_shared<MockDistributionEstimator>();

    MockDistributionEstimator* distributionEstimatorRef =
        static_cast<MockDistributionEstimator*>(distributionEstimator.get());

    EXPECT_CALL(*distributionEstimatorRef, getNumericalDistribution(numVariable))
        .WillRepeatedly(ReturnRef(histogram));
    EXPECT_CALL(*distributionEstimatorRef, getGlobalProbability(numVariable, 1.0)).Times(1)
        .WillOnce(Return(1.0/256.0));
    EXPECT_CALL(*distributionEstimatorRef, getGlobalProbability(numVariable, 2.0)).Times(1)
        .WillOnce(Return(1.0/128.0));
    EXPECT_CALL(*distributionEstimatorRef, getGlobalProbability(numVariable, 3.0)).Times(1)
        .WillOnce(Return(1.0/512.0));
    EXPECT_CALL(*distributionEstimatorRef, getGlobalProbability(numVariable, 5.0)).Times(1)
        .WillOnce(Return(1.0/1024.0));
    EXPECT_CALL(*distributionEstimatorRef, getRuleProbability(changeRangeRule.get())).Times(4)
        .WillRepeatedly(Return(1.0/8.0));

    DataEncodingEstimator encodingEstimator(calc, distributionEstimator);

    EXPECT_DOUBLE_EQ(1*(10-3) + 4*(9-3) + 5*(8-3) + 8*(7-3),
        encodingEstimator.estimateSavedBits(changeRangeRule.get(), 18));
}

TEST(DataEncodingEstimatorTest, SameBinButOutOfRangeTest)
{
    auto numVariable = std::make_shared<NumericalVariable>("num");
    numVariable->setAccuracy(1.0);
    auto valueRangeRule = std::make_unique<ValueRangeRule>(numVariable, 1.0, 3.0);
    Histogram histogram (std::vector<double_t>({-1.5, 5.5}),
        std::vector<int64_t>({12}));
    
    std::shared_ptr<MdlCalculator> calc = std::make_shared<MdlCalculator>();
    std::shared_ptr<DistributionEstimator> distributionEstimator =
        std::make_shared<MockDistributionEstimator>();

    MockDistributionEstimator* distributionEstimatorRef =
        static_cast<MockDistributionEstimator*>(distributionEstimator.get());

    EXPECT_CALL(*distributionEstimatorRef, getNumericalDistribution(numVariable))
        .WillRepeatedly(ReturnRef(histogram));
    EXPECT_CALL(*distributionEstimatorRef, getGlobalProbability(numVariable, -1.0)).Times(1)
        .WillOnce(Return(1.0/128.0));
    EXPECT_CALL(*distributionEstimatorRef, getRuleProbability(valueRangeRule.get())).Times(1)
        .WillRepeatedly(Return(1.0/2.0));

    DataEncodingEstimator encodingEstimator(calc, distributionEstimator);

    EXPECT_DOUBLE_EQ(9*(7-1), encodingEstimator.estimateSavedBits(valueRangeRule.get(), 9));
}

TEST(DataEncodingEstimatorTest, FrequencyVsProbabilityTest)
{
    auto numVariable = std::make_shared<NumericalVariable>("num");
    numVariable->setAccuracy(1.0);
    auto valueRangeRule = std::make_unique<ValueRangeRule>(numVariable, 1.0, 3.0);
    Histogram histogram (std::vector<double_t>({-1.5, 1.5, 2.5, 3.5, 5.5}),
        std::vector<int64_t>({4, 3, 2, 1}));
    
    std::shared_ptr<MdlCalculator> calc = std::make_shared<MdlCalculator>();
    std::shared_ptr<DistributionEstimator> distributionEstimator =
        std::make_shared<MockDistributionEstimator>();

    MockDistributionEstimator* distributionEstimatorRef =
        static_cast<MockDistributionEstimator*>(distributionEstimator.get());

    EXPECT_CALL(*distributionEstimatorRef, getNumericalDistribution(numVariable))
        .WillRepeatedly(ReturnRef(histogram));
    EXPECT_CALL(*distributionEstimatorRef, getGlobalProbability(numVariable, -1.0)).Times(1)
        .WillOnce(Return(1.0/128.0));
    EXPECT_CALL(*distributionEstimatorRef, getRuleProbability(valueRangeRule.get())).Times(1)
        .WillRepeatedly(Return(1.0/2.0));

    DataEncodingEstimator encodingEstimator(calc, distributionEstimator);

    EXPECT_DOUBLE_EQ(4*(7-1), encodingEstimator.estimateSavedBits(valueRangeRule.get(), 4));
}
