#include <gtest/gtest.h>
#include <memory>
#include <random>
#include "generators.hpp"
#include "../classes/datagenerator.hpp"
#include "../classes/scoring/conditionmatcher.hpp"
#include "../classes/datamodel/conditions/lowerboundcondition.hpp"
#include "../classes/datamodel/conditions/upperboundcondition.hpp"
#include "../classes/datamodel/conditions/valuecondition.hpp"
#include "../classes/datamodel/conditions/transitioncondition.hpp"
#include "../classes/datamodel/conditions/andoperator.hpp"
#include "../classes/datamodel/updaterules/valuerule.hpp"
#include "../classes/datamodel/updaterules/factorrule.hpp"
#include "../classes/datamodel/updaterules/differencerule.hpp"
#include "../classes/datamodel/updaterules/valuerangerule.hpp"
#include "../classes/datamodel/updaterules/changerangerule.hpp"
#include "../classes/datamodel/eventlogs/logcolumn.hpp"
#include <iostream>

TEST(DataGeneratorTest, EmptyModelTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::vector<std::shared_ptr<LogVariable>> variables {catVariable, numVariable};
    auto model = std::make_unique<RuleModel>();

    DataGenerator generator;
    auto logStatisticsPair = generator.generateEventLog(variables, model.get(), 3, 7);
    auto log = std::move(logStatisticsPair.first);
    auto statistics = logStatisticsPair.second;

    EXPECT_EQ(0, statistics.size());
    EXPECT_EQ(3, log->getTraces().size());
    EXPECT_EQ(7, log->getTraces().begin()->second->size());
}

TEST(DataGeneratorTest, AlwaysMatchingRuleTest)
{
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::vector<std::shared_ptr<LogVariable>> variables {numVariable};

    RuleModel::RulesType rules;
    auto condition = std::make_unique<Condition>();
    std::unique_ptr<UpdateRule> updateRule = std::make_unique<ValueRule>(numVariable, 42.0);
    rules.push_back(std::make_unique<ModificationRule>(condition, updateRule));
    auto model = std::make_unique<RuleModel>(rules);

    DataGenerator generator;
    auto logStatisticsPair = generator.generateEventLog(variables, model.get(), 3, 7);
    auto log = std::move(logStatisticsPair.first);
    auto statistics = logStatisticsPair.second;

    EXPECT_EQ(1, statistics.size());
    EXPECT_EQ(3*7, statistics.at(model->getRules()[0]));
    EXPECT_EQ(3, log->getTraces().size());
    EXPECT_EQ(7, log->getTraces().begin()->second->size());

    for (const auto& tracePair : log->getTraces())
    {
        auto numColumn = static_cast<LogColumn<double_t>*>(
            tracePair.second->getColumns().at(numVariable).get());
        for (const auto& entry : numColumn->getEntries())
        {
            EXPECT_EQ(42.0, entry);
        }
    }
}

TEST(DataGeneratorTest, ValueSetRuleTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::vector<std::shared_ptr<LogVariable>> variables {catVariable, numVariable};

    RuleModel::RulesType rules;
    Condition::ElementsType elements;
    elements.push_back(std::make_unique<LowerBoundCondition>(numVariable, 50.0));
    auto condition = std::make_unique<Condition>(elements);
    std::unique_ptr<UpdateRule> updateRule = std::make_unique<ValueSetRule>(
        catVariable, std::set<std::string>{"e", "i"});
    rules.push_back(std::make_unique<ModificationRule>(condition, updateRule));
    auto model = std::make_unique<RuleModel>(rules);

    DataGenerator generator;
    auto logStatisticsPair = generator.generateEventLog(variables, model.get(), 1, 100);
    auto log = std::move(logStatisticsPair.first);
    auto statistics = logStatisticsPair.second;

    auto trace = log->getTraces().begin()->second.get();
    auto targetColumn = static_cast<const LogColumn<std::string>*>(
        trace->getColumns().at(catVariable).get());
    EXPECT_EQ(1, statistics.size());
    ASSERT_EQ(100, trace->size());

    ConditionMatcher matcher;
    for (size_t i = 0; i < trace->size(); i++)
    {
        if (matcher.matches(model->getRules()[0]->getCondition(), trace, i))
        {
            std::set<std::string> allowedValues {"e", "i"};
            EXPECT_TRUE(allowedValues.find(targetColumn->getEntries()[i]) != allowedValues.end());
        }
        else
        {
            std::set<std::string> allowedValues {"a", "e", "i", "o", "u"};
            EXPECT_TRUE(allowedValues.find(targetColumn->getEntries()[i]) != allowedValues.end());
        }
    }
}

TEST(DataGeneratorTest, ValueRuleTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("cat",
        std::set<std::string>{"x", "o"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::vector<std::shared_ptr<LogVariable>> variables {catVariable, numVariable};

    RuleModel::RulesType rules;
    Condition::ElementsType elements;
    elements.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "x"));
    auto condition = std::make_unique<Condition>(elements);
    std::unique_ptr<UpdateRule> updateRule = std::make_unique<ValueRule>(numVariable, 42.12345);
    rules.push_back(std::make_unique<ModificationRule>(condition, updateRule));
    auto model = std::make_unique<RuleModel>(rules);

    DataGenerator generator;
    auto logStatisticsPair = generator.generateEventLog(variables, model.get(), 1, 100);
    auto log = std::move(logStatisticsPair.first);
    auto statistics = logStatisticsPair.second;

    auto trace = log->getTraces().begin()->second.get();
    auto targetColumn = static_cast<const LogColumn<double_t>*>(
        trace->getColumns().at(numVariable).get());
    EXPECT_EQ(1, statistics.size());
    ASSERT_EQ(100, trace->size());

    ConditionMatcher matcher;
    for (size_t i = 0; i < trace->size(); i++)
    {
        if (matcher.matches(model->getRules()[0]->getCondition(), trace, i))
        {
            EXPECT_DOUBLE_EQ(42.1, targetColumn->getEntries()[i]);
        }
        else
        {
            EXPECT_LE(0.0, targetColumn->getEntries()[i]);
            EXPECT_LE(targetColumn->getEntries()[i], 100.0);
        }
    }
}

TEST(DataGeneratorTest, FactorRuleTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("cat",
        std::set<std::string>{"x", "o"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::vector<std::shared_ptr<LogVariable>> variables {catVariable, numVariable};

    RuleModel::RulesType rules;
    Condition::ElementsType elements;
    elements.push_back(std::make_unique<TransitionCondition<std::string>>(catVariable, "x", "o"));
    auto condition = std::make_unique<Condition>(elements);
    std::unique_ptr<UpdateRule> updateRule = std::make_unique<FactorRule>(numVariable, 2.0);
    rules.push_back(std::make_unique<ModificationRule>(condition, updateRule));
    auto model = std::make_unique<RuleModel>(rules);

    DataGenerator generator;
    auto logStatisticsPair = generator.generateEventLog(variables, model.get(), 1, 100);
    auto log = std::move(logStatisticsPair.first);
    auto statistics = logStatisticsPair.second;

    auto trace = log->getTraces().begin()->second.get();
    auto targetColumn = static_cast<const LogColumn<double_t>*>(
        trace->getColumns().at(numVariable).get());
    EXPECT_EQ(1, statistics.size());
    ASSERT_EQ(100, trace->size());

    ConditionMatcher matcher;
    for (size_t i = 0; i < trace->size(); i++)
    {
        if (matcher.matches(model->getRules()[0]->getCondition(), trace, i))
        {
            EXPECT_NEAR(2.0 * targetColumn->getEntries()[i-1], targetColumn->getEntries()[i], 0.1);
        }
        else
        {
            EXPECT_LE(0.0, targetColumn->getEntries()[i]);
            EXPECT_LE(targetColumn->getEntries()[i], 100.0);
        }
    }
}

TEST(DataGeneratorTest, DifferenceRuleTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("cat",
        std::set<std::string>{"x", "o"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::vector<std::shared_ptr<LogVariable>> variables {catVariable, numVariable};

    RuleModel::RulesType rules;
    Condition::ElementsType elements;
    elements.push_back(std::make_unique<TransitionCondition<std::string>>(catVariable, "x", "o"));
    auto condition = std::make_unique<Condition>(elements);
    std::unique_ptr<UpdateRule> updateRule = std::make_unique<DifferenceRule>(numVariable, -11.0);
    rules.push_back(std::make_unique<ModificationRule>(condition, updateRule));
    auto model = std::make_unique<RuleModel>(rules);

    DataGenerator generator;
    auto logStatisticsPair = generator.generateEventLog(variables, model.get(), 1, 100);
    auto log = std::move(logStatisticsPair.first);
    auto statistics = logStatisticsPair.second;

    auto trace = log->getTraces().begin()->second.get();
    auto targetColumn = static_cast<const LogColumn<double_t>*>(
        trace->getColumns().at(numVariable).get());
    EXPECT_EQ(1, statistics.size());
    ASSERT_EQ(100, trace->size());

    ConditionMatcher matcher;
    for (size_t i = 0; i < trace->size(); i++)
    {
        if (matcher.matches(model->getRules()[0]->getCondition(), trace, i))
        {
            EXPECT_NEAR(targetColumn->getEntries()[i-1] - 11.0, targetColumn->getEntries()[i], 0.1);
        }
        else
        {
            EXPECT_LE(0.0, targetColumn->getEntries()[i]);
            EXPECT_LE(targetColumn->getEntries()[i], 100.0);
        }
    }
}

TEST(DataGeneratorTest, ValueRangeRuleTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("cat",
        std::set<std::string>{"x", "o"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::vector<std::shared_ptr<LogVariable>> variables {catVariable, numVariable};

    RuleModel::RulesType rules;
    Condition::ElementsType elements;
    elements.push_back(std::make_unique<TransitionCondition<std::string>>(catVariable, "x", "o"));
    auto condition = std::make_unique<Condition>(elements);
    std::unique_ptr<UpdateRule> updateRule = std::make_unique<ValueRangeRule>(
        numVariable, 42.0, 50.0);
    rules.push_back(std::make_unique<ModificationRule>(condition, updateRule));
    auto model = std::make_unique<RuleModel>(rules);

    DataGenerator generator;
    auto logStatisticsPair = generator.generateEventLog(variables, model.get(), 1, 100);
    auto log = std::move(logStatisticsPair.first);
    auto statistics = logStatisticsPair.second;

    auto trace = log->getTraces().begin()->second.get();
    auto targetColumn = static_cast<const LogColumn<double_t>*>(
        trace->getColumns().at(numVariable).get());
    EXPECT_EQ(1, statistics.size());
    ASSERT_EQ(100, trace->size());

    ConditionMatcher matcher;
    for (size_t i = 0; i < trace->size(); i++)
    {
        if (matcher.matches(model->getRules()[0]->getCondition(), trace, i))
        {
            EXPECT_LE(42.0, targetColumn->getEntries()[i]);
            EXPECT_LE(targetColumn->getEntries()[i], 50.0);
        }
        else
        {
            EXPECT_LE(0.0, targetColumn->getEntries()[i]);
            EXPECT_LE(targetColumn->getEntries()[i], 100.0);
        }
    }
}

TEST(DataGeneratorTest, ChangeRangeRuleTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("cat",
        std::set<std::string>{"x", "o"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::vector<std::shared_ptr<LogVariable>> variables {catVariable, numVariable};

    RuleModel::RulesType rules;
    Condition::ElementsType elements;
    elements.push_back(std::make_unique<TransitionCondition<std::string>>(catVariable, "x", "o"));
    auto condition = std::make_unique<Condition>(elements);
    std::unique_ptr<UpdateRule> updateRule = std::make_unique<ChangeRangeRule>(
        numVariable, 10.0, 15.0);
    rules.push_back(std::make_unique<ModificationRule>(condition, updateRule));
    auto model = std::make_unique<RuleModel>(rules);

    DataGenerator generator;
    auto logStatisticsPair = generator.generateEventLog(variables, model.get(), 1, 100);
    auto log = std::move(logStatisticsPair.first);
    auto statistics = logStatisticsPair.second;

    auto trace = log->getTraces().begin()->second.get();
    auto targetColumn = static_cast<const LogColumn<double_t>*>(
        trace->getColumns().at(numVariable).get());
    EXPECT_EQ(1, statistics.size());
    ASSERT_EQ(100, trace->size());

    ConditionMatcher matcher;
    for (size_t i = 0; i < trace->size(); i++)
    {
        if (matcher.matches(model->getRules()[0]->getCondition(), trace, i))
        {
            const auto change = targetColumn->getEntries()[i] - targetColumn->getEntries()[i-1];
            EXPECT_LE(10.0, change);
            EXPECT_LE(change, 15.0);
        }
        else
        {
            EXPECT_LE(0.0, targetColumn->getEntries()[i]);
            EXPECT_LE(targetColumn->getEntries()[i], 100.0);
        }
    }
}

TEST(DataGeneratorTest, OverlappingValueRuleTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("cat",
        std::set<std::string>{"x", "o"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::vector<std::shared_ptr<LogVariable>> variables {catVariable, numVariable};

    RuleModel::RulesType rules;
    Condition::ElementsType elements;
    elements.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "x"));
    auto condition = std::make_unique<Condition>(elements);
    std::unique_ptr<UpdateRule> updateRule = std::make_unique<ValueRule>(numVariable, 42);
    rules.push_back(std::make_unique<ModificationRule>(condition, updateRule));
    Condition::ElementsType elements2;
    elements2.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "x"));
    auto condition2 = std::make_unique<Condition>(elements2);
    std::unique_ptr<UpdateRule> updateRule2 = std::make_unique<ValueRule>(numVariable, 7.4332);
    rules.push_back(std::make_unique<ModificationRule>(condition2, updateRule2));
    auto model = std::make_unique<RuleModel>(rules);

    DataGenerator generator;
    auto logStatisticsPair = generator.generateEventLog(variables, model.get(), 1, 100);
    auto log = std::move(logStatisticsPair.first);
    auto statistics = logStatisticsPair.second;

    auto trace = log->getTraces().begin()->second.get();
    auto targetColumn = static_cast<const LogColumn<double_t>*>(
        trace->getColumns().at(numVariable).get());
    EXPECT_EQ(2, statistics.size());
    ASSERT_EQ(100, trace->size());

    ConditionMatcher matcher;
    for (size_t i = 0; i < trace->size(); i++)
    {
        if (matcher.matches(model->getRules()[0]->getCondition(), trace, i))
        {
            EXPECT_TRUE(
                42.0 == targetColumn->getEntries()[i] || 7.4 == targetColumn->getEntries()[i]);
        }
        else
        {
            EXPECT_LE(0.0, targetColumn->getEntries()[i]);
            EXPECT_LE(targetColumn->getEntries()[i], 100.0);
        }
    }
}

TEST(DataGeneratorTest, OverlappingSpecializationRuleTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>(
        "light1", std::set<std::string>{"red", "yellow", "green"});
    std::shared_ptr<CategoricalVariable> catVariable2 = std::make_shared<CategoricalVariable>(
        "light2", std::set<std::string>{"red", "yellow", "green"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::vector<std::shared_ptr<LogVariable>> variables {catVariable, catVariable2, numVariable};

    RuleModel::RulesType rules;
    Condition::ElementsType elements;
    elements.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "red"));
    auto condition = std::make_unique<Condition>(elements);
    std::unique_ptr<UpdateRule> updateRule = std::make_unique<DifferenceRule>(numVariable, -4);
    rules.push_back(std::make_unique<ModificationRule>(condition, updateRule));
    Condition::ElementsType elements2;
    elements2.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "red"));
    elements2.push_back(std::make_unique<ValueCondition<std::string>>(catVariable2, "red"));
    elements2.push_back(std::make_unique<AndOperator>());
    auto condition2 = std::make_unique<Condition>(elements2);
    std::unique_ptr<UpdateRule> updateRule2 = std::make_unique<DifferenceRule>(numVariable, -10);
    rules.push_back(std::make_unique<ModificationRule>(condition2, updateRule2));
    auto model = std::make_unique<RuleModel>(rules);

    DataGenerator generator;
    auto logStatisticsPair = generator.generateEventLog(variables, model.get(), 1, 100);
    auto log = std::move(logStatisticsPair.first);
    auto statistics = logStatisticsPair.second;

    auto trace = log->getTraces().begin()->second.get();
    auto targetColumn = static_cast<const LogColumn<double_t>*>(
        trace->getColumns().at(numVariable).get());
    EXPECT_EQ(2, statistics.size());
    ASSERT_EQ(100, trace->size());

    ConditionMatcher matcher;
    for (size_t i = 1; i < trace->size(); i++)
    {
        constexpr double_t tolerance = 1e-10;
        const auto change = targetColumn->getEntries()[i] - targetColumn->getEntries()[i-1];
        if (matcher.matches(model->getRules()[0]->getCondition(), trace, i)
            && matcher.matches(model->getRules()[1]->getCondition(), trace, i))
        {
            EXPECT_TRUE(std::abs(-4 - change) < tolerance || std::abs(-10 - change) < tolerance);
        }
        else if (matcher.matches(model->getRules()[0]->getCondition(), trace, i))
        {
            EXPECT_NEAR(-4, change, tolerance);
        }
        else
        {
            EXPECT_LE(0.0, targetColumn->getEntries()[i]);
            EXPECT_LE(targetColumn->getEntries()[i], 100.0);
        }
    }
}

TEST(DataGeneratorTest, PredictCategoricalTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("cat",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    auto sourceLog = Generators::generateEventLog(catVariable, numVariable);

    RuleModel::RulesType rules;
    Condition::ElementsType elements;
    elements.push_back(std::make_unique<ValueCondition<double_t>>(numVariable, 3));
    auto condition = std::make_unique<Condition>(elements);
    std::unique_ptr<UpdateRule> updateRule = std::make_unique<ValueSetRule>(
        catVariable, std::set<std::string>{"a", "e"});
    rules.push_back(std::make_unique<ModificationRule>(condition, updateRule));
    auto model = std::make_unique<RuleModel>(rules);

    DataGenerator generator;
    auto predictedLog = generator.predictEventLog(sourceLog.get(), model.get());

    ASSERT_EQ(sourceLog->getTraces().size(), predictedLog->getTraces().size());
    for (const auto& [caseId, trace] : predictedLog->getTraces())
    {
        ASSERT_EQ(trace->size(), sourceLog->getTraces().at(caseId)->size());
    }

    ConditionMatcher matcher;
    for (const auto& [caseId, trace] : predictedLog->getTraces())
    {
        auto targetColumn = static_cast<const LogColumn<std::string>*>(
            trace->getColumns().at(catVariable).get());
        for (size_t i = 0; i < trace->size(); i++)
        {
            if (matcher.matches(model->getRules()[0]->getCondition(),
                sourceLog->getTraces().at(caseId).get(), i))
            {
                std::set<std::string> allowedValues {"a", "e"};
                EXPECT_TRUE(allowedValues.find(targetColumn->getEntries()[i]) != allowedValues.end());
            }
            else
            {
                std::set<std::string> allowedValues {"a", "e", "i", "o", "u"};
                EXPECT_TRUE(allowedValues.find(targetColumn->getEntries()[i]) != allowedValues.end());
            }
        }
    }
}

TEST(DataGeneratorTest, PredictNumericalTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("cat",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    auto sourceLog = Generators::generateEventLog(catVariable, numVariable);

    RuleModel::RulesType rules;
    Condition::ElementsType elements;
    elements.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "a"));
    auto condition = std::make_unique<Condition>(elements);
    std::unique_ptr<UpdateRule> updateRule = std::make_unique<ValueRule>(numVariable, 11.0);
    rules.push_back(std::make_unique<ModificationRule>(condition, updateRule));
    auto model = std::make_unique<RuleModel>(rules);

    DataGenerator generator;
    auto predictedLog = generator.predictEventLog(sourceLog.get(), model.get());

    ASSERT_EQ(sourceLog->getTraces().size(), predictedLog->getTraces().size());
    for (const auto& [caseId, trace] : predictedLog->getTraces())
    {
        ASSERT_EQ(trace->size(), sourceLog->getTraces().at(caseId)->size());
    }

    ConditionMatcher matcher;
    for (const auto& [caseId, trace] : predictedLog->getTraces())
    {
        auto targetColumn = static_cast<const LogColumn<double_t>*>(
            trace->getColumns().at(numVariable).get());
        for (size_t i = 0; i < trace->size(); i++)
        {
            if (matcher.matches(model->getRules()[0]->getCondition(),
                sourceLog->getTraces().at(caseId).get(), i))
            {
                EXPECT_DOUBLE_EQ(11.0, targetColumn->getEntries()[i]);
            }
            else
            {
                EXPECT_LE(0.0, targetColumn->getEntries()[i]);
                EXPECT_LE(targetColumn->getEntries()[i], 100.0);
            }
        }
    }
}

TEST(DataGeneratorTest, PredictNumericalChangeTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("cat",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    auto sourceLog = Generators::generateEventLog(catVariable, numVariable);

    RuleModel::RulesType rules;
    Condition::ElementsType elements;
    elements.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "a"));
    auto condition = std::make_unique<Condition>(elements);
    std::unique_ptr<UpdateRule> updateRule = std::make_unique<DifferenceRule>(numVariable, 2.0);
    rules.push_back(std::make_unique<ModificationRule>(condition, updateRule));
    auto model = std::make_unique<RuleModel>(rules);

    DataGenerator generator;
    auto predictedLog = generator.predictEventLog(sourceLog.get(), model.get());

    ASSERT_EQ(sourceLog->getTraces().size(), predictedLog->getTraces().size());
    for (const auto& [caseId, trace] : predictedLog->getTraces())
    {
        ASSERT_EQ(trace->size(), sourceLog->getTraces().at(caseId)->size());
    }

    ConditionMatcher matcher;
    for (const auto& [caseId, trace] : predictedLog->getTraces())
    {
        auto targetColumn = static_cast<const LogColumn<double_t>*>(
            trace->getColumns().at(numVariable).get());
        auto sourceColumn = static_cast<const LogColumn<double_t>*>(
            sourceLog->getTraces().at(caseId)->getColumns().at(numVariable).get());
        for (size_t i = 0; i < trace->size(); i++)
        {
            const auto change = targetColumn->getEntries()[i] - sourceColumn->getEntries()[i-1];
            if (matcher.matches(model->getRules()[0]->getCondition(),
                sourceLog->getTraces().at(caseId).get(), i))
            {
                EXPECT_DOUBLE_EQ(2.0, change);
            }
            else
            {
                EXPECT_LE(0.0, targetColumn->getEntries()[i]);
                EXPECT_LE(targetColumn->getEntries()[i], 100.0);
            }
        }
    }
}

TEST(DataGeneratorTest, PredictValueSetsOverlapTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("cat",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    auto sourceLog = Generators::generateEventLog(catVariable, numVariable);

    RuleModel::RulesType rules;
    Condition::ElementsType elements;
    elements.push_back(std::make_unique<ValueCondition<double_t>>(numVariable, 3));
    auto condition = std::make_unique<Condition>(elements);
    std::unique_ptr<UpdateRule> updateRule = std::make_unique<ValueSetRule>(
        catVariable, std::set<std::string>{"a", "e"});
    rules.push_back(std::make_unique<ModificationRule>(condition, updateRule));
    Condition::ElementsType elements2;
    elements2.push_back(std::make_unique<UpperBoundCondition>(numVariable, 5));
    auto condition2 = std::make_unique<Condition>(elements2);
    std::unique_ptr<UpdateRule> updateRule2 = std::make_unique<ValueSetRule>(
        catVariable, std::set<std::string>{"o"});
    rules.push_back(std::make_unique<ModificationRule>(condition2, updateRule2));
    auto model = std::make_unique<RuleModel>(rules);

    DataGenerator generator;
    auto valueSets = generator.predictValueSets(sourceLog.get(), model.get());

    EXPECT_EQ(sourceLog->size(), valueSets.at(catVariable).size());

    std::vector<std::set<std::string>> expectedSets{
        {"a", "e", "i", "o", "u"},
        {"a", "e", "i", "o", "u"},
        {"a", "e", "i", "o", "u"},
        {"a", "e", "i", "o", "u"},
        {"o"},
        {"a", "e", "o"},
        {"a", "e", "o"},
        {"a", "e", "o"},
        {"o"},
        {"a", "e", "i", "o", "u"}
    };
    EXPECT_EQ(expectedSets, valueSets.at(catVariable));
}

TEST(DataGeneratorTest, PredictMeanValueRangeRuleTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("cat",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    auto sourceLog = Generators::generateEventLog(catVariable, numVariable);

    RuleModel::RulesType rules;
    Condition::ElementsType elements;
    elements.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "a"));
    auto condition = std::make_unique<Condition>(elements);
    std::unique_ptr<UpdateRule> updateRule = std::make_unique<ValueRangeRule>(numVariable,
        42.0, 50.0);
    rules.push_back(std::make_unique<ModificationRule>(condition, updateRule));
    auto model = std::make_unique<RuleModel>(rules);

    DataGenerator generator;
    auto predictedLog = generator.predictEventLog(sourceLog.get(), model.get(), true);

    ASSERT_EQ(sourceLog->getTraces().size(), predictedLog->getTraces().size());
    for (const auto& [caseId, trace] : predictedLog->getTraces())
    {
        ASSERT_EQ(trace->size(), sourceLog->getTraces().at(caseId)->size());
    }

    ConditionMatcher matcher;
    for (const auto& [caseId, trace] : predictedLog->getTraces())
    {
        auto targetColumn = static_cast<const LogColumn<double_t>*>(
            trace->getColumns().at(numVariable).get());
        for (size_t i = 0; i < trace->size(); i++)
        {
            if (matcher.matches(model->getRules()[0]->getCondition(),
                sourceLog->getTraces().at(caseId).get(), i))
            {
                EXPECT_EQ(46.0, targetColumn->getEntries()[i]);
            }
            else
            {
                EXPECT_EQ(50.0, targetColumn->getEntries()[i]);
            }
        }
    }
}

TEST(DataGeneratorTest, PredictMeanChangeRangeRuleTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("cat",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    auto sourceLog = Generators::generateEventLog(catVariable, numVariable);

    RuleModel::RulesType rules;
    Condition::ElementsType elements;
    elements.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "a"));
    auto condition = std::make_unique<Condition>(elements);
    std::unique_ptr<UpdateRule> updateRule = std::make_unique<ChangeRangeRule>(
        numVariable, 10.0, 15.0);
    rules.push_back(std::make_unique<ModificationRule>(condition, updateRule));
    auto model = std::make_unique<RuleModel>(rules);

    DataGenerator generator;
    auto predictedLog = generator.predictEventLog(sourceLog.get(), model.get(), true);

    ASSERT_EQ(sourceLog->getTraces().size(), predictedLog->getTraces().size());
    for (const auto& [caseId, trace] : predictedLog->getTraces())
    {
        ASSERT_EQ(trace->size(), sourceLog->getTraces().at(caseId)->size());
    }

    ConditionMatcher matcher;
    for (const auto& [caseId, trace] : predictedLog->getTraces())
    {
        auto targetColumn = static_cast<const LogColumn<double_t>*>(
            trace->getColumns().at(numVariable).get());
        auto sourceColumn = static_cast<const LogColumn<double_t>*>(
            sourceLog->getTraces().at(caseId)->getColumns().at(numVariable).get());
        for (size_t i = 0; i < trace->size(); i++)
        {
            if (matcher.matches(model->getRules()[0]->getCondition(),
                sourceLog->getTraces().at(caseId).get(), i))
            {
                const auto change = targetColumn->getEntries()[i] - sourceColumn->getEntries()[i-1];
                EXPECT_EQ(12.5, change);
            }
            else
            {
                EXPECT_EQ(50.0, targetColumn->getEntries()[i]);
            }
        }
    }
}

TEST(DataGeneratorTest, PredictMeanOverlappingValueRuleTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("cat",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    auto sourceLog = Generators::generateEventLog(catVariable, numVariable);
    numVariable->setAccuracy(0.1);

    RuleModel::RulesType rules;
    Condition::ElementsType elements;
    elements.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "a"));
    auto condition = std::make_unique<Condition>(elements);
    std::unique_ptr<UpdateRule> updateRule = std::make_unique<ValueRule>(numVariable, 22);
    rules.push_back(std::make_unique<ModificationRule>(condition, updateRule));
    Condition::ElementsType elements2;
    elements2.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "a"));
    auto condition2 = std::make_unique<Condition>(elements2);
    std::unique_ptr<UpdateRule> updateRule2 = std::make_unique<ValueRule>(numVariable, 33);
    rules.push_back(std::make_unique<ModificationRule>(condition2, updateRule2));
    auto model = std::make_unique<RuleModel>(rules);

    DataGenerator generator;
    auto predictedLog = generator.predictEventLog(sourceLog.get(), model.get(), true);

    ASSERT_EQ(sourceLog->getTraces().size(), predictedLog->getTraces().size());
    for (const auto& [caseId, trace] : predictedLog->getTraces())
    {
        ASSERT_EQ(trace->size(), sourceLog->getTraces().at(caseId)->size());
    }

    ConditionMatcher matcher;
    for (const auto& [caseId, trace] : predictedLog->getTraces())
    {
        auto targetColumn = static_cast<const LogColumn<double_t>*>(
            trace->getColumns().at(numVariable).get());
        for (size_t i = 0; i < trace->size(); i++)
        {
            if (matcher.matches(model->getRules()[0]->getCondition(),
                sourceLog->getTraces().at(caseId).get(), i))
            {
                EXPECT_EQ(27.5, targetColumn->getEntries()[i]);
            }
            else
            {
                EXPECT_EQ(50.0, targetColumn->getEntries()[i]);
            }
        }
    }
}

TEST(DataGeneratorTest, PredictMeanOverlappingSpecializationRuleTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("cat",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<CategoricalVariable> otherVariable = std::make_shared<CategoricalVariable>("other",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    auto sourceLog = Generators::generateEventLogWithNumericalUpdateRule(
        catVariable, otherVariable, numVariable);
    numVariable->setAccuracy(1.0);

    RuleModel::RulesType rules;
    Condition::ElementsType elements;
    elements.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "a"));
    auto condition = std::make_unique<Condition>(elements);
    std::unique_ptr<UpdateRule> updateRule = std::make_unique<DifferenceRule>(numVariable, -4);
    rules.push_back(std::make_unique<ModificationRule>(condition, updateRule));
    Condition::ElementsType elements2;
    elements2.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "a"));
    elements2.push_back(std::make_unique<ValueCondition<std::string>>(otherVariable, "i"));
    elements2.push_back(std::make_unique<AndOperator>());
    auto condition2 = std::make_unique<Condition>(elements2);
    std::unique_ptr<UpdateRule> updateRule2 = std::make_unique<DifferenceRule>(numVariable, -10);
    rules.push_back(std::make_unique<ModificationRule>(condition2, updateRule2));
    auto model = std::make_unique<RuleModel>(rules);

    DataGenerator generator;
    auto predictedLog = generator.predictEventLog(sourceLog.get(), model.get(), true);

    ASSERT_EQ(sourceLog->getTraces().size(), predictedLog->getTraces().size());
    for (const auto& [caseId, trace] : predictedLog->getTraces())
    {
        ASSERT_EQ(trace->size(), sourceLog->getTraces().at(caseId)->size());
    }

    ConditionMatcher matcher;
    for (const auto& [caseId, trace] : predictedLog->getTraces())
    {
        auto targetColumn = static_cast<const LogColumn<double_t>*>(
            trace->getColumns().at(numVariable).get());
        auto sourceColumn = static_cast<const LogColumn<double_t>*>(
            sourceLog->getTraces().at(caseId)->getColumns().at(numVariable).get());
        for (size_t i = 0; i < trace->size(); i++)
        {
            constexpr double_t tolerance = 1e-10;
            const auto change = targetColumn->getEntries()[i] - sourceColumn->getEntries()[i-1];
            if (matcher.matches(model->getRules()[0]->getCondition(),
                    sourceLog->getTraces().at(caseId).get(), i)
                && matcher.matches(model->getRules()[1]->getCondition(),
                    sourceLog->getTraces().at(caseId).get(), i))
            {
                EXPECT_NEAR(-7, change, tolerance);
            }
            else if (matcher.matches(model->getRules()[0]->getCondition(),
                sourceLog->getTraces().at(caseId).get(), i))
            {
                EXPECT_NEAR(-4, change, tolerance);
            }
            else
            {
                EXPECT_EQ(50.0, targetColumn->getEntries()[i]);
            }
        }
    }
}

TEST(DataGeneratorTest, PredictMeanOverlappingOrderedListTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("cat",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    auto sourceLog = Generators::generateEventLog(catVariable, numVariable);
    numVariable->setAccuracy(0.1);

    RuleModel::RulesType rules;
    Condition::ElementsType elements;
    elements.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "a"));
    auto condition = std::make_unique<Condition>(elements);
    std::unique_ptr<UpdateRule> updateRule = std::make_unique<ValueRule>(numVariable, 22);
    rules.push_back(std::make_unique<ModificationRule>(condition, updateRule));
    Condition::ElementsType elements2;
    elements2.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "a"));
    auto condition2 = std::make_unique<Condition>(elements2);
    std::unique_ptr<UpdateRule> updateRule2 = std::make_unique<ValueRule>(numVariable, 33);
    rules.push_back(std::make_unique<ModificationRule>(condition2, updateRule2));
    auto model = std::make_unique<RuleModel>(rules);

    DataGenerator generator;
    auto predictedLog = generator.predictEventLog(sourceLog.get(), model.get(), true, true);

    ASSERT_EQ(sourceLog->getTraces().size(), predictedLog->getTraces().size());
    for (const auto& [caseId, trace] : predictedLog->getTraces())
    {
        ASSERT_EQ(trace->size(), sourceLog->getTraces().at(caseId)->size());
    }

    ConditionMatcher matcher;
    for (const auto& [caseId, trace] : predictedLog->getTraces())
    {
        auto targetColumn = static_cast<const LogColumn<double_t>*>(
            trace->getColumns().at(numVariable).get());
        for (size_t i = 0; i < trace->size(); i++)
        {
            if (matcher.matches(model->getRules()[0]->getCondition(),
                sourceLog->getTraces().at(caseId).get(), i))
            {
                EXPECT_EQ(22, targetColumn->getEntries()[i]);
            }
            else
            {
                EXPECT_EQ(50.0, targetColumn->getEntries()[i]);
            }
        }
    }
}

TEST(DataGeneratorTest, PredictValueSetsOverlapOrderedListTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("cat",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    auto sourceLog = Generators::generateEventLog(catVariable, numVariable);

    RuleModel::RulesType rules;
    Condition::ElementsType elements;
    elements.push_back(std::make_unique<ValueCondition<double_t>>(numVariable, 3));
    auto condition = std::make_unique<Condition>(elements);
    std::unique_ptr<UpdateRule> updateRule = std::make_unique<ValueSetRule>(
        catVariable, std::set<std::string>{"a", "e"});
    rules.push_back(std::make_unique<ModificationRule>(condition, updateRule));
    Condition::ElementsType elements2;
    elements2.push_back(std::make_unique<UpperBoundCondition>(numVariable, 5));
    auto condition2 = std::make_unique<Condition>(elements2);
    std::unique_ptr<UpdateRule> updateRule2 = std::make_unique<ValueSetRule>(
        catVariable, std::set<std::string>{"o"});
    rules.push_back(std::make_unique<ModificationRule>(condition2, updateRule2));
    auto model = std::make_unique<RuleModel>(rules);

    DataGenerator generator;
    auto valueSets = generator.predictValueSets(sourceLog.get(), model.get(), true);

    EXPECT_EQ(sourceLog->size(), valueSets.at(catVariable).size());

    std::vector<std::set<std::string>> expectedSets{
        {"a", "e", "i", "o", "u"},
        {"a", "e", "i", "o", "u"},
        {"a", "e", "i", "o", "u"},
        {"a", "e", "i", "o", "u"},
        {"o"},
        {"a", "e"},
        {"a", "e"},
        {"a", "e"},
        {"o"},
        {"a", "e", "i", "o", "u"}
    };
    EXPECT_EQ(expectedSets, valueSets.at(catVariable));
}
