#include <gtest/gtest.h>
#include "generators.hpp"
#include "../classes/preprocessing/distributionestimator.hpp"
#include "../classes/preprocessing/nmlhistogramwrapper.hpp"
#include "../classes/datamodel/updaterules/abstractrangerule.hpp"
#include "../classes/datamodel/updaterules/valuerangerule.hpp"
#include "../classes/datamodel/updaterules/valuerule.hpp"
#include "../classes/datamodel/eventlogs/logcolumn.hpp"
#include "mock_histogramestimator.hpp"
#include "mock_columncollector.hpp"

using ::testing::Return;
using ::testing::_;

TEST(DistributionEstimatorTest, GlobalProbabilityTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::unique_ptr<EventLog> log = Generators::generateEventLog(catVariable);

    std::unique_ptr<HistogramEstimator> histEstimator = std::make_unique<NmlHistogramWrapper>();
    std::unique_ptr<ColumnCollector> collector = std::make_unique<ColumnCollector>();
    DistributionEstimator estimator(histEstimator, collector, false);
    estimator.calculateGlobalDistributions(log.get());

    EXPECT_DOUBLE_EQ(0.1, estimator.getGlobalProbability(catVariable, "i"));
    EXPECT_DOUBLE_EQ(0.2, estimator.getGlobalProbability(catVariable, "e"));
    EXPECT_DOUBLE_EQ(0.3, estimator.getGlobalProbability(catVariable, "a"));
    EXPECT_DOUBLE_EQ(0.4, estimator.getGlobalProbability(catVariable, "u"));
}

TEST(DistributionEstimatorTest, InvalidGlobalProbabilityTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::unique_ptr<EventLog> log = Generators::generateEventLog(catVariable);

    std::unique_ptr<HistogramEstimator> histEstimator = std::make_unique<NmlHistogramWrapper>();
    std::unique_ptr<ColumnCollector> collector = std::make_unique<ColumnCollector>();
    DistributionEstimator estimator(histEstimator, collector, false);
    estimator.calculateGlobalDistributions(log.get());

    EXPECT_THROW(estimator.getGlobalProbability(catVariable, "o"), std::invalid_argument);
}

TEST(DistributionEstimatorTest, RuleProbabilityTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::unique_ptr<EventLog> log = Generators::generateEventLog(catVariable);

    std::unique_ptr<ValueSetRule> rule = 
        std::make_unique<ValueSetRule>(catVariable, ValueSetRule::SetType {"i", "e"});

    std::unique_ptr<HistogramEstimator> histEstimator = std::make_unique<NmlHistogramWrapper>();
    std::unique_ptr<ColumnCollector> collector = std::make_unique<ColumnCollector>();
    DistributionEstimator estimator(histEstimator, collector, false);
    estimator.calculateGlobalDistributions(log.get());

    EXPECT_DOUBLE_EQ(1.0/3.0, estimator.getRuleProbability(rule.get(), "i"));
    EXPECT_DOUBLE_EQ(2.0/3.0, estimator.getRuleProbability(rule.get(), "e"));
}

TEST(DistributionEstimatorTest, InvalidRuleProbabilityTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::unique_ptr<EventLog> log = Generators::generateEventLog(catVariable);

    std::unique_ptr<ValueSetRule> rule = 
        std::make_unique<ValueSetRule>(catVariable, ValueSetRule::SetType {"i", "e"});

    std::unique_ptr<HistogramEstimator> histEstimator = std::make_unique<NmlHistogramWrapper>();
    std::unique_ptr<ColumnCollector> collector = std::make_unique<ColumnCollector>();
    DistributionEstimator estimator(histEstimator, collector, false);
    estimator.calculateGlobalDistributions(log.get());

    EXPECT_THROW(estimator.getRuleProbability(rule.get(), "u"), std::invalid_argument);
}

TEST(DistributionEstimatorTest, WrongCategoricalVariableTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::unique_ptr<EventLog> log = Generators::generateEventLog();

    std::unique_ptr<HistogramEstimator> histEstimator = std::make_unique<NmlHistogramWrapper>();
    std::unique_ptr<ColumnCollector> collector = std::make_unique<ColumnCollector>();
    DistributionEstimator estimator(histEstimator, collector, false);
    estimator.calculateGlobalDistributions(log.get());

    EXPECT_THROW(estimator.getGlobalProbability(catVariable, "u"), std::invalid_argument);
}

TEST(DistributionEstimatorTest, ComplexHistogramTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<EventLog> log = Generators::generateEventLog(catVariable, numVariable);

    std::unique_ptr<HistogramEstimator> histEstimator = std::make_unique<MockHistogramEstimator>();
    MockHistogramEstimator* histEstimatorReference = static_cast<MockHistogramEstimator *>(
        histEstimator.get());
    std::unique_ptr<ColumnCollector> collector = std::make_unique<ColumnCollector>();
    DistributionEstimator estimator(histEstimator, collector, false);

    Histogram histogram(
        std::vector<double_t>{0.5, 2.5, 4.5, 5.5, 9.5}, std::vector<int64_t>{1, 3, 1, 5});
    EXPECT_CALL(*histEstimatorReference, computeHistogram(_, 50, 1.0, 1.0))
        .Times(1).WillOnce(Return(histogram));

    estimator.calculateGlobalDistributions(log.get());
    EXPECT_DOUBLE_EQ(0.05, estimator.getGlobalProbability(numVariable, 1.0));
    EXPECT_DOUBLE_EQ(0.05, estimator.getGlobalProbability(numVariable, 2.0));
    EXPECT_DOUBLE_EQ(0.15, estimator.getGlobalProbability(numVariable, 3.0));
    EXPECT_DOUBLE_EQ(0.15, estimator.getGlobalProbability(numVariable, 4.0));
    EXPECT_DOUBLE_EQ(0.1, estimator.getGlobalProbability(numVariable, 5.0));
    EXPECT_DOUBLE_EQ(0.125, estimator.getGlobalProbability(numVariable, 6.0));
    EXPECT_DOUBLE_EQ(0.125, estimator.getGlobalProbability(numVariable, 7.0));
    EXPECT_DOUBLE_EQ(0.125, estimator.getGlobalProbability(numVariable, 8.0));
    EXPECT_DOUBLE_EQ(0.125, estimator.getGlobalProbability(numVariable, 9.0));
}

TEST(DistributionEstimatorTest, SingleBinHistogramTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<EventLog> log = Generators::generateEventLog(catVariable, numVariable);

    std::unique_ptr<HistogramEstimator> histEstimator = std::make_unique<MockHistogramEstimator>();
    MockHistogramEstimator* histEstimatorReference = static_cast<MockHistogramEstimator *>(
        histEstimator.get());
    std::unique_ptr<ColumnCollector> collector = std::make_unique<ColumnCollector>();
    DistributionEstimator estimator(histEstimator, collector, false);

    Histogram histogram(
        std::vector<double_t>{0.5, 9.5}, std::vector<int64_t>{10});
    EXPECT_CALL(*histEstimatorReference, computeHistogram(_, 50, 1.0, 1.0))
        .Times(1).WillOnce(Return(histogram));

    estimator.calculateGlobalDistributions(log.get());
    EXPECT_DOUBLE_EQ(1.0/9.0, estimator.getGlobalProbability(numVariable, 1.0));
    EXPECT_DOUBLE_EQ(1.0/9.0, estimator.getGlobalProbability(numVariable, 2.0));
    EXPECT_DOUBLE_EQ(1.0/9.0, estimator.getGlobalProbability(numVariable, 3.0));
    EXPECT_DOUBLE_EQ(1.0/9.0, estimator.getGlobalProbability(numVariable, 4.0));
    EXPECT_DOUBLE_EQ(1.0/9.0, estimator.getGlobalProbability(numVariable, 5.0));
    EXPECT_DOUBLE_EQ(1.0/9.0, estimator.getGlobalProbability(numVariable, 6.0));
    EXPECT_DOUBLE_EQ(1.0/9.0, estimator.getGlobalProbability(numVariable, 7.0));
    EXPECT_DOUBLE_EQ(1.0/9.0, estimator.getGlobalProbability(numVariable, 8.0));
    EXPECT_DOUBLE_EQ(1.0/9.0, estimator.getGlobalProbability(numVariable, 9.0));
}

TEST(DistributionEstimatorTest, OutsideHistogramRangeTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<EventLog> log = Generators::generateEventLog(catVariable, numVariable);

    std::unique_ptr<HistogramEstimator> histEstimator = std::make_unique<MockHistogramEstimator>();
    MockHistogramEstimator* histEstimatorReference = static_cast<MockHistogramEstimator *>(
        histEstimator.get());
    std::unique_ptr<ColumnCollector> collector = std::make_unique<ColumnCollector>();
    DistributionEstimator estimator(histEstimator, collector, false);

    Histogram histogram(
        std::vector<double_t>{0.5, 9.5}, std::vector<int64_t>{10});
    EXPECT_CALL(*histEstimatorReference, computeHistogram(_, 50, 1.0, 1.0))
        .Times(1).WillOnce(Return(histogram));

    estimator.calculateGlobalDistributions(log.get());
    EXPECT_THROW(estimator.getGlobalProbability(numVariable, 9.5001), std::invalid_argument);
}

TEST(DistributionEstimatorTest, WrongNumericalVariableTest)
{
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<EventLog> log = Generators::generateEventLog();

    std::unique_ptr<HistogramEstimator> histEstimator = std::make_unique<NmlHistogramWrapper>();
    std::unique_ptr<ColumnCollector> collector = std::make_unique<ColumnCollector>();
    DistributionEstimator estimator(histEstimator, collector, false);
    estimator.calculateGlobalDistributions(log.get());

    EXPECT_THROW(estimator.getGlobalProbability(numVariable, 4.0), std::invalid_argument);
}
