#include "generators.hpp"
#include <vector>
#include <cmath>
#include "../classes/datamodel/eventlogs/logcolumn.hpp"
#include "../classes/datamodel/eventlogs/eventlog.hpp"
#include "../classes/datamodel/conditions/transitioncondition.hpp"
#include "../classes/datamodel/conditions/condition.hpp"
#include "../classes/datamodel/conditions/andoperator.hpp"
#include "../classes/datamodel/conditions/oroperator.hpp"
#include "../classes/datamodel/conditions/valuecondition.hpp"
#include "../classes/datamodel/conditions/lowerboundcondition.hpp"
#include "../classes/datamodel/conditions/upperboundcondition.hpp"
#include "../classes/datamodel/updaterules/updaterule.hpp"
#include "../classes/datamodel/updaterules/valuerule.hpp"
#include "../classes/datamodel/updaterules/valuesetrule.hpp"
#include "../classes/datamodel/updaterules/valuerangerule.hpp"
#include "../classes/datamodel/updaterules/changerangerule.hpp"
#include "../classes/datamodel/updaterules/factorrule.hpp"
#include "../classes/datamodel/updaterules/differencerule.hpp"
#include "../classes/datamodel/modificationrule.hpp"

std::unique_ptr<Trace> Generators::generateComplexTrace(
    std::shared_ptr<CategoricalVariable> catVariable,
    std::shared_ptr<CategoricalVariable> otherVariable,
    std::shared_ptr<NumericalVariable> numVariable)
{
    numVariable->setAccuracy(1.0);
    std::vector<std::string> entries      {"u", "a", "u", "e", "a", "e", "u"};
    std::vector<std::string> otherEntries {"u", "i", "i", "a", "u", "e", "u"};
    std::vector<double_t> numEntries   {4.0, 4.0, 32.0, 42.0, -3.0, 6.0, 7.0};
    Trace::ColumnsType columns;
    columns.insert({catVariable, std::make_unique<LogColumn<std::string>>(entries)});
    columns.insert({otherVariable, std::make_unique<LogColumn<std::string>>(otherEntries)});
    columns.insert({numVariable, std::make_unique<LogColumn<double_t>>(numEntries)});

    return std::make_unique<Trace>(columns);
}

std::unique_ptr<EventLog> Generators::generateEventLog(std::shared_ptr<CategoricalVariable> catVariable,
    std::shared_ptr<NumericalVariable> numVariable)
{
    numVariable->setAccuracy(1.0);
    EventLog::TracesType traces;

    std::vector<std::string> entries1 {"u", "a", "u", "e"};
    std::vector<double_t> numEntries1 {9, 9, 7, 7};
    Trace::ColumnsType columns1;
    columns1.insert({catVariable, std::make_unique<LogColumn<std::string>>(entries1)});
    columns1.insert({numVariable, std::make_unique<LogColumn<double_t>>(numEntries1)});
    traces.insert({"0", std::make_unique<Trace>(columns1)});

    std::vector<std::string> entries2 {"i", "a", "u", "e", "a", "u"};
    std::vector<double_t> numEntries2 {1, 3, 3, 3, 5, 6};
    Trace::ColumnsType columns2;
    columns2.insert({catVariable, std::make_unique<LogColumn<std::string>>(entries2)});
    columns2.insert({numVariable, std::make_unique<LogColumn<double_t>>(numEntries2)});
    traces.insert({"1", std::make_unique<Trace>(columns2)});

    return std::make_unique<EventLog>(traces);
}

std::unique_ptr<EventLog> Generators::generateEventLogWithNumericalUpdateRule(
    std::shared_ptr<CategoricalVariable> catVariable,
    std::shared_ptr<CategoricalVariable> otherVariable,
    std::shared_ptr<NumericalVariable> numVariable)
{
    numVariable->setAccuracy(42.0);
    EventLog::TracesType traces;

    for (size_t i = 0; i < 40; i++)
    {
        std::vector<std::string> entries1 {"u", "a", "u", "a", "o"};
        std::vector<std::string> otherEntries1 {"e", "i", "i", "e", "a"};
        std::vector<double_t> numEntries1 {0, 42, 0, 0, 0};
        Trace::ColumnsType columns1;
        columns1.insert({catVariable, std::make_unique<LogColumn<std::string>>(entries1)});
        columns1.insert({otherVariable, std::make_unique<LogColumn<std::string>>(otherEntries1)});
        columns1.insert({numVariable, std::make_unique<LogColumn<double_t>>(numEntries1)});
        traces.insert({"0:" + std::to_string(i), std::make_unique<Trace>(columns1)});

        std::vector<std::string> entries2 {"i", "a", "u", "e", "a", "a"};
        std::vector<std::string> otherEntries2 {"i", "i", "i", "o", "i", "u"};
        std::vector<double_t> numEntries2 {0, 42, 0, 0, 42, 0};
        Trace::ColumnsType columns2;
        columns2.insert({catVariable, std::make_unique<LogColumn<std::string>>(entries2)});
        columns2.insert({otherVariable, std::make_unique<LogColumn<std::string>>(otherEntries2)});
        columns2.insert({numVariable, std::make_unique<LogColumn<double_t>>(numEntries2)});
        traces.insert({"1:" + std::to_string(i), std::make_unique<Trace>(columns2)});
    }

    return std::make_unique<EventLog>(traces);
}

std::unique_ptr<EventLog> Generators::generateEventLogWithCategoricalUpdateRule(
    std::shared_ptr<CategoricalVariable> catVariable,
    std::shared_ptr<NumericalVariable> numVariable)
{
    numVariable->setAccuracy(0.5);
    EventLog::TracesType traces;

    for (size_t i = 0; i < 40; i++)
    {
        std::vector<std::string> entries1 {"u", "a", "u", "a", "o"};
        std::vector<double_t> numEntries1 {-1.2, 11.0, 9.0, 15.0, 2.3};
        Trace::ColumnsType columns1;
        columns1.insert({catVariable, std::make_unique<LogColumn<std::string>>(entries1)});
        columns1.insert({numVariable, std::make_unique<LogColumn<double_t>>(numEntries1)});
        traces.insert({"0:" + std::to_string(i), std::make_unique<Trace>(columns1)});

        std::vector<std::string> entries2 {"i", "a", "u", "e", "a", "o"};
        std::vector<double_t> numEntries2 {0.0, 12.3, 4.2, 7.5, 21.7, -0.7};
        Trace::ColumnsType columns2;
        columns2.insert({catVariable, std::make_unique<LogColumn<std::string>>(entries2)});
        columns2.insert({numVariable, std::make_unique<LogColumn<double_t>>(numEntries2)});
        traces.insert({"1:" + std::to_string(i), std::make_unique<Trace>(columns2)});
    }

    return std::make_unique<EventLog>(traces);
}

std::unique_ptr<EventLog> Generators::generateEventLogWithTransitionalUpdateRule(
    std::shared_ptr<CategoricalVariable> catVariable,
    std::shared_ptr<NumericalVariable> numVariable)
{
    numVariable->setAccuracy(42.0);
    EventLog::TracesType traces;

    for (size_t i = 0; i < 40; i++)
    {
        std::vector<std::string> entries1 {"a", "e", "i", "o", "u"};
        std::vector<double_t> numEntries1 {0, 0, 0, 0, 42};
        Trace::ColumnsType columns1;
        columns1.insert({catVariable, std::make_unique<LogColumn<std::string>>(entries1)});
        columns1.insert({numVariable, std::make_unique<LogColumn<double_t>>(numEntries1)});
        traces.insert({"0:" + std::to_string(i), std::make_unique<Trace>(columns1)});

        std::vector<std::string> entries2 {"o", "u", "u", "o", "u", "u"};
        std::vector<double_t> numEntries2 {0, 42, 0, 0, 42, 0};
        Trace::ColumnsType columns2;
        columns2.insert({catVariable, std::make_unique<LogColumn<std::string>>(entries2)});
        columns2.insert({numVariable, std::make_unique<LogColumn<double_t>>(numEntries2)});
        traces.insert({"1:" + std::to_string(i), std::make_unique<Trace>(columns2)});
    }

    return std::make_unique<EventLog>(traces);
}

std::unique_ptr<EventLog> Generators::generateEventLogWithMissingValues(
    std::shared_ptr<CategoricalVariable> catVariable,
    std::shared_ptr<NumericalVariable> numVariable)
{
    numVariable->setAccuracy(0.1);
    EventLog::TracesType traces;

    std::vector<std::string> entries1 {"u", "nan", "u", "e"};
    std::vector<double_t> numEntries1 {nan(""), 9, 7, 7};
    Trace::ColumnsType columns1;
    columns1.insert({catVariable, std::make_unique<LogColumn<std::string>>(entries1)});
    columns1.insert({numVariable, std::make_unique<LogColumn<double_t>>(numEntries1)});
    traces.insert({"0", std::make_unique<Trace>(columns1)});

    std::vector<std::string> entries2 {"i", "a", "nan", "e", "a", "nan"};
    std::vector<double_t> numEntries2 {1, nan(""), 3, 3, 5, nan("")};
    Trace::ColumnsType columns2;
    columns2.insert({catVariable, std::make_unique<LogColumn<std::string>>(entries2)});
    columns2.insert({numVariable, std::make_unique<LogColumn<double_t>>(numEntries2)});
    traces.insert({"1", std::make_unique<Trace>(columns2)});

    return std::make_unique<EventLog>(traces);
}

std::unique_ptr<RuleModel> Generators::generateRuleModel(
    std::shared_ptr<CategoricalVariable> catVariable,
    std::shared_ptr<NumericalVariable> numVariable)
{
    RuleModel::RulesType rules;

    std::unique_ptr<Condition> condition = std::make_unique<Condition>();
    std::unique_ptr<UpdateRule> updateRule =
        std::make_unique<ValueRule>(numVariable, 4.0);
    rules.push_back(std::make_unique<ModificationRule>(condition, updateRule));

    std::unique_ptr<Condition> condition2 = std::make_unique<Condition>();
    std::unique_ptr<UpdateRule> updateRule2 =
        std::make_unique<ValueSetRule>(catVariable, std::set<std::string>{"e"});
    rules.push_back(std::make_unique<ModificationRule>(condition2, updateRule2));

    std::unique_ptr<Condition> condition3 = std::make_unique<Condition>();
    std::unique_ptr<UpdateRule> updateRule3 =
        std::make_unique<ValueRangeRule>(numVariable, 1.0, 2.0);
    rules.push_back(std::make_unique<ModificationRule>(condition3, updateRule3));

    return std::make_unique<RuleModel>(rules);
}

std::unique_ptr<RuleModel> Generators::generateCompleteRuleModel(
    std::shared_ptr<CategoricalVariable> catVariable,
    std::shared_ptr<NumericalVariable> numVariable)
{
    RuleModel::RulesType rules;

    Condition::ElementsType elements;
    std::unique_ptr<Condition> condition;
    std::unique_ptr<UpdateRule> updateRule;

    elements = Condition::ElementsType();
    elements.push_back(std::make_unique<AndOperator>());
    elements.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "o"));
    condition = std::make_unique<Condition>(elements);
    updateRule = std::make_unique<ValueSetRule>(
        catVariable, std::set<std::string>{"e", "i"});
    rules.push_back(std::make_unique<ModificationRule>(condition, updateRule));

    elements = Condition::ElementsType();
    elements.push_back(std::make_unique<OrOperator>());
    elements.push_back(std::make_unique<ValueCondition<double_t>>(numVariable, 1.0));
    condition = std::make_unique<Condition>(elements);
    updateRule = std::make_unique<ValueRule>(numVariable, 42.0);
    rules.push_back(std::make_unique<ModificationRule>(condition, updateRule));

    elements = Condition::ElementsType();
    elements.push_back(std::make_unique<LowerBoundCondition>(numVariable, 2.0));
    condition = std::make_unique<Condition>(elements);
    updateRule = std::make_unique<ValueRangeRule>(numVariable, 13.0, 17.0);
    rules.push_back(std::make_unique<ModificationRule>(condition, updateRule));

    elements = Condition::ElementsType();
    elements.push_back(std::make_unique<UpperBoundCondition>(numVariable, 3.0));
    condition = std::make_unique<Condition>(elements);
    updateRule = std::make_unique<ChangeRangeRule>(numVariable, 20.0, 40.0);
    rules.push_back(std::make_unique<ModificationRule>(condition, updateRule));

    elements = Condition::ElementsType();
    elements.push_back(std::make_unique<TransitionCondition<std::string>>(catVariable, "o", "e"));
    condition = std::make_unique<Condition>(elements);
    updateRule = std::make_unique<FactorRule>(numVariable, 0.9);
    rules.push_back(std::make_unique<ModificationRule>(condition, updateRule));

    elements = Condition::ElementsType();
    elements.push_back(std::make_unique<TransitionCondition<double_t>>(numVariable, -1.0, -4.0));
    condition = std::make_unique<Condition>(elements);
    updateRule = std::make_unique<DifferenceRule>(numVariable, -42.0);
    rules.push_back(std::make_unique<ModificationRule>(condition, updateRule));

    return std::make_unique<RuleModel>(rules);
}

std::unique_ptr<RuleModel> Generators::generateRuleModelForRuleMerging(
    std::shared_ptr<CategoricalVariable> catVariable,
    std::shared_ptr<NumericalVariable> numVariable)
{
    RuleModel::RulesType rules;

    Condition::ElementsType elements;
    std::unique_ptr<Condition> condition;
    std::unique_ptr<UpdateRule> updateRule;

    elements = Condition::ElementsType();
    elements.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "o"));
    condition = std::make_unique<Condition>(elements);
    updateRule = std::make_unique<ValueRule>(numVariable, 42.0);
    rules.push_back(std::make_unique<ModificationRule>(condition, updateRule));

    return std::make_unique<RuleModel>(rules);
}

std::unique_ptr<Trace> Generators::generateBelowTrace(std::shared_ptr<NumericalVariable> numVariable)
{
    numVariable->setAccuracy(1.0);
    std::vector<double_t> numEntries {3.0};
    Trace::ColumnsType columns;
    columns.insert({numVariable, std::make_unique<LogColumn<double_t>>(numEntries)});
    return std::make_unique<Trace>(columns);
}

std::unique_ptr<Trace> Generators::generateEqualTrace(std::shared_ptr<NumericalVariable> numVariable)
{
    numVariable->setAccuracy(1.0);
    std::vector<double_t> numEntries {4.0};
    Trace::ColumnsType columns;
    columns.insert({numVariable, std::make_unique<LogColumn<double_t>>(numEntries)});
    return std::make_unique<Trace>(columns);
}

std::unique_ptr<Trace> Generators::generateAboveTrace(std::shared_ptr<NumericalVariable> numVariable)
{
    numVariable->setAccuracy(1.0);
    std::vector<double_t> numEntries {5.0};
    Trace::ColumnsType columns;
    columns.insert({numVariable, std::make_unique<LogColumn<double_t>>(numEntries)});
    return std::make_unique<Trace>(columns);
}

std::pair<std::unique_ptr<AtomicCondition>, std::unique_ptr<Trace>> Generators::generateTransitionData()
{
    std::set<std::string> vowelDomain = {"a", "e", "i", "o", "u"};
    std::shared_ptr<CategoricalVariable> catVariable =
        std::make_shared<CategoricalVariable>("vowel", vowelDomain);
    std::unique_ptr<AtomicCondition> cond =
        std::make_unique<TransitionCondition<std::string>>(catVariable, "a", "u");
    
    std::vector<std::string> entries{"u", "a", "u", "e", "a", "e", "u"};
    Trace::ColumnsType columns;
    columns.insert({catVariable, std::make_unique<LogColumn<std::string>>(entries)});
    std::unique_ptr<Trace> trace = std::make_unique<Trace>(columns);

    return std::make_pair(std::move(cond), std::move(trace));
}
