#include <gtest/gtest.h>
#include "../classes/preprocessing/histogram.hpp"

TEST(HistogramTest, EmptyHistogramTest)
{
    std::vector<double_t> limits;
    std::vector<int64_t> counts;

    EXPECT_THROW(Histogram histogram(limits, counts), std::invalid_argument);
}

TEST(HistogramTest, SizeMismatchTest)
{
    std::vector<double_t> limits{1.2, 4.5, 8.9, 11.4};
    std::vector<int64_t> counts{5, 22};

    EXPECT_THROW(Histogram histogram(limits, counts), std::invalid_argument);
}

TEST(HistogramTest, UnorderedTest)
{
    std::vector<double_t> limits{1.2, 4.5, 8.9, 11.4, 5.4};
    std::vector<int64_t> counts{5, 22, 7, 44};

    EXPECT_THROW(Histogram histogram(limits, counts), std::invalid_argument);
}

TEST(HistogramTest, NegativeCountTest)
{
    std::vector<double_t> limits{1.2, 4.5, 8.9, 11.4};
    std::vector<int64_t> counts{5, 22, -7};

    EXPECT_THROW(Histogram histogram(limits, counts), std::invalid_argument);
}

TEST(HistogramTest, TooSmallValueTest)
{
    std::vector<double_t> limits{1.2, 4.5, 8.9, 11.4};
    std::vector<int64_t> counts{5, 22, 7};
    Histogram histogram(limits, counts);

    EXPECT_THROW(histogram.getCount(1.1999), std::invalid_argument);
}

TEST(HistogramTest, TooLargeValueTest)
{
    std::vector<double_t> limits{1.2, 4.5, 8.9, 11.4};
    std::vector<int64_t> counts{5, 22, 7};
    Histogram histogram(limits, counts);

    EXPECT_THROW(histogram.getBinWidth(11.4001), std::invalid_argument);
}

TEST(HistogramTest, ValueAtLowerBoundTest)
{
    std::vector<double_t> limits{1.2, 4.5, 8.9, 11.4};
    std::vector<int64_t> counts{5, 22, 7};
    Histogram histogram(limits, counts);

    EXPECT_EQ(5, histogram.getCount(1.2));
    EXPECT_EQ(4.5 - 1.2, histogram.getBinWidth(1.2));
}

TEST(HistogramTest, ValueAtUpperBoundTest)
{
    std::vector<double_t> limits{1.2, 4.5, 8.9, 11.4};
    std::vector<int64_t> counts{5, 22, 7};
    Histogram histogram(limits, counts);

    EXPECT_EQ(7, histogram.getCount(11.4));
    EXPECT_EQ(11.4 - 8.9, histogram.getBinWidth(11.4));
}

TEST(HistogramTest, ValueBetweenBinsTest)
{
    std::vector<double_t> limits{1.2, 4.5, 8.9, 11.4};
    std::vector<int64_t> counts{5, 22, 7};
    Histogram histogram(limits, counts);

    EXPECT_EQ(22, histogram.getCount(4.5));
    EXPECT_EQ(8.9 - 4.5, histogram.getBinWidth(4.5));
}

TEST(HistogramTest, ValueWithinBinTest)
{
    std::vector<double_t> limits{1.2, 4.5, 8.9, 11.4};
    std::vector<int64_t> counts{5, 22, 7};
    Histogram histogram(limits, counts);

    EXPECT_EQ(7, histogram.getCount(10.1));
    EXPECT_EQ(11.4 - 8.9, histogram.getBinWidth(10.1));
}

TEST(HistogramTest, TotalCountTest)
{
    std::vector<double_t> limits{1.2, 4.5, 8.9, 11.4};
    std::vector<int64_t> counts{5, 22, 7};
    Histogram histogram(limits, counts);

    EXPECT_EQ(5 + 22 + 7, histogram.getTotalCount());
}
