#include <gtest/gtest.h>
#include <stdexcept>
#include "../classes/scoring/mdlcalculator.hpp"

TEST(PrequentialEncoderTest, ZeroEpsilonTest)
{
    MdlCalculator calc;
    std::unordered_map<std::string, int64_t> symbolCounts = {{"x", 1}, {"c", 1}};
    EXPECT_THROW(calc.codePrequential<std::string>(symbolCounts, 0.0), std::domain_error);
}

TEST(PrequentialEncoderTest, NegativeEpsilonTest)
{
    MdlCalculator calc;
    std::unordered_map<std::string, int64_t> symbolCounts = {{"x", 1}, {"c", 1}};
    EXPECT_THROW(calc.codePrequential<std::string>(symbolCounts, -1e-13), std::domain_error);
}

TEST(PrequentialEncoderTest, OneSymbolTest)
{
    MdlCalculator calc;
    std::unordered_map<std::string, int64_t> symbolCounts = {{"x", 0}, {"c", 1}};
    EXPECT_DOUBLE_EQ(1.0, calc.codePrequential<std::string>(symbolCounts));
}

TEST(PrequentialEncoderTest, SymbolSequenceTest)
{
    MdlCalculator calc;
    std::unordered_map<std::string, int64_t> symbolCounts = {{"x", 1}, {"c", 0}};
    double_t currentBits = 1.0;
    EXPECT_NEAR(currentBits, calc.codePrequential<std::string>(symbolCounts),
        currentBits * std::pow(10.0, -(DBL_DIG - 1)));
    symbolCounts["x"] += 1;
    currentBits += 0.4150374992788438;
    EXPECT_NEAR(currentBits, calc.codePrequential<std::string>(symbolCounts),
        currentBits * std::pow(10.0, -(DBL_DIG - 1)));
    symbolCounts["x"] += 1;
    currentBits += 0.26303440583379395;
    EXPECT_NEAR(currentBits, calc.codePrequential<std::string>(symbolCounts),
        currentBits * std::pow(10.0, -(DBL_DIG - 1)));
    symbolCounts["x"] += 1;
    currentBits += 0.1926450779424;
    EXPECT_NEAR(currentBits, calc.codePrequential<std::string>(symbolCounts),
        currentBits * std::pow(10.0, -(DBL_DIG - 1)));
    symbolCounts["c"] += 1;
    currentBits += 3.321928094887362;
    EXPECT_NEAR(currentBits, calc.codePrequential<std::string>(symbolCounts),
        currentBits * std::pow(10.0, -(DBL_DIG - 1)));
    symbolCounts["c"] += 1;
    currentBits += 2.0;
    EXPECT_NEAR(currentBits, calc.codePrequential<std::string>(symbolCounts),
        currentBits * std::pow(10.0, -(DBL_DIG - 1)));
    symbolCounts["c"] += 1;
    currentBits += 1.485426827170242;
    EXPECT_NEAR(currentBits, calc.codePrequential<std::string>(symbolCounts),
        currentBits * std::pow(10.0, -(DBL_DIG - 1)));
}

TEST(PrequentialEncoderTest, NonStringTest)
{
    MdlCalculator calc;
    std::unordered_map<int64_t, int64_t> symbolCounts = {{42, 1}, {666, 0}};
    EXPECT_DOUBLE_EQ(1.0, calc.codePrequential<int64_t>(symbolCounts));
}

TEST(PrequentialEncoderTest, LargeNumbersTest)
{
    MdlCalculator calc;
    // generate a long sequence (many symbols)
    std::unordered_map<int64_t, int64_t> symbolCounts = {{0, 20000}, {1, 10000}};
    // generate many unused symbols
    for (int64_t i = 2; i < 10000; i++)
    {
        symbolCounts.insert(std::pair<int64_t, int64_t>(i, 0));
    }
    double_t currentBits = calc.codePrequential<int64_t>(symbolCounts);
    symbolCounts[0] += 1;
    currentBits += 0.9999278670512641;
    EXPECT_NEAR(currentBits, calc.codePrequential<int64_t>(symbolCounts),
        currentBits * 1e-4);
    symbolCounts[1] += 1;
    currentBits += 1.9998918046340932;
    EXPECT_NEAR(currentBits, calc.codePrequential<int64_t>(symbolCounts),
        currentBits * 1e-4);
    symbolCounts[2] += 1;
    currentBits += 11.828352893860888;
    EXPECT_NEAR(currentBits, calc.codePrequential<int64_t>(symbolCounts),
        currentBits * 1e-4);
}

TEST(PrequentialEncoderTest, NonDefaultEpsilonTest)
{
    MdlCalculator calc;
    std::unordered_map<std::string, int64_t> symbolCounts = {{"x", 1}, {"c", 0}};
    double_t currentBits = 1.0;
    EXPECT_DOUBLE_EQ(currentBits, calc.codePrequential<std::string>(symbolCounts, 1.0));
    symbolCounts["x"] += 1;
    currentBits += 0.5849625007211563;
    EXPECT_DOUBLE_EQ(currentBits, calc.codePrequential<std::string>(symbolCounts, 1.0));
}

TEST(PrequentialEncoderTest, EmptySequenceTest)
{
    MdlCalculator calc;
    std::unordered_map<std::string, int64_t> symbolCounts = {{"x", 0}};
    EXPECT_DOUBLE_EQ(0.0, calc.codePrequential<std::string>(symbolCounts));
}

TEST(PrequentialEncoderTest, TrulyEmptySequenceTest)
{
    MdlCalculator calc;
    std::unordered_map<std::string, int64_t> symbolCounts;
    EXPECT_DOUBLE_EQ(0.0, calc.codePrequential<std::string>(symbolCounts));
}

TEST(PrequentialEncoderTest, SingleZeroEpsilonTest)
{
    MdlCalculator calc;
    std::unordered_map<std::string, int64_t> symbolCounts = {{"x", 1}, {"c", 1}};
    EXPECT_THROW(calc.codeSinglePrequential<std::string>(symbolCounts, "x", 0.0),
        std::domain_error);
}

TEST(PrequentialEncoderTest, SingleNegativeEpsilonTest)
{
    MdlCalculator calc;
    std::unordered_map<std::string, int64_t> symbolCounts = {{"x", 1}, {"c", 1}};
    EXPECT_THROW(calc.codeSinglePrequential<std::string>(symbolCounts, "x", -1e-13),
        std::domain_error);
}

TEST(PrequentialEncoderTest, SingleSymbolNotInMapTest)
{
    MdlCalculator calc;
    std::unordered_map<std::string, int64_t> symbolCounts = {{"x", 1}, {"c", 1}};
    EXPECT_THROW(calc.codeSinglePrequential<std::string>(symbolCounts, "z"), std::invalid_argument);
}

TEST(PrequentialEncoderTest, SingleNoEntropyTest)
{
    MdlCalculator calc;
    std::unordered_map<std::string, int64_t> symbolCounts = {{"x", 1}};
    EXPECT_DOUBLE_EQ(0.0, calc.codeSinglePrequential<std::string>(symbolCounts, "x"));
}

TEST(PrequentialEncoderTest, SingleFirstElementTest)
{
    MdlCalculator calc;
    std::unordered_map<std::string, int64_t> symbolCounts = {{"x", 0}, {"c", 0}};
    EXPECT_DOUBLE_EQ(1.0, calc.codeSinglePrequential<std::string>(symbolCounts, "x"));
}

TEST(PrequentialEncoderTest, SingleMidSequenceTest)
{
    MdlCalculator calc;
    std::unordered_map<std::string, int64_t> symbolCounts = {{"x", 123}, {"c", 58}};
    EXPECT_DOUBLE_EQ(0.5594274086140186, calc.codeSinglePrequential<std::string>(symbolCounts, "x"));
}

TEST(PrequentialEncoderTest, SingleNonDefaultEpsilonTest)
{
    MdlCalculator calc;
    std::unordered_map<std::string, int64_t> symbolCounts = {{"x", 1}, {"c", 0}};
    EXPECT_DOUBLE_EQ(0.584962500721156, calc.codeSinglePrequential<std::string>(
        symbolCounts, "x", 1.0));
}
