#include <gtest/gtest.h>
#include "../classes/scoring/modelencoder.hpp"
#include "../classes/datamodel/updaterules/valuesetrule.hpp"
#include "../classes/datamodel/updaterules/valuerangerule.hpp"
#include "../classes/datamodel/updaterules/valuerule.hpp"
#include "../classes/datamodel/conditions/condition.hpp"
#include "../classes/datamodel/conditions/valuecondition.hpp"
#include "../classes/datamodel/conditions/transitioncondition.hpp"
#include "../classes/datamodel/conditions/andoperator.hpp"
#include "../classes/datamodel/conditions/oroperator.hpp"
#include "../classes/datamodel/conditions/notoperator.hpp"
#include "../classes/datamodel/rulemodel.hpp"
#include "../classes/datamodel/modificationrule.hpp"
#include "mock_mdlcalculator.hpp"
#include "mock_modelencoder.hpp"

using ::testing::Return;

TEST(ModelEncoderTest, ValueSetRuleTest)
{
    std::shared_ptr<MockMdlCalculator> calc = std::make_shared<MockMdlCalculator>();
    ModelEncoder enc(calc, 7, 5);
    std::set<std::string> vowelDomain = {"a", "e", "i", "o", "u"};
    std::shared_ptr<CategoricalVariable> catVariable =
        std::make_shared<CategoricalVariable>("vowel", vowelDomain);
    ValueSetRule::SetType set = {"e", "i"};
    std::unique_ptr<ValueSetRule> rule = std::make_unique<ValueSetRule>(catVariable, set);
    EXPECT_CALL(*calc, codeUniformPossibilities(6)).Times(1).WillOnce(Return(3));
    EXPECT_CALL(*calc, codeUniformPossibilities(7)).Times(1).WillOnce(Return(13));
    EXPECT_CALL(*calc, codeUniformPossibilities(4)).Times(1).WillOnce(Return(11));
    EXPECT_CALL(*calc, codeBinomial(5, 2)).Times(1).WillOnce(Return(9));
    EXPECT_DOUBLE_EQ(3+13+11+9, enc.encodeUpdateRule(rule.get()));
}

TEST(ModelEncoderTest, ValueRangeRuleTest)
{
    std::shared_ptr<MockMdlCalculator> calc = std::make_shared<MockMdlCalculator>();
    ModelEncoder enc(calc, 7, 5);
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<ValueRangeRule> rule = std::make_unique<ValueRangeRule>(numVariable, -3.0, 4.0);
    EXPECT_CALL(*calc, codeUniformPossibilities(6)).Times(1).WillOnce(Return(3));
    EXPECT_CALL(*calc, codeUniformPossibilities(7)).Times(1).WillOnce(Return(13));
    EXPECT_CALL(*calc, codeReal(-3.0, 5)).Times(1).WillOnce(Return(11));
    EXPECT_CALL(*calc, codeReal(4.0, 5)).Times(1).WillOnce(Return(9));
    EXPECT_DOUBLE_EQ(3+13+11+9, enc.encodeUpdateRule(rule.get()));
}

TEST(ModelEncoderTest, ValueRuleTest)
{
    std::shared_ptr<MockMdlCalculator> calc = std::make_shared<MockMdlCalculator>();
    ModelEncoder enc(calc, 7, 5);
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<ValueRule> rule = std::make_unique<ValueRule>(numVariable, 6.6);
    EXPECT_CALL(*calc, codeUniformPossibilities(6)).Times(1).WillOnce(Return(3));
    EXPECT_CALL(*calc, codeUniformPossibilities(7)).Times(1).WillOnce(Return(13));
    EXPECT_CALL(*calc, codeReal(6.6, 5)).Times(1).WillOnce(Return(11));
    EXPECT_DOUBLE_EQ(3+13+11, enc.encodeUpdateRule(rule.get()));
}

TEST(ModelEncoderTest, ConditionTest)
{
    std::shared_ptr<MockMdlCalculator> calc = std::make_shared<MockMdlCalculator>();
    ModelEncoder enc(calc, 2, 5);
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::set<std::string> vowelDomain = {"a", "e", "i", "o", "u"};
    std::shared_ptr<CategoricalVariable> catVariable =
        std::make_shared<CategoricalVariable>("vowel", vowelDomain);

    Condition::ElementsType elements;
    elements.push_back(std::make_unique<ValueCondition<double_t>>(numVariable, 42.0));
    elements.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "u"));
    elements.push_back(std::make_unique<AndOperator>());
    elements.push_back(std::make_unique<TransitionCondition<double_t>>(numVariable, 1.0, 3.0));
    elements.push_back(std::make_unique<TransitionCondition<std::string>>(catVariable, "i", "e"));
    elements.push_back(std::make_unique<AndOperator>());
    elements.push_back(std::make_unique<OrOperator>());
    elements.push_back(std::make_unique<NotOperator>());

    std::unique_ptr<Condition> condition = std::make_unique<Condition>(elements);

    EXPECT_CALL(*calc, codePositiveInteger(9)).Times(1).WillOnce(Return(14));
    EXPECT_CALL(*calc, codeUniformPossibilities(2)).Times(1).WillOnce(Return(11));
    EXPECT_CALL(*calc, codeUniformPossibilities(7)).Times(1).WillOnce(Return(17));
    EXPECT_CALL(*calc, codeReal(42.0, 5)).Times(1).WillOnce(Return(6));
    EXPECT_CALL(*calc, codeUniformPossibilities(5)).Times(2).WillOnce(Return(3)).WillOnce(Return(8));
    EXPECT_CALL(*calc, codeReal(1.0, 5)).Times(1).WillOnce(Return(13));
    EXPECT_CALL(*calc, codeReal(3.0, 5)).Times(1).WillOnce(Return(2));

    EXPECT_DOUBLE_EQ(14 + 4*11 + 8*17 + 6 + 3 + 2*8 + 13 + 2, enc.encodeCondition(condition.get()));
}

TEST(ModelEncoderTest, RuleModelTest)
{
    std::shared_ptr<MockMdlCalculator> calc = std::make_shared<MockMdlCalculator>();
    MockModelEncoder enc(calc, 7, 5);
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");

    RuleModel::RulesType rules;
    std::unique_ptr<Condition> condition1 = std::make_unique<Condition>();
    std::unique_ptr<UpdateRule> rule1 = std::make_unique<ValueRule>(numVariable, 42.0);
    rules.push_back(std::make_unique<ModificationRule>(condition1, rule1));
    std::unique_ptr<Condition> condition2 = std::make_unique<Condition>();
    std::unique_ptr<UpdateRule> rule2 = std::make_unique<ValueRule>(numVariable, 42.0);
    rules.push_back(std::make_unique<ModificationRule>(condition2, rule2));
    std::unique_ptr<Condition> condition3 = std::make_unique<Condition>();
    std::unique_ptr<UpdateRule> rule3 = std::make_unique<ValueRule>(numVariable, 42.0);
    rules.push_back(std::make_unique<ModificationRule>(condition3, rule3));

    std::unique_ptr<RuleModel> model = std::make_unique<RuleModel>(rules);

    EXPECT_CALL(*calc, codePositiveInteger(4)).Times(1).WillOnce(Return(14));
    EXPECT_CALL(enc, encodeCondition).Times(3).WillOnce(Return(2)).WillOnce(Return(3)).WillOnce(Return(5));
    EXPECT_CALL(enc, encodeUpdateRule).Times(3).WillOnce(Return(7)).WillOnce(Return(11)).WillOnce(Return(13));

    EXPECT_DOUBLE_EQ(14 + 2 + 3 + 5 + 7 + 11 + 13, enc.encodeRuleModel(model.get()));
}
