#include <gtest/gtest.h>
#include "generators.hpp"
#include "../classes/search/modelfinder.hpp"
#include "../classes/search/forwardrulegenerator.hpp"
#include "../classes/search/backwardrulegenerator.hpp"
#include "../classes/datamodel/updaterules/valuerule.hpp"
#include "../classes/datamodel/conditions/valuecondition.hpp"
#include "../classes/datamodel/conditions/oroperator.hpp"

TEST(ModelFinderTest, SimpleExecutionTest)
{
    auto log = Generators::generateEventLog();
    ModelFinder finder;
    auto condition = std::make_unique<Condition>();
    auto distributionEstimator = std::make_shared<DistributionEstimator>();
    distributionEstimator->calculateGlobalDistributions(log.get());
    ForwardRuleGenerator forwardGen (distributionEstimator, log.get(), condition.get());
    auto model = finder.findRuleModel(log.get(), &forwardGen);
    BackwardRuleGenerator backwardGen (distributionEstimator);
    auto reverseModel = finder.findRuleModel(log.get(), &backwardGen);
    EXPECT_TRUE(true);
}

TEST(ModelFinderTest, AddRuleToEmptyModel)
{
    std::unique_ptr<RuleModel> model = std::make_unique<RuleModel>();
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");

    std::unique_ptr<Condition> condition = std::make_unique<Condition>();
    std::unique_ptr<UpdateRule> updateRule =
        std::make_unique<ValueRule>(numVariable, 4.0);
    std::unique_ptr<ModificationRule> ruleToAdd =
        std::make_unique<ModificationRule>(condition, updateRule);
    
    ModelFinder finder;
    finder.addRuleWhileMergingConditions(model.get(), std::move(ruleToAdd));
    ASSERT_EQ(1, model->getRules().size());
    EXPECT_EQ(0, model->getRules()[0]->getCondition()->getElements().size());
    auto retrievedUpdateRule = dynamic_cast<const ValueRule*>(model->getRules()[0]->getUpdateRule());
    ASSERT_TRUE(retrievedUpdateRule != nullptr);
    EXPECT_DOUBLE_EQ(4.0, retrievedUpdateRule->getConstant());
}

TEST(ModelFinderTest, RuleMergedPredictionSlightlyOff)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>(
        "vowel", std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<RuleModel> model = Generators::generateRuleModelForRuleMerging(
        catVariable, numVariable);

    Condition::ElementsType elements;
    elements.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "a"));
    std::unique_ptr<Condition> condition = std::make_unique<Condition>(elements);
    std::unique_ptr<UpdateRule> updateRule = std::make_unique<ValueRule>(numVariable, 42.00000049);
    auto ruleToAdd = std::make_unique<ModificationRule>(condition, updateRule);

    ModelFinder finder;
    finder.addRuleWhileMergingConditions(model.get(), std::move(ruleToAdd));
    ASSERT_EQ(1, model->getRules().size());
    auto retrievedUpdateRule = dynamic_cast<const ValueRule*>(model->getRules()[0]->getUpdateRule());
    ASSERT_TRUE(retrievedUpdateRule != nullptr);
    EXPECT_DOUBLE_EQ(42.0, retrievedUpdateRule->getConstant());

    const Condition::ElementsType& elements1 = model->getRules()[0]->getCondition()->getElements();
    ASSERT_EQ(3, elements1.size());
    auto firstCondition = dynamic_cast<const ValueCondition<std::string>*>(elements1[0].get());
    auto secondCondition = dynamic_cast<const ValueCondition<std::string>*>(elements1[1].get());
    auto conditionOperator = dynamic_cast<const OrOperator*>(elements1[2].get());
    ASSERT_TRUE(firstCondition != nullptr);
    ASSERT_TRUE(secondCondition != nullptr);
    ASSERT_TRUE(conditionOperator != nullptr);
    EXPECT_EQ("o", firstCondition->getValue());
    EXPECT_EQ("a", secondCondition->getValue());
}

TEST(ModelFinderTest, RuleMergedEmptyCondition)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>(
        "vowel", std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<RuleModel> model = Generators::generateRuleModelForRuleMerging(
        catVariable, numVariable);

    std::unique_ptr<Condition> condition = std::make_unique<Condition>();
    std::unique_ptr<UpdateRule> updateRule = std::make_unique<ValueRule>(numVariable, 42.0);
    auto ruleToAdd = std::make_unique<ModificationRule>(condition, updateRule);

    ModelFinder finder;
    finder.addRuleWhileMergingConditions(model.get(), std::move(ruleToAdd));
    ASSERT_EQ(1, model->getRules().size());
    auto retrievedUpdateRule = dynamic_cast<const ValueRule*>(model->getRules()[0]->getUpdateRule());
    ASSERT_TRUE(retrievedUpdateRule != nullptr);
    EXPECT_DOUBLE_EQ(42.0, retrievedUpdateRule->getConstant());
    EXPECT_EQ(0, model->getRules()[0]->getCondition()->getElements().size());
}

TEST(ModelFinderTest, RuleNotMergedPredictionOff)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>(
        "vowel", std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<RuleModel> model = Generators::generateRuleModelForRuleMerging(
        catVariable, numVariable);

    Condition::ElementsType elements;
    elements.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "a"));
    std::unique_ptr<Condition> condition = std::make_unique<Condition>(elements);
    std::unique_ptr<UpdateRule> updateRule = std::make_unique<ValueRule>(numVariable, 42.000001);
    auto ruleToAdd = std::make_unique<ModificationRule>(condition, updateRule);

    ModelFinder finder;
    finder.addRuleWhileMergingConditions(model.get(), std::move(ruleToAdd));
    ASSERT_EQ(2, model->getRules().size());
    
    const Condition::ElementsType& elements1 = model->getRules()[0]->getCondition()->getElements();
    ASSERT_EQ(1, elements1.size());
    auto firstCondition = dynamic_cast<const ValueCondition<std::string>*>(elements1[0].get());
    ASSERT_TRUE(firstCondition != nullptr);
    EXPECT_EQ("o", firstCondition->getValue());
    auto firstUpdateRule = dynamic_cast<const ValueRule*>(model->getRules()[0]->getUpdateRule());
    ASSERT_TRUE(firstUpdateRule != nullptr);
    EXPECT_DOUBLE_EQ(42.0, firstUpdateRule->getConstant());

    const Condition::ElementsType& elements2 = model->getRules()[1]->getCondition()->getElements();
    ASSERT_EQ(1, elements2.size());
    auto secondCondition = dynamic_cast<const ValueCondition<std::string>*>(elements2[0].get());
    ASSERT_TRUE(secondCondition != nullptr);
    EXPECT_EQ("a", secondCondition->getValue());
    auto secondUpdateRule = dynamic_cast<const ValueRule*>(model->getRules()[1]->getUpdateRule());
    ASSERT_TRUE(secondUpdateRule != nullptr);
    EXPECT_DOUBLE_EQ(42.000001, secondUpdateRule->getConstant());
}

TEST(ModelFinderTest, RuleNotMergedVariableOff)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>(
        "vowel", std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::shared_ptr<NumericalVariable> numVariable2 = std::make_shared<NumericalVariable>("num2");
    std::unique_ptr<RuleModel> model = Generators::generateRuleModelForRuleMerging(
        catVariable, numVariable);

    Condition::ElementsType elements;
    elements.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "a"));
    std::unique_ptr<Condition> condition = std::make_unique<Condition>(elements);
    std::unique_ptr<UpdateRule> updateRule = std::make_unique<ValueRule>(numVariable2, 42.0);
    auto ruleToAdd = std::make_unique<ModificationRule>(condition, updateRule);

    ModelFinder finder;
    finder.addRuleWhileMergingConditions(model.get(), std::move(ruleToAdd));
    ASSERT_EQ(2, model->getRules().size());
    
    const Condition::ElementsType& elements1 = model->getRules()[0]->getCondition()->getElements();
    ASSERT_EQ(1, elements1.size());
    auto firstCondition = dynamic_cast<const ValueCondition<std::string>*>(elements1[0].get());
    ASSERT_TRUE(firstCondition != nullptr);
    EXPECT_EQ("o", firstCondition->getValue());
    auto firstUpdateRule = dynamic_cast<const ValueRule*>(model->getRules()[0]->getUpdateRule());
    ASSERT_TRUE(firstUpdateRule != nullptr);
    EXPECT_DOUBLE_EQ(42.0, firstUpdateRule->getConstant());
    EXPECT_EQ(numVariable, firstUpdateRule->getVariable());

    const Condition::ElementsType& elements2 = model->getRules()[1]->getCondition()->getElements();
    ASSERT_EQ(1, elements2.size());
    auto secondCondition = dynamic_cast<const ValueCondition<std::string>*>(elements2[0].get());
    ASSERT_TRUE(secondCondition != nullptr);
    EXPECT_EQ("a", secondCondition->getValue());
    auto secondUpdateRule = dynamic_cast<const ValueRule*>(model->getRules()[1]->getUpdateRule());
    ASSERT_TRUE(secondUpdateRule != nullptr);
    EXPECT_DOUBLE_EQ(42.0, secondUpdateRule->getConstant());
    EXPECT_EQ(numVariable2, secondUpdateRule->getVariable());
}
