#include <gtest/gtest.h>
#include "../classes/datamodel/modificationrule.hpp"
#include "../classes/datamodel/updaterules/valuerule.hpp"
#include "../classes/datamodel/updaterules/factorrule.hpp"
#include "../classes/datamodel/conditions/andoperator.hpp"

TEST(ModificationRuleTest, ConstructorSuccessTest)
{
    std::unique_ptr<Condition> condition = std::make_unique<Condition>();
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<UpdateRule> updateRule = std::make_unique<ValueRule>(numVariable, 1.0);
    ASSERT_TRUE(updateRule != nullptr);
    ModificationRule modRule(condition, updateRule);
    EXPECT_TRUE(updateRule == nullptr);
    const ValueRule* updateRuleInModRule = dynamic_cast<const ValueRule*>(modRule.getUpdateRule());
    ASSERT_TRUE(updateRuleInModRule != nullptr);
    EXPECT_DOUBLE_EQ(1.0, updateRuleInModRule->getConstant());
}

TEST(ModificationRuleTest, ConstructorExceptionUpdateRuleTest)
{
    std::unique_ptr<Condition> condition = std::make_unique<Condition>();
    std::unique_ptr<UpdateRule> updateRule;
    EXPECT_THROW(ModificationRule modRule(condition, updateRule), std::invalid_argument);
}

TEST(ModificationRuleTest, ConstructorExceptionConditionTest)
{
    std::unique_ptr<Condition> condition;
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<UpdateRule> updateRule = std::make_unique<ValueRule>(numVariable, 1.0);
    EXPECT_THROW(ModificationRule modRule(condition, updateRule), std::invalid_argument);
}

TEST(ModificationRuleTest, UpdateRuleSetSuccessTest)
{
    std::unique_ptr<Condition> condition = std::make_unique<Condition>();
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<UpdateRule> updateRule = std::make_unique<ValueRule>(numVariable, 1.0);
    ModificationRule modRule(condition, updateRule);
    std::unique_ptr<UpdateRule> newUpdateRule = std::make_unique<FactorRule>(numVariable, 2.0);
    ASSERT_TRUE(newUpdateRule != nullptr);
    modRule.setUpdateRule(newUpdateRule);
    EXPECT_TRUE(newUpdateRule == nullptr);
    const FactorRule* updateRuleInModRule = dynamic_cast<const FactorRule*>(modRule.getUpdateRule());
    ASSERT_TRUE(updateRuleInModRule != nullptr);
    EXPECT_DOUBLE_EQ(2.0, updateRuleInModRule->getConstant());
}

TEST(ModificationRuleTest, UpdateRuleSetExceptionTest)
{
    std::unique_ptr<Condition> condition = std::make_unique<Condition>();
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<UpdateRule> updateRule = std::make_unique<ValueRule>(numVariable, 1.0);
    ModificationRule modRule(condition, updateRule);
    std::unique_ptr<UpdateRule> newUpdateRule;
    EXPECT_THROW(modRule.setUpdateRule(newUpdateRule), std::invalid_argument);
}

TEST(ModificationRuleTest, ConditionSetSuccessTest)
{
    std::unique_ptr<Condition> condition = std::make_unique<Condition>();
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<UpdateRule> updateRule = std::make_unique<ValueRule>(numVariable, 1.0);
    ModificationRule modRule(condition, updateRule);
    Condition::ElementsType elements;
    elements.push_back(std::make_unique<AndOperator>());
    std::unique_ptr<Condition> newCondition = std::make_unique<Condition>(elements);
    ASSERT_TRUE(newCondition != nullptr);
    modRule.setCondition(newCondition);
    EXPECT_TRUE(newCondition == nullptr);
    EXPECT_TRUE(dynamic_cast<const AndOperator *>(
        modRule.getCondition()->getElements()[0].get()) != nullptr);
}

TEST(ModificationRuleTest, ConditionSetExceptionTest)
{
    std::unique_ptr<Condition> condition = std::make_unique<Condition>();
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<UpdateRule> updateRule = std::make_unique<ValueRule>(numVariable, 1.0);
    ModificationRule modRule(condition, updateRule);
    std::unique_ptr<Condition> newCondition;
    EXPECT_THROW(modRule.setCondition(newCondition), std::invalid_argument);
}
