#include <gtest/gtest.h>
#include "generators.hpp"
#include "../classes/scoring/rulefinder.hpp"
#include "../classes/datamodel/eventlogs/logcolumn.hpp"
#include "../classes/datamodel/conditions/valuecondition.hpp"
#include "../classes/datamodel/updaterules/valuesetrule.hpp"
#include "../classes/datamodel/updaterules/valuerule.hpp"
#include "../classes/datamodel/updaterules/factorrule.hpp"
#include "../classes/datamodel/updaterules/differencerule.hpp"

TEST(RuleFinderTest, EmptyModelTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::unique_ptr<Trace> trace = Generators::generateComplexTrace(catVariable);

    RuleModel::RulesType rules;
    std::unique_ptr<RuleModel> model = std::make_unique<RuleModel>(rules);

    RuleFinder finder;
    std::vector<const ModificationRule*> result = finder.findRulesThatApply(
        model.get(), catVariable.get(), trace.get(), 0);
    EXPECT_EQ(0, result.size());
}

TEST(RuleFinderTest, NoMatchingRuleTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<CategoricalVariable> otherVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::unique_ptr<Trace> trace = Generators::generateComplexTrace(catVariable, otherVariable);

    Condition::ElementsType elements;
    elements.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "o"));
    std::unique_ptr<Condition> condition = std::make_unique<Condition>(elements);
    std::unique_ptr<UpdateRule> update = std::make_unique<ValueSetRule>(otherVariable, std::set<std::string>{"o"});
    RuleModel::RulesType rules;
    rules.push_back(std::make_unique<ModificationRule>(condition, update));
    std::unique_ptr<RuleModel> model = std::make_unique<RuleModel>(rules);

    RuleFinder finder;
    std::vector<const ModificationRule*> result = finder.findRulesThatApply(
        model.get(), otherVariable.get(), trace.get(), 0);
    EXPECT_EQ(0, result.size());
}

TEST(RuleFinderTest, OnlyRuleForOtherVariableTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<CategoricalVariable> otherVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::unique_ptr<Trace> trace = Generators::generateComplexTrace(catVariable, otherVariable);

    Condition::ElementsType elements;
    elements.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "e"));
    std::unique_ptr<Condition> condition = std::make_unique<Condition>(elements);
    std::unique_ptr<UpdateRule> update = std::make_unique<ValueSetRule>(otherVariable, std::set<std::string>{"e"});
    RuleModel::RulesType rules;
    rules.push_back(std::make_unique<ModificationRule>(condition, update));
    std::unique_ptr<RuleModel> model = std::make_unique<RuleModel>(rules);

    RuleFinder finder;
    std::vector<const ModificationRule*> result = finder.findRulesThatApply(
        model.get(), catVariable.get(), trace.get(), 5);
    EXPECT_EQ(0, result.size());
}

TEST(RuleFinderTest, OneRuleTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<CategoricalVariable> otherVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<Trace> trace = Generators::generateComplexTrace(catVariable, otherVariable, numVariable);

    Condition::ElementsType elements;
    elements.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "e"));
    std::unique_ptr<Condition> condition = std::make_unique<Condition>(elements);
    std::unique_ptr<UpdateRule> update = std::make_unique<ValueSetRule>(otherVariable, std::set<std::string>{"e"});

    Condition::ElementsType elements2;
    elements2.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "e"));
    std::unique_ptr<Condition> condition2 = std::make_unique<Condition>(elements);
    std::unique_ptr<UpdateRule> update2 = std::make_unique<ValueRule>(numVariable, 6.0);

    RuleModel::RulesType rules;
    std::unique_ptr<ModificationRule> ruleToBeFound = std::make_unique<ModificationRule>(condition, update);
    const ModificationRule* ruleToBeFoundCopy = ruleToBeFound.get();
    rules.push_back(std::move(ruleToBeFound));
    rules.push_back(std::make_unique<ModificationRule>(condition2, update2));
    std::unique_ptr<RuleModel> model = std::make_unique<RuleModel>(rules);

    RuleFinder finder;
    std::vector<const ModificationRule*> result = finder.findRulesThatApply(
        model.get(), otherVariable.get(), trace.get(), 5);
    EXPECT_EQ(1, result.size());
    EXPECT_EQ(result[0], ruleToBeFoundCopy);
}

TEST(RuleFinderTest, MultipleRulesTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<CategoricalVariable> otherVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<Trace> trace = Generators::generateComplexTrace(catVariable, otherVariable, numVariable);

    Condition::ElementsType elements;
    elements.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "e"));
    std::unique_ptr<Condition> condition = std::make_unique<Condition>(elements);
    std::unique_ptr<UpdateRule> update = std::make_unique<ValueSetRule>(otherVariable, std::set<std::string>{"e"});

    Condition::ElementsType elements2;
    elements2.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "e"));
    std::unique_ptr<Condition> condition2 = std::make_unique<Condition>(elements);
    std::unique_ptr<UpdateRule> update2 = std::make_unique<ValueRule>(numVariable, 6.0);

    Condition::ElementsType elements3;
    elements3.push_back(std::make_unique<ValueCondition<double_t>>(numVariable, 6.0));
    std::unique_ptr<Condition> condition3 = std::make_unique<Condition>(elements);
    std::unique_ptr<UpdateRule> update3 = std::make_unique<ValueSetRule>(otherVariable, std::set<std::string>{"e"});

    RuleModel::RulesType rules;
    std::unique_ptr<ModificationRule> ruleToBeFound = std::make_unique<ModificationRule>(condition, update);
    const ModificationRule* ruleToBeFoundCopy = ruleToBeFound.get();
    rules.push_back(std::move(ruleToBeFound));
    rules.push_back(std::make_unique<ModificationRule>(condition2, update2));
    std::unique_ptr<ModificationRule> ruleToBeFound2 = std::make_unique<ModificationRule>(condition3, update3);
    const ModificationRule* ruleToBeFoundCopy2 = ruleToBeFound2.get();
    rules.push_back(std::move(ruleToBeFound2));
    std::unique_ptr<RuleModel> model = std::make_unique<RuleModel>(rules);

    RuleFinder finder;
    std::vector<const ModificationRule*> result = finder.findRulesThatApply(
        model.get(), otherVariable.get(), trace.get(), 5);
    EXPECT_EQ(2, result.size());
    EXPECT_EQ(result[0], ruleToBeFoundCopy);
    EXPECT_EQ(result[1], ruleToBeFoundCopy2);
}

TEST(RuleFinderTest, ChangeRuleAtFirstEntryTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<CategoricalVariable> otherVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<Trace> trace = Generators::generateComplexTrace(catVariable, otherVariable, numVariable);

    Condition::ElementsType elements;
    elements.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "u"));
    std::unique_ptr<Condition> condition = std::make_unique<Condition>(elements);
    std::unique_ptr<UpdateRule> update = std::make_unique<FactorRule>(numVariable, 2.0);

    RuleModel::RulesType rules;
    rules.push_back(std::make_unique<ModificationRule>(condition, update));
    std::unique_ptr<RuleModel> model = std::make_unique<RuleModel>(rules);

    RuleFinder finder;
    std::vector<const ModificationRule*> result = finder.findRulesThatApply(
        model.get(), numVariable.get(), trace.get(), 0);
    EXPECT_EQ(0, result.size());
}

TEST(RuleFinderTest, ChangeRuleAfterMissingValueTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    numVariable->setAccuracy(0.1);

    std::vector<std::string> entries {"a", "e", "i", "o", "u"};
    std::vector<double_t> numEntries {1, 9, 4, nan(""), 2};
    Trace::ColumnsType columns;
    columns.insert({catVariable, std::make_unique<LogColumn<std::string>>(entries)});
    columns.insert({numVariable, std::make_unique<LogColumn<double_t>>(numEntries)});
    std::unique_ptr<Trace> trace = std::make_unique<Trace>(columns);

    Condition::ElementsType elements;
    elements.push_back(std::make_unique<ValueCondition<std::string>>(catVariable, "u"));
    std::unique_ptr<Condition> condition = std::make_unique<Condition>(elements);
    std::unique_ptr<UpdateRule> update = std::make_unique<DifferenceRule>(numVariable, -4.0);

    RuleModel::RulesType rules;
    rules.push_back(std::make_unique<ModificationRule>(condition, update));
    std::unique_ptr<RuleModel> model = std::make_unique<RuleModel>(rules);

    RuleFinder finder;
    std::vector<const ModificationRule*> result = finder.findRulesThatApply(
        model.get(), numVariable.get(), trace.get(), 4);
    EXPECT_EQ(0, result.size());
}
