#include <gtest/gtest.h>
#include <fcntl.h>
#include <capnp/serialize.h>
#include "generators.hpp"
#include "../classes/serialobjectconverter.hpp"
#include "../classes/datamodel/eventlogs/logcolumn.hpp"
#include "../classes/datamodel/conditions/andoperator.hpp"
#include "../classes/datamodel/conditions/oroperator.hpp"
#include "../classes/datamodel/conditions/transitioncondition.hpp"
#include "../classes/datamodel/updaterules/differencerule.hpp"

std::shared_ptr<CategoricalVariable> getCategoricalVariableFromLog(const EventLog* const log, const std::string& traceId)
{
    std::shared_ptr<CategoricalVariable> catVariable;
    for (const auto& columnPair : log->getTraces().at(traceId)->getColumns())
    {
        catVariable = std::dynamic_pointer_cast<CategoricalVariable>(columnPair.first);
        if (catVariable != nullptr)
        {
            break;
        }
    }
    return catVariable;
}

TEST(SerialObjectConverterTest, EventLogReconstuctionTest)
{
    const std::string filePath = "/tmp/mySerialEventLog.bin";

    ::capnp::MallocMessageBuilder outputMessage;
    serial::EventLog::Builder builder = outputMessage.initRoot<serial::EventLog>();

    std::unique_ptr<EventLog> log = Generators::generateEventLog();
    SerialObjectConverter converter;

    converter.eventLogToSerial(log.get(), builder);

    int writeFileDescriptor = open(filePath.c_str(), O_CREAT | O_WRONLY);
    writeMessageToFd(writeFileDescriptor, outputMessage);
    ASSERT_EQ(0, close(writeFileDescriptor));
    
    int readFileDescriptor = open(filePath.c_str(), O_RDONLY);
    ::capnp::StreamFdMessageReader inputMessage(readFileDescriptor);
    serial::EventLog::Reader reader = inputMessage.getRoot<serial::EventLog>();
    
    std::unique_ptr<EventLog> logFromFile = converter.eventLogFromSerial(reader);

    EXPECT_EQ(log->getVariables().size(), logFromFile->getVariables().size());
    EXPECT_EQ(log->getTraces().size(), logFromFile->getTraces().size());
    EXPECT_EQ(log->getTraces().at("0")->size(),
        logFromFile->getTraces().at("0")->size());

    std::shared_ptr<CategoricalVariable> catVariable =
        getCategoricalVariableFromLog(log.get(), "0");
    std::shared_ptr<CategoricalVariable> catVariableFromFile =
        getCategoricalVariableFromLog(logFromFile.get(), "0");

    EXPECT_EQ(catVariable->getDomain(), catVariableFromFile->getDomain());

    auto catColumn = static_cast<const LogColumn<std::string>*>(
        log->getTraces().at("0")->getColumns().at(catVariable).get());
    auto catColumnFromFile = static_cast<const LogColumn<std::string>*>(
        logFromFile->getTraces().at("0")->getColumns().at(catVariableFromFile).get());
    EXPECT_EQ(catColumn->getEntries(), catColumnFromFile->getEntries());

    ASSERT_EQ(0, close(readFileDescriptor));
}

TEST(SerialObjectConverterTest, RuleModelReconstuctionTest)
{
    const std::string filePath = "/tmp/mySerialModel.bin";

    ::capnp::MallocMessageBuilder outputMessage;
    serial::RuleModel::Builder builder = outputMessage.initRoot<serial::RuleModel>();

    std::unique_ptr<RuleModel> model = Generators::generateCompleteRuleModel();
    SerialObjectConverter converter;

    converter.ruleModelToSerial(model.get(), builder);

    int writeFileDescriptor = open(filePath.c_str(), O_CREAT | O_WRONLY);
    writeMessageToFd(writeFileDescriptor, outputMessage);
    ASSERT_EQ(0, close(writeFileDescriptor));
    
    int readFileDescriptor = open(filePath.c_str(), O_RDONLY);
    ::capnp::StreamFdMessageReader inputMessage(readFileDescriptor);
    serial::RuleModel::Reader reader = inputMessage.getRoot<serial::RuleModel>();
    
    std::unique_ptr<RuleModel> modelFromFile = converter.ruleModelFromSerial(reader);

    EXPECT_EQ(model->getRules().size(), modelFromFile->getRules().size());

    auto andOperator = dynamic_cast<const AndOperator*>(
        modelFromFile->getRules()[0]->getCondition()->getElements()[0].get());
    EXPECT_TRUE(andOperator != nullptr);

    auto orOperator = dynamic_cast<const OrOperator*>(
        modelFromFile->getRules()[1]->getCondition()->getElements()[0].get());
    EXPECT_TRUE(orOperator != nullptr);

    auto numTransitionCondition = dynamic_cast<const TransitionCondition<double_t>*>(
        modelFromFile->getRules()[5]->getCondition()->getElements()[0].get());
    ASSERT_TRUE(numTransitionCondition != nullptr);
    EXPECT_EQ(-1.0, numTransitionCondition->getBefore());
    EXPECT_EQ(-4.0, numTransitionCondition->getAfter());

    auto differenceRule = dynamic_cast<const DifferenceRule*>(
        modelFromFile->getRules()[5]->getUpdateRule());
    ASSERT_TRUE(differenceRule != nullptr);
    EXPECT_EQ(-42.0, differenceRule->getConstant());

    ASSERT_EQ(0, close(readFileDescriptor));
}
