#include <gtest/gtest.h>
#include <algorithm>
#include <random>
#include "../classes/statistics.hpp"
#include "../classes/datamodel/eventlogs/categoricalvariable.hpp"
#include "../classes/datamodel/eventlogs/numericalvariable.hpp"

TEST(StatisticsTest, FrequentValuesMoreThanExistTest)
{
    Statistics stats;
    std::default_random_engine rng;
    std::vector<std::string> input {"a", "b", "b", "b", "c", "c", "d", "d", "d", "d", "d"};
    std::shuffle(input.begin(), input.end(), rng);
    auto result = stats.getMostFrequentValues(input, 5, -1);
    EXPECT_EQ(Statistics::ValueFrequencies<std::string>(
        {{"d", 5}, {"b", 3}, {"c", 2}, {"a", 1}}), result);
}

TEST(StatisticsTest, FrequentValuesAllTest)
{
    Statistics stats;
    std::default_random_engine rng;
    std::vector<std::string> input {"a", "b", "b", "b", "c", "c", "d", "d", "d", "d", "d"};
    std::shuffle(input.begin(), input.end(), rng);
    auto result = stats.getMostFrequentValues(input, 4, -1);
    EXPECT_EQ(Statistics::ValueFrequencies<std::string>(
        {{"d", 5}, {"b", 3}, {"c", 2}, {"a", 1}}), result);
}

TEST(StatisticsTest, FrequentValuesOneTest)
{
    Statistics stats;
    std::default_random_engine rng;
    std::vector<std::string> input {"a", "b", "b", "b", "c", "c", "d", "d", "d", "d", "d"};
    std::shuffle(input.begin(), input.end(), rng);
    auto result = stats.getMostFrequentValues(input, 1, -1);
    EXPECT_EQ(Statistics::ValueFrequencies<std::string>({{"d", 5}}), result);
}

TEST(StatisticsTest, FrequentValuesNoneTest)
{
    Statistics stats;
    std::default_random_engine rng;
    std::vector<std::string> input {"a", "b", "b", "b", "c", "c", "d", "d", "d", "d", "d"};
    std::shuffle(input.begin(), input.end(), rng);
    auto result = stats.getMostFrequentValues(input, 0, -1);
    EXPECT_EQ(0, result.size());
}

TEST(StatisticsTest, AllCleanFrequentValuesTest)
{
    Statistics stats;
    std::default_random_engine rng;
    std::vector<std::string> input {"a", "b", "b", "b", "c", "c", "d", "d", "d", "d", "d"};
    std::shuffle(input.begin(), input.end(), rng);
    auto result = stats.getMostFrequentValuesClean(input, 4, -1);
    EXPECT_EQ(std::vector<std::string>({"d", "b", "c", "a"}), result);
}

TEST(StatisticsTest, FrequentValuesAccuracyTest)
{
    Statistics stats;
    std::default_random_engine rng;
    std::vector<double_t> input {0.999, 1, 1, 1.001, 1.999, 2, 2.001, 2.999, 3};
    std::shuffle(input.begin(), input.end(), rng);
    auto result = stats.getMostFrequentValues(input, 4, 1.0);
    for (const auto& [value, frequency] : result)
    {
        if (std::abs(1.0 - value) < 0.5)
        {
            EXPECT_EQ(4, frequency);
        }
        else if (std::abs(2.0 - value) < 0.5)
        {
            EXPECT_EQ(3, frequency);
        }
        else if (std::abs(3.0 - value) < 0.5)
        {
            EXPECT_EQ(2, frequency);
        }
    }
}

TEST(StatisticsTest, QuantileIntervalNoneTest)
{
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    numVariable->setAccuracy(1.0);
    Statistics stats;
    std::default_random_engine rng;
    std::vector<double_t> input {1, 1, 2, 4, 5, 6, 6, 7, 7, 7, 8, 9, 9, 9, 9, 9};
    std::shuffle(input.begin(), input.end(), rng);
    auto result = stats.getQuantileIntervals(input, numVariable, 0);
    EXPECT_EQ(0, result.size());
}

TEST(StatisticsTest, QuantileIntervalOneTest)
{
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    numVariable->setAccuracy(1.0);
    Statistics stats;
    std::default_random_engine rng;
    std::vector<double_t> input {1, 1, 2, 4, 5, 6, 6, 7, 7, 7, 8, 9, 9, 9, 9, 9};
    std::shuffle(input.begin(), input.end(), rng);
    auto result = stats.getQuantileIntervals(input, numVariable, 1);
    EXPECT_EQ(Statistics::IntervalsType({{{1, 9}, 16}}), result);
}

TEST(StatisticsTest, QuantileIntervalAllTest)
{
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    numVariable->setAccuracy(1.0);
    Statistics stats;
    std::default_random_engine rng;
    std::vector<double_t> input {1, 1, 2, 4, 5, 6, 6, 7, 7, 7, 8, 9, 9, 9, 9, 9};
    std::shuffle(input.begin(), input.end(), rng);
    auto result = stats.getQuantileIntervals(input, numVariable, 3);
    EXPECT_EQ(Statistics::IntervalsType({{{1, 9}, 16}, {{5, 9}, 12}, {{6, 7}, 5}}), result);
}

TEST(StatisticsTest, QuantileIntervalTooManyTest)
{
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    numVariable->setAccuracy(1.0);
    Statistics stats;
    std::default_random_engine rng;
    std::vector<double_t> input {1, 1, 2, 4, 5, 6, 6, 7, 7, 7, 8, 9, 9, 9, 9, 9};
    std::shuffle(input.begin(), input.end(), rng);
    auto result = stats.getQuantileIntervals(input, numVariable, 4);
    EXPECT_EQ(Statistics::IntervalsType({{{1, 9}, 16}, {{5, 9}, 12}, {{6, 7}, 5}}), result);
}

TEST(StatisticsTest, QuantileIntervalUnevenTest)
{
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    numVariable->setAccuracy(1.0);
    Statistics stats;
    std::default_random_engine rng;
    std::vector<double_t> input {1, 2, 3, 4, 5, 6, 7, 8, 9};
    std::shuffle(input.begin(), input.end(), rng);
    auto result = stats.getQuantileIntervals(input, numVariable, 4);
    EXPECT_EQ(Statistics::IntervalsType({{{1, 9}, 9}, {{3, 7}, 5}, {{4, 6}, 3}}), result);
}

TEST(StatisticsTest, QuantileDivisionsNoneTest)
{
    Statistics stats;
    std::default_random_engine rng;
    std::vector<double_t> input {1, 1, 2, 4, 5, 6, 6, 7, 7, 7, 8, 9, 9, 9, 9, 9};
    std::shuffle(input.begin(), input.end(), rng);
    auto result = stats.getQuantileDivisions(input, 1.0, 0);
    EXPECT_EQ(0, result.size());
}

TEST(StatisticsTest, QuantileDivisionsOneTest)
{
    Statistics stats;
    std::default_random_engine rng;
    std::vector<double_t> input {1, 1, 2, 4, 5, 6, 6, 7, 7, 7, 8, 9, 9, 9, 9, 9};
    std::shuffle(input.begin(), input.end(), rng);
    auto result = stats.getQuantileDivisions(input, 1.0, 1);
    EXPECT_EQ(std::vector<double_t>({7}), result);
}

TEST(StatisticsTest, QuantileDivisionsSomeTest)
{
    Statistics stats;
    std::default_random_engine rng;
    std::vector<double_t> input {1, 1, 2, 4, 5, 6, 6, 7, 7, 7, 8, 9, 9, 9, 9, 9};
    std::shuffle(input.begin(), input.end(), rng);
    auto result = stats.getQuantileDivisions(input, 1.0, 4);
    EXPECT_EQ(std::vector<double_t>({4, 7, 8, 9}), result);
}

TEST(StatisticsTest, QuantileDivisionsTooManyTest)
{
    Statistics stats;
    std::default_random_engine rng;
    std::vector<double_t> input {1, 1, 2, 4, 5, 6, 6, 7, 7, 7, 8, 9, 9, 9, 9, 9};
    std::shuffle(input.begin(), input.end(), rng);
    auto result = stats.getQuantileDivisions(input, 1.0, 16);
    EXPECT_EQ(std::vector<double_t>({1, 2, 4, 5, 6, 7, 8, 9}), result);
}

TEST(StatisticsTest, QuantileDivisionsPerfectSplitTest)
{
    Statistics stats;
    std::default_random_engine rng;
    std::vector<double_t> input {1, 2, 3, 4, 5, 6, 7};
    std::shuffle(input.begin(), input.end(), rng);
    auto result = stats.getQuantileDivisions(input, 1.0, 3);
    EXPECT_EQ(std::vector<double_t>({2, 4, 6}), result);
}
