#include <gtest/gtest.h>
#include "generators.hpp"
#include "../classes/search/treeconditiongenerator.hpp"
#include "../classes/datamodel/updaterules/valuerule.hpp"
#include "../classes/datamodel/updaterules/valuesetrule.hpp"
#include "../classes/datamodel/conditions/valuecondition.hpp"
#include "../classes/datamodel/conditions/andoperator.hpp"
#include "../classes/datamodel/conditions/lowerboundcondition.hpp"
#include "../classes/datamodel/conditions/transitioncondition.hpp"

TEST(TreeConditionGeneratorTest, GenerateConditionsCategoricalSplitTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<CategoricalVariable> otherVariable = std::make_shared<CategoricalVariable>("other",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");

    auto log = Generators::generateEventLogWithNumericalUpdateRule(catVariable, otherVariable, numVariable);
    std::unique_ptr<UpdateRule> updateRule = std::make_unique<ValueRule>(numVariable, 42);

    TreeConditionGenerator conditionGenerator;
    auto conditions = conditionGenerator.generateConditions(log.get(), updateRule.get(), 3);

    ASSERT_EQ(1, conditions.size());
    ASSERT_EQ(1, conditions[0]->getElements().size());
    EXPECT_EQ("vowel", conditions[0]->getElements()[0]->getVariable()->getName());
    auto catValueCondition = dynamic_cast<const ValueCondition<std::string>*>(
        conditions[0]->getElements()[0].get());
    ASSERT_TRUE(catValueCondition != nullptr);
    EXPECT_EQ("a", catValueCondition->getValue());
}

TEST(TreeConditionGeneratorTest, GenerateConditionsNumericalSplitTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");

    auto log = Generators::generateEventLogWithCategoricalUpdateRule(catVariable, numVariable);
    std::unique_ptr<UpdateRule> updateRule = std::make_unique<ValueSetRule>(
        catVariable, ValueSetRule::SetType{"a"});

    TreeConditionGenerator conditionGenerator;
    auto conditions = conditionGenerator.generateConditions(log.get(), updateRule.get());

    ASSERT_EQ(1, conditions.size());
    ASSERT_EQ(1, conditions[0]->getElements().size());
    EXPECT_EQ("num", conditions[0]->getElements()[0]->getVariable()->getName());
    auto boundCondition = dynamic_cast<const LowerBoundCondition*>(
        conditions[0]->getElements()[0].get());
    ASSERT_TRUE(boundCondition != nullptr);
    EXPECT_EQ(10.0, boundCondition->getValue());
}

TEST(TreeConditionGeneratorTest, GenerateConditionsTransitionalSplitTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");

    auto log = Generators::generateEventLogWithTransitionalUpdateRule(catVariable, numVariable);
    std::unique_ptr<UpdateRule> updateRule = std::make_unique<ValueRule>(numVariable, 42);

    TreeConditionGenerator conditionGenerator;
    auto conditions = conditionGenerator.generateConditions(log.get(), updateRule.get());

    ASSERT_EQ(1, conditions.size());
    ASSERT_EQ(1, conditions[0]->getElements().size());
    EXPECT_EQ("vowel", conditions[0]->getElements()[0]->getVariable()->getName());
    auto transitionCondition = dynamic_cast<const TransitionCondition<std::string>*>(
        conditions[0]->getElements()[0].get());
    ASSERT_TRUE(transitionCondition != nullptr);
    EXPECT_EQ("o", transitionCondition->getBefore());
    EXPECT_EQ("u", transitionCondition->getAfter());
}
