#include <gtest/gtest.h>
#include "generators.hpp"
#include "../classes/scoring/updaterulematcher.hpp"
#include "../classes/datamodel/updaterules/valuerangerule.hpp"
#include "../classes/datamodel/updaterules/changerangerule.hpp"
#include "../classes/datamodel/updaterules/valuerule.hpp"
#include "../classes/datamodel/updaterules/factorrule.hpp"
#include "../classes/datamodel/updaterules/differencerule.hpp"

TEST(UpdateRuleMatcherTest, InValueSetTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::unique_ptr<Trace> trace = Generators::generateComplexTrace(catVariable);
    std::unique_ptr<ValueSetRule> rule = std::make_unique<ValueSetRule>(catVariable, std::set<std::string>{"a", "e"});

    UpdateRuleMatcher matcher;
    EXPECT_TRUE(matcher.matches(rule.get(), trace.get(), 1));
}

TEST(UpdateRuleMatcherTest, NotInValueSetTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::unique_ptr<Trace> trace = Generators::generateComplexTrace(catVariable);
    std::unique_ptr<ValueSetRule> rule = std::make_unique<ValueSetRule>(catVariable, std::set<std::string>{"a", "e"});

    UpdateRuleMatcher matcher;
    EXPECT_FALSE(matcher.matches(rule.get(), trace.get(), 0));
}

TEST(UpdateRuleMatcherTest, BelowMinValueRangeTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<CategoricalVariable> otherVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<Trace> trace = Generators::generateComplexTrace(catVariable, otherVariable, numVariable);
    std::unique_ptr<ValueRangeRule> rule = std::make_unique<ValueRangeRule>(numVariable, 5.0, 6.0);

    UpdateRuleMatcher matcher;
    EXPECT_FALSE(matcher.matches(rule.get(), trace.get(), 0));
}

TEST(UpdateRuleMatcherTest, AtMinValueRangeTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<CategoricalVariable> otherVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<Trace> trace = Generators::generateComplexTrace(catVariable, otherVariable, numVariable);
    std::unique_ptr<ValueRangeRule> rule = std::make_unique<ValueRangeRule>(numVariable, 4.0, 5.0);

    UpdateRuleMatcher matcher;
    EXPECT_TRUE(matcher.matches(rule.get(), trace.get(), 0));
}

TEST(UpdateRuleMatcherTest, AboveMinValueRangeTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<CategoricalVariable> otherVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<Trace> trace = Generators::generateComplexTrace(catVariable, otherVariable, numVariable);
    std::unique_ptr<ValueRangeRule> rule = std::make_unique<ValueRangeRule>(numVariable, 3.9, 5.0);

    UpdateRuleMatcher matcher;
    EXPECT_TRUE(matcher.matches(rule.get(), trace.get(), 0));
}

TEST(UpdateRuleMatcherTest, BelowMaxValueRangeTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<CategoricalVariable> otherVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<Trace> trace = Generators::generateComplexTrace(catVariable, otherVariable, numVariable);
    std::unique_ptr<ValueRangeRule> rule = std::make_unique<ValueRangeRule>(numVariable, 3.0, 4.1);

    UpdateRuleMatcher matcher;
    EXPECT_TRUE(matcher.matches(rule.get(), trace.get(), 0));
}

TEST(UpdateRuleMatcherTest, AtMaxValueRangeTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<CategoricalVariable> otherVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<Trace> trace = Generators::generateComplexTrace(catVariable, otherVariable, numVariable);
    std::unique_ptr<ValueRangeRule> rule = std::make_unique<ValueRangeRule>(numVariable, 3.0, 4.0);

    UpdateRuleMatcher matcher;
    EXPECT_TRUE(matcher.matches(rule.get(), trace.get(), 0));
}

TEST(UpdateRuleMatcherTest, AboveMaxValueRangeTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<CategoricalVariable> otherVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<Trace> trace = Generators::generateComplexTrace(catVariable, otherVariable, numVariable);
    std::unique_ptr<ValueRangeRule> rule = std::make_unique<ValueRangeRule>(numVariable, 2.0, 3.0);

    UpdateRuleMatcher matcher;
    EXPECT_FALSE(matcher.matches(rule.get(), trace.get(), 0));
}

TEST(UpdateRuleMatcherTest, BelowMinChangeRangeTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<CategoricalVariable> otherVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<Trace> trace = Generators::generateComplexTrace(catVariable, otherVariable, numVariable);
    std::unique_ptr<ChangeRangeRule> rule = std::make_unique<ChangeRangeRule>(numVariable, 10.0, 12.0);

    UpdateRuleMatcher matcher;
    EXPECT_FALSE(matcher.matches(rule.get(), trace.get(), 5));
}

TEST(UpdateRuleMatcherTest, AtMinChangeRangeTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<CategoricalVariable> otherVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<Trace> trace = Generators::generateComplexTrace(catVariable, otherVariable, numVariable);
    std::unique_ptr<ChangeRangeRule> rule = std::make_unique<ChangeRangeRule>(numVariable, 9.0, 10.0);

    UpdateRuleMatcher matcher;
    EXPECT_TRUE(matcher.matches(rule.get(), trace.get(), 5));
}

TEST(UpdateRuleMatcherTest, AboveMinChangeRangeTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<CategoricalVariable> otherVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<Trace> trace = Generators::generateComplexTrace(catVariable, otherVariable, numVariable);
    std::unique_ptr<ChangeRangeRule> rule = std::make_unique<ChangeRangeRule>(numVariable, 8.9, 10.0);

    UpdateRuleMatcher matcher;
    EXPECT_TRUE(matcher.matches(rule.get(), trace.get(), 5));
}

TEST(UpdateRuleMatcherTest, BelowMaxChangeRangeTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<CategoricalVariable> otherVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<Trace> trace = Generators::generateComplexTrace(catVariable, otherVariable, numVariable);
    std::unique_ptr<ChangeRangeRule> rule = std::make_unique<ChangeRangeRule>(numVariable, 8.0, 10.0);

    UpdateRuleMatcher matcher;
    EXPECT_TRUE(matcher.matches(rule.get(), trace.get(), 5));
}

TEST(UpdateRuleMatcherTest, AtMaxChangeRangeTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<CategoricalVariable> otherVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<Trace> trace = Generators::generateComplexTrace(catVariable, otherVariable, numVariable);
    std::unique_ptr<ChangeRangeRule> rule = std::make_unique<ChangeRangeRule>(numVariable, 8.0, 9.0);

    UpdateRuleMatcher matcher;
    EXPECT_TRUE(matcher.matches(rule.get(), trace.get(), 5));
}

TEST(UpdateRuleMatcherTest, AboveMaxChangeRangeTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<CategoricalVariable> otherVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<Trace> trace = Generators::generateComplexTrace(catVariable, otherVariable, numVariable);
    std::unique_ptr<ChangeRangeRule> rule = std::make_unique<ChangeRangeRule>(numVariable, 6.0, 8.0);

    UpdateRuleMatcher matcher;
    EXPECT_FALSE(matcher.matches(rule.get(), trace.get(), 5));
}

TEST(UpdateRuleMatcherTest, ChangeRangeAtFirstIndexTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<CategoricalVariable> otherVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<Trace> trace = Generators::generateComplexTrace(catVariable, otherVariable, numVariable);
    std::unique_ptr<ChangeRangeRule> rule = std::make_unique<ChangeRangeRule>(numVariable, 8.0, 10.0);

    UpdateRuleMatcher matcher;
    EXPECT_FALSE(matcher.matches(rule.get(), trace.get(), 0));
}

TEST(UpdateRuleMatcherTest, ValueRuleMatchTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<CategoricalVariable> otherVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<Trace> trace = Generators::generateComplexTrace(catVariable, otherVariable, numVariable);
    std::unique_ptr<AbstractValueRule> rule = std::make_unique<ValueRule>(numVariable, 4.0);

    UpdateRuleMatcher matcher;
    EXPECT_TRUE(matcher.matches(rule.get(), trace.get(), 0));
}

TEST(UpdateRuleMatcherTest, ValueRuleToleranceTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<CategoricalVariable> otherVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<Trace> trace = Generators::generateComplexTrace(catVariable, otherVariable, numVariable);
    std::unique_ptr<AbstractValueRule> rule = std::make_unique<ValueRule>(numVariable, 4.499);

    UpdateRuleMatcher matcher;
    EXPECT_TRUE(matcher.matches(rule.get(), trace.get(), 0));
}

TEST(UpdateRuleMatcherTest, ValueRuleMismatchTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<CategoricalVariable> otherVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<Trace> trace = Generators::generateComplexTrace(catVariable, otherVariable, numVariable);
    std::unique_ptr<AbstractValueRule> rule = std::make_unique<ValueRule>(numVariable, 5.0);

    UpdateRuleMatcher matcher;
    EXPECT_FALSE(matcher.matches(rule.get(), trace.get(), 0));
}

TEST(UpdateRuleMatcherTest, DifferenceRuleMatchTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<CategoricalVariable> otherVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<Trace> trace = Generators::generateComplexTrace(catVariable, otherVariable, numVariable);
    std::unique_ptr<AbstractValueRule> rule = std::make_unique<DifferenceRule>(numVariable, 0.0);

    UpdateRuleMatcher matcher;
    EXPECT_TRUE(matcher.matches(rule.get(), trace.get(), 1));
}

TEST(UpdateRuleMatcherTest, DifferenceRuleToleranceTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<CategoricalVariable> otherVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<Trace> trace = Generators::generateComplexTrace(catVariable, otherVariable, numVariable);
    std::unique_ptr<AbstractValueRule> rule = std::make_unique<DifferenceRule>(numVariable, -0.499);

    UpdateRuleMatcher matcher;
    EXPECT_TRUE(matcher.matches(rule.get(), trace.get(), 1));
}

TEST(UpdateRuleMatcherTest, DifferenceRuleMismatchTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<CategoricalVariable> otherVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<Trace> trace = Generators::generateComplexTrace(catVariable, otherVariable, numVariable);
    std::unique_ptr<AbstractValueRule> rule = std::make_unique<DifferenceRule>(numVariable, -1.0);

    UpdateRuleMatcher matcher;
    EXPECT_FALSE(matcher.matches(rule.get(), trace.get(), 1));
}

TEST(UpdateRuleMatcherTest, FactorRuleMatchTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<CategoricalVariable> otherVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<Trace> trace = Generators::generateComplexTrace(catVariable, otherVariable, numVariable);
    std::unique_ptr<AbstractValueRule> rule = std::make_unique<FactorRule>(numVariable, 1.0);

    UpdateRuleMatcher matcher;
    EXPECT_TRUE(matcher.matches(rule.get(), trace.get(), 1));
}

TEST(UpdateRuleMatcherTest, FactorRuleToleranceTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<CategoricalVariable> otherVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<Trace> trace = Generators::generateComplexTrace(catVariable, otherVariable, numVariable);
    std::unique_ptr<AbstractValueRule> rule = std::make_unique<FactorRule>(numVariable, 1.12499);

    UpdateRuleMatcher matcher;
    EXPECT_TRUE(matcher.matches(rule.get(), trace.get(), 1));
}

TEST(UpdateRuleMatcherTest, FactorRuleMismatchTest)
{
    std::shared_ptr<CategoricalVariable> catVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<CategoricalVariable> otherVariable = std::make_shared<CategoricalVariable>("vowel",
        std::set<std::string>{"a", "e", "i", "o", "u"});
    std::shared_ptr<NumericalVariable> numVariable = std::make_shared<NumericalVariable>("num");
    std::unique_ptr<Trace> trace = Generators::generateComplexTrace(catVariable, otherVariable, numVariable);
    std::unique_ptr<AbstractValueRule> rule = std::make_unique<FactorRule>(numVariable, 1.25);

    UpdateRuleMatcher matcher;
    EXPECT_FALSE(matcher.matches(rule.get(), trace.get(), 1));
}
