# fix working directory
import os
abspath = os.path.abspath(__file__)
dname = os.path.dirname(abspath)
os.chdir(dname)

from rulelist.rulelist import RuleList
import re
import sys
sys.path.append("../..")
from experiments.dataframetocapnp import dataframeFromMessage
import capnp
capnp.remove_import_hook()
interface_capnp = capnp.load("../../cpp-tool/capnproto/interface.capnp", imports=["/usr/local/include"])

class ServerImplementation(interface_capnp.SimpleProcessDataUpdates.Server):
    def findModel(self, log, targetVariables, **kwargs):
        # get dataframe and remove case ID
        dataframe = dataframeFromMessage(log)
        caseIdColumn = "case:concept:name"
        dataframe.drop(caseIdColumn, axis=1, inplace=True)

        # no given target variables => all variables are considered as targets
        if len(targetVariables) == 0:
            targetVariables = sorted(dataframe.columns)
        else:
            targetVariables = sorted(targetVariables)
            
        discoveredRules = []
        # run SSD++ for all possible target variables
        for targetName in targetVariables:
            # prepare inputs
            y = dataframe[targetName]
            X = dataframe.drop(targetName, axis="columns")

            # run SSD++
            target_model = self.getTargetModelForVariableName(log, targetName)
            subgroupModel = RuleList(task="discovery", target_model=target_model)
            subgroupModel.fit(X, y)
            
            # convert SSD++ result to data modification rules and add them to the result
            discoveredRules.extend(self.ssdppModelToRules(subgroupModel, targetName))

        return interface_capnp.RuleModel.new_message(
            rules=discoveredRules, variables=list(log.variables))

    def getTargetModelForVariableName(self, log, variableName):
        # search name in variables
        for variable in log.variables:
            if variable.name == variableName:
                # convert variable types to target model names
                if variable.which() == "numerical":
                    return "gaussian"
                elif variable.which() == "categorical":
                    return "categorical"
                else:
                    raise RuntimeError("Variable is neither categorical nor numerical")

    def createBoundCondition(self, variableName, bound, conditionType):
        # create atomic condition
        element = interface_capnp.ConditionElement.new_message()
        element.init("atomicCondition")
        element.atomicCondition.variable = variableName
        # initialize bound condition of desired type and set bound
        if conditionType == "lower":
            element.atomicCondition.init("lowerBoundCondition")
            element.atomicCondition.lowerBoundCondition.value = bound
        elif conditionType == "upper":
            element.atomicCondition.init("upperBoundCondition")
            element.atomicCondition.upperBoundCondition.value = bound
        else:
            raise RuntimeError("Neither lower nor upper bound condition type requested")
        # return the created condition
        return element

    def addAndOperatorIfNecessary(self, elements):
        # only add operator if there are at least 2 condition elements
        if len(elements) > 1:
            # create and add AND operator
            element = interface_capnp.ConditionElement.new_message()
            element.andOperator = None
            elements.append(element)

    def ssdppModelToRules(self, subgroupModel, targetName):
        rules = []
        # create a modification rule for each discovered subgroup
        for subgroup in subgroupModel._rulelist.subgroups:
            # construct condition
            elements = []
            # construct a condition element for each pattern/condition of the subgroup
            for pattern in subgroup.pattern:
                # match a pattern that specifies an interval for a numeric variable
                bothBounds = re.match(r"(\d+\.?\d*) <= (.+) < (\d+\.?\d*)", pattern.description)
                # match a pattern that specifies a single bound for a numeric variable
                oneBound = re.match(r"^([^<>=]+) (<|>=) (\d+\.?\d*)", pattern.description)
                # match a pattern that specifies a single value for a categorical variable
                # (SSD++ does not model single values for numeric variables)
                equals = re.match(r"(.+) = (.+)", pattern.description)

                if bothBounds is not None:
                    # create two bound conditions for the interval and connect them with AND operators
                    lowCondition = self.createBoundCondition(
                        bothBounds.group(2), float(bothBounds.group(1)), "lower")
                    highCondition = self.createBoundCondition(
                        bothBounds.group(2), float(bothBounds.group(3)), "upper")
                    elements.append(lowCondition)
                    # SSD++ only has AND connectors
                    self.addAndOperatorIfNecessary(elements)
                    elements.append(highCondition)
                    # SSD++ only has AND connectors
                    self.addAndOperatorIfNecessary(elements)
                elif oneBound is not None:
                    # create the single bound condition that was discovered
                    conditionType = ""
                    if oneBound.group(2) == "<":
                        conditionType = "upper"
                    else:
                        conditionType = "lower"
                    
                    condition = self.createBoundCondition(
                        oneBound.group(1), float(oneBound.group(3)), conditionType)
                    elements.append(condition)
                    # SSD++ only has AND connectors
                    self.addAndOperatorIfNecessary(elements)
                elif equals is not None:
                    # create the condition that selects a categorical value
                    element = interface_capnp.ConditionElement.new_message()
                    element.init("atomicCondition")
                    element.atomicCondition.variable = equals.group(1)
                    element.atomicCondition.init("categoricalValueCondition")
                    element.atomicCondition.categoricalValueCondition.value = equals.group(2)
                    elements.append(element)
                    # SSD++ only has AND connectors
                    self.addAndOperatorIfNecessary(elements)

            # insert condition into rule
            rule = interface_capnp.ModificationRule.new_message()
            rule.condition.init("elements", len(elements))
            for i, element in enumerate(elements):
                rule.condition.elements[i] = element
            
            # construct update rule
            rule.updateRule.variable = targetName
            if subgroupModel.target_model == "gaussian":
                # find the numerical value predicted in the subgroup
                prediction = subgroup.statistics.mean[0].item()

                # create numerical value rule
                rule.updateRule.init("valueRule")
                rule.updateRule.valueRule.constant = prediction
            elif subgroupModel.target_model == "categorical":
                # find the (most frequent) categorical value in the subgroup
                categoryUsages = subgroup.statistics.usage_per_class
                prediction = max(categoryUsages[targetName], 
                    key=lambda x: categoryUsages[targetName][x])
                
                # create categorical value set rule with one element in the set
                rule.updateRule.init("valueSetRule")
                rule.updateRule.valueSetRule.init("set", 1)
                rule.updateRule.valueSetRule.set[0] = prediction
            else:
                raise RuntimeError("Model is neither gaussian nor categorical")
            
            # add rule to result
            rules.append(rule)
        return rules
        
if __name__ == "__main__":
    # start server
    server = capnp.TwoPartyServer("*:21652", bootstrap=ServerImplementation())
    server.run_forever()
