import pandas as pd
import capnp
from tqdm import tqdm
import os
abspath = os.path.abspath(__file__)
directoryName = os.path.dirname(abspath)
capnp.remove_import_hook()
interface_capnp = capnp.load(directoryName + "/../cpp-tool/capnproto/interface.capnp",
                             imports=["/usr/local/include"])

def determineColumnDatatype(column):
    if column.dtype.kind in "iuf": # integer, unsigned integer or float
        return "numerical"
    elif column.dtype.kind in "Ob": # object (aka string) or bool
        return "categorical"
    else:
        raise TypeError("Data type of column not supported")

def initializeVariableList(eventLogMessage, dataframe, columnsToEncode, disableProgressBar=False):
    variables = eventLogMessage.init("variables", len(columnsToEncode))
    # each dataframe column is a variable in the event log
    for i in tqdm(range(len(columnsToEncode)), desc="Constructing variables", 
                  disable=disableProgressBar):
        # set variable name and determine its type
        variables[i].name = columnsToEncode[i]
        columnType = determineColumnDatatype(dataframe[columnsToEncode[i]])
        if columnType == "numerical":
            # numerical variables don't need further information
            variables[i].numerical = None
        elif columnType == "categorical":
            variables[i].init("categorical")
            # determine and set the domain (possible values) for a categorical variable
            domainInData = list(set(dataframe[columnsToEncode[i]].dropna()))
            domainInMessage = variables[i].categorical.init("domain", len(domainInData))
            for j in range(len(domainInData)):
                domainInMessage[j] = str(domainInData[j])

def initializeTraces(eventLogMessage, dataframe, columnsToEncode,
        caseIdColumn="case:concept:name", fillMissingValues=False, disableProgressBar=False):
    caseIds = list(set(dataframe[caseIdColumn]))
    traces = eventLogMessage.init("traces", len(caseIds))
    # each trace is identified by its case ID
    for i in tqdm(range(len(caseIds)), desc="Constructing traces", disable=disableProgressBar):
        # filter data for the current case ID
        mask = dataframe[caseIdColumn] == caseIds[i]
        # (optionally) fill missing values within a trace
        if fillMissingValues:
            dataframe.loc[mask] = dataframe.loc[mask].fillna(method="ffill")
        caseData = dataframe.loc[mask]
        # set case ID of trace
        traces[i].caseId = caseIds[i]
        # fill the columns of the trace with data
        traces[i].init("columns", len(columnsToEncode))
        for j in range(len(columnsToEncode)):
            initializeColumn(traces[i].columns[j], caseData, columnsToEncode[j])

def initializeColumn(targetColumn, caseData, columnName):
    # set variable name and determine its type
    targetColumn.variableName = columnName
    columnType = determineColumnDatatype(caseData[columnName])
    if columnType == "numerical":
        # fill in values of numerical column
        targetColumn.init("numerical")
        targetColumn.numerical.init("entries", len(caseData))
        columnAsList = list(caseData[columnName])
        for i in range(len(caseData)):
            targetColumn.numerical.entries[i] = columnAsList[i]
    elif columnType == "categorical":
        # fill in values of categorical column
        targetColumn.init("categorical")
        targetColumn.categorical.init("entries", len(caseData))
        columnAsList = list(caseData[columnName])
        for i in range(len(caseData)):
            # convert to string since the column may be boolean
            targetColumn.categorical.entries[i] = str(columnAsList[i])

def dataframeToMessage(dataframe, variables=None, caseIdColumn="case:concept:name",
                       fillMissingValues=False, disableProgressBar=False):
    # encode all except for the case ID column
    columnsToEncode = list(dataframe.columns)
    columnsToEncode.remove(caseIdColumn)

    # create and populate capnp message
    eventLogMessage = interface_capnp.EventLog.new_message()
    if variables == None:
        initializeVariableList(eventLogMessage, dataframe, columnsToEncode, 
                               disableProgressBar=disableProgressBar)
    else:
        eventLogMessage.variables = variables
    initializeTraces(eventLogMessage, dataframe, columnsToEncode, caseIdColumn, fillMissingValues, 
                     disableProgressBar=disableProgressBar)

    return eventLogMessage

def dataframeFromMessage(eventLogMessage, disableProgressBar=False):
    caseIdColumn = "case:concept:name"
    columnNames = [caseIdColumn]
    for variable in eventLogMessage.variables:
        columnNames.append(variable.name)
    
    eventLogDictionary = {x: [] for x in columnNames}

    for trace in tqdm(eventLogMessage.traces, disable=disableProgressBar):
        traceLength = len(getEntries(trace.columns[0]))
        eventLogDictionary[caseIdColumn].extend([trace.caseId] * traceLength)
        for column in trace.columns:
            if traceLength != len(getEntries(column)):
                raise IndexError("Columns in trace do not have equal size")
            
            entries = getEntries(column)
            fixedEntries = [convertStringEntry(x) for x in entries]
            eventLogDictionary[column.variableName].extend(fixedEntries)
    
    # ensure the dataframe is sorted by case ID
    result = pd.DataFrame(eventLogDictionary)
    result.sort_values("case:concept:name", inplace=True, kind="stable")
    result.reset_index(drop=True, inplace=True)

    return result

def getEntries(traceColumn):
    if traceColumn.which() == "numerical":
        return traceColumn.numerical.entries
    elif traceColumn.which() == "categorical":
        return traceColumn.categorical.entries

def convertStringEntry(entry):
    if entry == "True":
        return True
    elif entry == "False":
        return False
    elif entry in ["nan", "NaN"]:
        return float("nan")
    else:
        return entry
