import re
import os
import pandas as pd
from .dataframetocapnp import dataframeFromMessage
from .dataframetocapnp import dataframeToMessage
from sklearn.metrics import f1_score
from sklearn.metrics import mean_absolute_error
from sklearn.metrics import balanced_accuracy_score
from sklearn.metrics import root_mean_squared_error
import matplotlib.pyplot as plt
import numpy as np
from tqdm import tqdm
from .decisiontree import DecisionTreeAdapter
import joblib
import random

class Evaluator:

    def __init__(self, interface_capnp, processDataUpdatesCpp) -> None:
        self.interface_capnp = interface_capnp
        self.processDataUpdatesCpp = processDataUpdatesCpp
        self.allMetrics = ["strictF1", "looseF1", "predictionF1", "complexity", "valueSetF1", 
                           "predictionMAE", "valueSetBalancedAccuracy", "predictionRMSE"]
        self.descriptionLookup = {
            "strictF1": "F1 score of discovered rules",
            "looseF1": "F1 score of discovered rules with swaps",
            "predictionF1": "F1 score of individual categorical predictions",
            "complexity": "Model complexity",
            "valueSetF1": "F1 score of categorical set predictions", 
            "predictionMAE": "Mean absolute error of expected value predictions",
            "valueSetBalancedAccuracy": "Balanced accuracy of categorical set predictions",
            "predictionRMSE": "Root mean squared error of expected value predictions",
            "noiseProportion": "Proportion of randomly swapped values per attribute",
            "logLength": "Number of events",
            "conditionCount": "Number of conditions per type and non-target variable",
            "runtime": "Execution time [sec]"
        }

    def swapPossible(self, ruleString):
        return re.match(r"IF (\w+) = (\w+) THEN (\w+) IN {(\w+)}", ruleString) is not None

    def performSwap(self, ruleString):
        match = re.match(r"IF (\w+) = (\w+) THEN (\w+) IN {(\w+)}", ruleString)
        return "IF " + match.group(3) + " = " + match.group(4) + " THEN "\
            + match.group(1) + " IN {" + match.group(2) + "}"

    def computeRelativeMatches(self, baseRuleString, matchingRuleString, allowSwaps=False, numericDigits=1):
        relativeMatches = 0
        # for all base rules (lines in the string) try to find a matching rule
        for baseRule in baseRuleString.splitlines():
            baseRule = re.sub(r"\d+\.\d+", lambda x: str(round(float(x.group(0)), numericDigits)), baseRule)
            # search through matching rules (lines in the string) for the base rule
            for matchingRule in matchingRuleString.splitlines():
                matchingRule = re.sub(r"\d+\.\d+", lambda x: str(round(float(x.group(0)), numericDigits)), matchingRule)
                # if allowed, swap condition and update rule of modification rule string
                swappedRule = ""
                if allowSwaps and self.swapPossible(matchingRule):
                    swappedRule = self.performSwap(matchingRule)
                # count if base rule matches and abort search (inner loop)
                if baseRule == matchingRule or baseRule == swappedRule:
                    relativeMatches += 1
                    break
        
        # convert absolute count of matches into relative proportion
        if len(baseRuleString.splitlines()) > 0:
            relativeMatches /= len(baseRuleString.splitlines())
        
        return relativeMatches

    def computeF1Score(self, trueRuleModel, estimatedRuleModel, allowSwaps=False):
        # convert models to string
        trueModelString = self.processDataUpdatesCpp.modelToString(trueRuleModel).wait().string
        estimatedModelString = self.processDataUpdatesCpp.modelToString(estimatedRuleModel).wait().string

        # compute building blocks of F1 score
        precision = self.computeRelativeMatches(estimatedModelString, trueModelString, allowSwaps)
        recall = self.computeRelativeMatches(trueModelString, estimatedModelString, allowSwaps)

        # compute F1 score
        if precision + recall > 0:
            return 2 * (precision * recall) / (precision + recall)
        else:
            return 0.0

    def computeComplexity(self, model):
        complexity = 0

        for rule in model.rules:
            # count atomic conditions in condition
            for element in rule.condition.elements:
                if element.which() == "atomicCondition":
                    complexity += 1
            # cout elements of update rule
            if rule.updateRule.which() == "valueSetRule":
                complexity += len(rule.updateRule.valueSetRule.set)
            else:
                complexity += 1

        return complexity
    
    def computeTreeComplexity(self, treeModel):
        complexity = 0

        for _, decisionTree in treeModel.items():
            complexity += decisionTree.tree_.node_count

        return complexity

    def computePredictions(self, model, eventLogMessage, trainingLogMessage = None,
                           predictMeans=True, modelIsOrderedList=False):
        # get predictions from a decision tree model
        if type(model) == dict:
            predictions = self.getPredictionsFromTree(eventLogMessage, model)
            predictionDataframe = pd.DataFrame(predictions)
            return dataframeToMessage(predictionDataframe,
                                      list(eventLogMessage.as_builder().variables),
                                      disableProgressBar=True)

        if trainingLogMessage is None: # we use uniform distributions when predicting without a rule
            return self.processDataUpdatesCpp.predictEventLog(
                eventLogMessage, model, predictMeans=predictMeans,
                modelIsOrderedList=modelIsOrderedList).wait().log
        else: # we use the training log's distributions when predicting without a rule
            return self.processDataUpdatesCpp.predictRealEventLog(
                trainingLogMessage, eventLogMessage, model, predictMeans=predictMeans,
                modelIsOrderedList=modelIsOrderedList).wait().log

    def computePredictionF1(self, eventLogMessage, predictions):
        # get true and predicted data as dataframe
        logDataframe = dataframeFromMessage(eventLogMessage, disableProgressBar=True)
        predictionsDataframe = dataframeFromMessage(predictions, disableProgressBar=True)
        logDataframe.sort_values("case:concept:name", inplace=True)
        predictionsDataframe.sort_values("case:concept:name", inplace=True)

        # average over all variable's F1 scores
        f1scores = []
        for column in predictionsDataframe.columns:
            # only compute F1 score on categorical columns
            if getVariableType(column, eventLogMessage.variables) != "categorical":
                continue
            score = self.computeSinglePredictionMetric(logDataframe[column],
                                                       predictionsDataframe[column],
                                                       f1_score, average="macro")
            f1scores.append(score)

        if len(f1scores) == 0:
            return float("nan")

        return sum(f1scores) / len(f1scores)
    
    def computeSingleClassMetric(self, trueDataframe, predictedDataframe, column, className,
                                 sklearnMetric, **kwargs):
        nanMask = trueDataframe[column].notna()
        trueValues = trueDataframe[column][nanMask]
        predictedValues = predictedDataframe[column][nanMask]
        return sklearnMetric(trueValues, predictedValues, labels=[className],
                             average="macro", **kwargs)
    
    def computeNumericPredictionMetric(self, trueModel, eventLogMessage, predictions,
                                       sklearnMetric, **kwargs):
        # get predicted events as dataframe
        modelDataframe = dataframeFromMessage(predictions, disableProgressBar=True)

        # get true events as dataframe
        if trueModel is None: # we can only use the real-world event log as the ground-truth
            trueDataframe = dataframeFromMessage(eventLogMessage, disableProgressBar=True)
        else: # we use the true model to generate the ground-truth log
            truePredictions = self.processDataUpdatesCpp.predictEventLog(
                eventLogMessage, trueModel, predictMeans=True).wait().log
            trueDataframe = dataframeFromMessage(truePredictions, disableProgressBar=True)

        trueDataframe.sort_values("case:concept:name", inplace=True)
        modelDataframe.sort_values("case:concept:name", inplace=True)

        # average over all variable's scores
        scores = []
        for column in modelDataframe.columns:
            # only compute score on numerical columns
            if getVariableType(column, eventLogMessage.variables) != "numerical":
                continue
            score = self.computeSinglePredictionMetric(trueDataframe[column], modelDataframe[column],
                                                       sklearnMetric, **kwargs)
            scores.append(score)

        if len(scores) == 0:
            return float("nan")

        return sum(scores) / len(scores)

    def computeSinglePredictionMetric(self, trueSeries, predictedSeries, sklearnMetric,
                                      predictingIndices=None, **kwargs):
        # no indices given => use all
        if predictingIndices is None:
            predictingIndices = [x for x in range(len(trueSeries))]
        # only use values in the series that have a prediction by the model
        trueValues = trueSeries[predictingIndices]
        predictedValues = predictedSeries[predictingIndices]
        # remove missing values
        nanMask = trueValues.notna()
        trueValues = trueValues[nanMask]
        predictedValues = predictedValues[nanMask]
        # compute the metric
        return sklearnMetric(trueValues, predictedValues, **kwargs)

    def setPredictionsToPython(self, setPredictionsMessage):
        return {column.variable : [frozenset(values) for values in column.predictions]
                for column in setPredictionsMessage}

    def computeValueSetMetric(self, modelToEvaluate, trueModel, eventLogMessage, evaluationModelIsList,
                              sklearnMetric, **kwargs):
        # compute set predictions for both models
        trueSetPredictions = self.processDataUpdatesCpp.predictValueSets(
            eventLogMessage, trueModel).wait().setPredictions
        trueSetPredictions = self.setPredictionsToPython(trueSetPredictions)
        if type(modelToEvaluate) == dict: # model of decision trees
            modelSetPredictions = self.getValueSetPredictionsFromTree(eventLogMessage,
                                                                      modelToEvaluate)
        else:
            # compute predictions through Cap'n Proto interface
            modelSetPredictions = self.processDataUpdatesCpp.predictValueSets(
                eventLogMessage, modelToEvaluate, evaluationModelIsList).wait().setPredictions
            modelSetPredictions = self.setPredictionsToPython(modelSetPredictions)

        # compute scores for each target variable
        scores = []
        for variableName in trueSetPredictions.keys():
            # obtain predictions for target variable
            trueValueSets = trueSetPredictions[variableName]
            modelValueSets = modelSetPredictions[variableName]
            
            # convert each set to a class (its ID respectively)
            valueSetsClasses = {valueSet: classId for classId, valueSet in 
                                enumerate(set(trueValueSets + modelValueSets))}
            trueClasses = [valueSetsClasses[valueSet] for valueSet in trueValueSets]
            modelClasses = [valueSetsClasses[valueSet] for valueSet in modelValueSets]

            # compute and append score
            score = sklearnMetric(trueClasses, modelClasses, **kwargs)
            scores.append(score)

        if len(scores) == 0:
            return float("nan")

        # average over scores
        return sum(scores) / len(scores)
        

    def computeMetric(self, metricName, modelToEvaluate, trueModel, eventLogMessage, evaluationModelIsList=False):
        if metricName == "strictF1":
            return self.computeF1Score(trueModel, modelToEvaluate, allowSwaps=False)
        elif metricName == "looseF1":
            return self.computeF1Score(trueModel, modelToEvaluate, allowSwaps=True)
        elif metricName == "predictionF1":
            predictions = self.computePredictions(modelToEvaluate, eventLogMessage)
            return self.computePredictionF1(modelToEvaluate, predictions)
        elif metricName == "complexity" and type(modelToEvaluate) == dict:
            return self.computeTreeComplexity(modelToEvaluate)
        elif metricName == "complexity":
            return self.computeComplexity(modelToEvaluate)
        elif metricName == "valueSetF1":
            return self.computeValueSetMetric(modelToEvaluate, trueModel, eventLogMessage,
                                              evaluationModelIsList, f1_score, average="macro")
        elif metricName == "valueSetBalancedAccuracy":
            return self.computeValueSetMetric(modelToEvaluate, trueModel, eventLogMessage, 
                                              evaluationModelIsList, balanced_accuracy_score)
        elif metricName == "predictionMAE":
            predictions = self.computePredictions(modelToEvaluate, eventLogMessage,
                                                  modelIsOrderedList=evaluationModelIsList)
            return self.computeNumericPredictionMetric(trueModel, eventLogMessage, predictions,
                                                       mean_absolute_error)
        elif metricName == "predictionRMSE":
            predictions = self.computePredictions(modelToEvaluate, eventLogMessage,
                                                  modelIsOrderedList=evaluationModelIsList)
            return self.computeNumericPredictionMetric(trueModel, eventLogMessage, predictions,
                                                       root_mean_squared_error)
        else:
            raise NotImplementedError("Metric not implemented")

    def createExperimentTable(self, directory, fileRegex, attributeNames):
        # generate columns for each attribute that was searched over
        valueDict = {x: [] for x in attributeNames}
        # generate column for the model's file name
        valueDict["filename"] = []
        # create entry for each file
        for filename in os.listdir(directory):
            # extract attributes form filename with regex
            matches = re.match(fileRegex, filename)
            # skip files that do not match
            if matches == None:
                continue
            # set all values extracted using the regex
            valuesTuple = matches.groups()
            for value, attribute in zip(valuesTuple, attributeNames):
                valueDict[attribute].append(value)
            # set filename
            valueDict["filename"].append(filename)
        
        return pd.DataFrame(valueDict)

    def addEvaluationToExperimentTable(self, directory, metricName, experimentTable):
        scoreColumn = []
        # compute evaluation metric for each row in the table
        for filename in tqdm(experimentTable["filename"], desc="Computing " + metricName):
            # get estimated model
            if "tree" in filename:
                # load sklearn decision tree model
                modelToEvaluate = joblib.load(directory + "/" + filename)
            else:
                # load model in Cap'n Proto format
                with open(directory + "/" + filename, "rb") as file:
                    modelToEvaluate = self.interface_capnp.RuleModel.read(file)
            # check whether estimated model is a list (for SSD++ models)
            evaluationModelIsList = False
            if "ssdpp" in filename:
                evaluationModelIsList = True
            # get ground-truth model
            trueModelId = experimentTable["modelID"][experimentTable["filename"] == filename].iloc[0]
            with open(directory + "/--" + trueModelId + "--true-model.bin", "rb") as file:
                trueModel = self.interface_capnp.RuleModel.read(file)
            # generate event log as test set for predictions
            testLog = None
            if metricName in ["predictionF1", "valueSetF1", "predictionMAE", "predictionRMSE",
                              "valueSetBalancedAccuracy"]:
                testLog = self.processDataUpdatesCpp.generateEventLog(trueModel, 200, 10).wait().log
            # compute and set the evaluation score
            score = self.computeMetric(metricName, modelToEvaluate, trueModel, testLog, evaluationModelIsList)
            scoreColumn.append(score)
        # add evaluation scores to table
        experimentTable[metricName] = scoreColumn
    
    def addAllMetricsToExperimentTable(self, directory, experimentTable):
        self.addMetricsToExperimentTable(directory, experimentTable, self.allMetrics)

    def addMetricsToExperimentTable(self, directory, experimentTable, metrics):
        for metricName in metrics:
            self.addEvaluationToExperimentTable(directory, metricName, experimentTable)

    def plotHyperparameterExperiment(self, aggregatedData, metricName, xAxisName, seriesesName):
        xAxis = aggregatedData.index.get_level_values(xAxisName)
        serieses = aggregatedData.index.get_level_values(seriesesName)

        for series in set(serieses):
            # draw a line for the median
            plt.plot(xAxis[serieses == series],
                    aggregatedData[metricName].midIqr[serieses == series], marker="x", 
                    label=seriesesName + " " + str(series))
            # fill the interquartile range
            plt.fill_between(xAxis[serieses == series],
                aggregatedData[metricName].lowIqr[serieses == series],
                aggregatedData[metricName].highIqr[serieses == series],
                alpha = 0.3, label=seriesesName + " " + str(series) + " IQR")
        
        plt.xlabel(self.descriptionLookup[xAxisName] + " (" + xAxisName + ")")
        plt.ylabel("Median " + self.descriptionLookup[metricName] + " (" + metricName + ")")
        plt.legend()
        plt.show()

    def plotScoreWithIqr(self, aggregatedData, metricName, xAxisName, constantColumnNames,
                         methodNames=["my", "turs", "true", "empty"], showLastMethod=True):
        # find columns that do not change
        constantColumns = {}
        for column in constantColumnNames:
            uniqueValues = aggregatedData.index.get_level_values(column).dropna().unique()
            if len(uniqueValues) > 1:
                raise RuntimeError("Constant column " + str(column) + " has more than one value")
            constantColumns[column] = uniqueValues[0]

        xAxis = aggregatedData.index.get_level_values(xAxisName)
        method = aggregatedData.index.get_level_values("method")
        methodNamesToPlot = methodNames[:-2]
        for methodName in methodNamesToPlot:
            # draw a line for the median
            plt.plot(xAxis[method == methodName],
                    aggregatedData[metricName].midIqr[method == methodName], marker="x", 
                    label=methodName + " models median")
            # fill the interquartile range
            plt.fill_between(xAxis[method == methodName],
                aggregatedData[metricName].lowIqr[method == methodName],
                aggregatedData[metricName].highIqr[method == methodName],
                alpha = 0.3, label=methodName + " models IQR")
            
        # draw a line for the perfect model
        perfectScore = aggregatedData[metricName].midIqr[method == methodNames[-2]].iloc[0]
        plt.plot((min(xAxis), max(xAxis)), (perfectScore, perfectScore), 
                 label=methodNames[-2] + " model")
        if showLastMethod:
            # draw a line for the empty model
            emptyScore = aggregatedData[metricName].midIqr[method == methodNames[-1]].iloc[0]
            lowEmptyScore = aggregatedData[metricName].lowIqr[method == methodNames[-1]].iloc[0]
            highEmptyScore = aggregatedData[metricName].highIqr[method == methodNames[-1]].iloc[0]
            emptyModelLine, = plt.plot((min(xAxis), max(xAxis)), (emptyScore, emptyScore), 
                                       label=methodNames[-1] + " models median")
            # fill IQR for empty model
            plt.fill_between((min(xAxis), max(xAxis)), (lowEmptyScore, lowEmptyScore),
                            (highEmptyScore, highEmptyScore), alpha = 0.3, 
                            facecolor=emptyModelLine.get_color(), 
                            label=methodNames[-1] + " models IQR")
        
        # add constant columns to title
        title = self.descriptionLookup[metricName] + " at "
        for column, value in constantColumns.items():
            title += str(column) + "=" + str(value) + " "
        plt.title(title)
        plt.xlabel(self.descriptionLookup[xAxisName] + " (" + xAxisName + ")")
        plt.ylabel(self.descriptionLookup[metricName] + " (" + metricName + ")")
        plt.legend()
        plt.show()

    def lowIqr(self, x):
        return np.quantile(x, q=0.25)
    def midIqr(self, x):
        return np.quantile(x, q=0.5)
    def highIqr(self, x):
        return np.quantile(x, q=0.75)

    def aggregateMetrics(self, table, groupingColumns, metrics=None):
        if metrics is None:
            metrics = self.allMetrics
        aggregationFunctions = [self.lowIqr, self.midIqr, self.highIqr]
        aggregationDict = {x: aggregationFunctions for x in metrics}
        return table.groupby(groupingColumns, dropna=False).agg(aggregationDict)
    
    def generateSingleRuleModels(self, model):
        result = []
        # generate models with one rule only
        for rule in model.rules:
            singleRuleModel = self.interface_capnp.RuleModel.new_message()
            singleRuleModel.variables = list(model.as_builder().variables)
            rules = singleRuleModel.init("rules", 1)
            rules[0] = rule.as_builder()
            result.append(singleRuleModel)
        return result
    
    def getPredictingPositionsAsIndices(self, eventLogMessage, eventLogDataframe, model, variable,
                                        caseIdColumn="case:concept:name"):
        result = []
        # get all predicting positions
        positionsMessage = self.processDataUpdatesCpp.getPredictingPositions(
            eventLog=eventLogMessage, model=model).wait().positions
        # find positions for variable
        for mapping in positionsMessage:
            if mapping.targetVariable == variable:
                # convert two-part to one-part index
                for position in mapping.positions:
                    traceStart = eventLogDataframe[caseIdColumn].searchsorted(position.caseId)
                    onePartIndex = traceStart + position.index
                    result.append(onePartIndex)
                return result
    
    def getPredictionsFromTree(self, eventLogMessage, treeModel, caseIdColumn="case:concept:name"):
        result = dict()
        result[caseIdColumn] = list(dataframeFromMessage(eventLogMessage, True)[caseIdColumn])
        treeAdapter = DecisionTreeAdapter()

        for targetVariable in eventLogMessage.variables:
            targetVariableName = targetVariable.name
            if targetVariableName in treeModel:
                # use tree to predict values
                decisionTree = treeModel[targetVariableName]
                X, _, catLabelMapping = treeAdapter.eventLogToDecisionTreeDataset(eventLogMessage,
                                                                                targetVariableName)
                yPredicted = decisionTree.predict(X)
                if getVariableType(targetVariableName, eventLogMessage.variables) == "categorical":
                    # map integer-encode categories back to category names
                    result[targetVariableName] = [catLabelMapping[item] for item in yPredicted]
                else:
                    # use numeric predictions by the tree
                    result[targetVariableName] = [item for item in yPredicted]
            else:
                # choose random value from domain
                if targetVariable.which() == "numerical":
                    result[targetVariableName] = [random.random()*100 for x in result[caseIdColumn]]
                else:
                    result[targetVariableName] = [random.choice(targetVariable.categorical.domain)
                                                  for x in result[caseIdColumn]]

        return result

    def getValueSetPredictionsFromTree(self, eventLogMessage, treeModel,
                                       caseIdColumn="case:concept:name"):
        result = dict()
        result[caseIdColumn] = list(dataframeFromMessage(eventLogMessage, True)[caseIdColumn])
        treeAdapter = DecisionTreeAdapter()

        for targetVariableName, decisionTree in treeModel.items():
            # only create value set predictions for categorical variables
            if getVariableType(targetVariableName, eventLogMessage.variables) == "categorical":
                X, _, catLabelMapping = treeAdapter.eventLogToDecisionTreeDataset(eventLogMessage,
                                                                                  targetVariableName)
                yPredicted = decisionTree.predict(X)
                yPredictedLabels = [frozenset([catLabelMapping[item]]) for item in yPredicted]
                result[targetVariableName] = yPredictedLabels

        return result

    def evaluateRules(self, eventLogMessage, datasetName, model, modelName, categoricalTable, numericalTable):
        # get sorted dataframe
        dataframe = dataframeFromMessage(eventLogMessage, disableProgressBar=True)

        # evaluate each rule separately
        singleRuleModels = self.generateSingleRuleModels(model)
        for i, m in enumerate(tqdm(singleRuleModels)):
            targetVariableName = m.rules[0].updateRule.variable
            # get predictions and positions where they are based on the rule
            predictions = self.computePredictions(m, eventLogMessage)
            predictedDataframe = dataframeFromMessage(predictions, disableProgressBar=True)
            predictingIndices = self.getPredictingPositionsAsIndices(eventLogMessage, dataframe,
                                                                     m, targetVariableName)
            # compute F1 score for categorical targets
            if getVariableType(targetVariableName, model.variables) == "categorical":
                targetValue = m.rules[0].updateRule.valueSetRule.set[0]
                score = self.computeSinglePredictionMetric(dataframe[targetVariableName],
                                                           predictedDataframe[targetVariableName],
                                                           f1_score, predictingIndices,
                                                           average="macro", labels=[targetValue])
                newRow = pd.DataFrame([{"model": modelName,
                                        "dataset": datasetName,
                                        "ruleNumber": i,
                                        "targetVariable": targetVariableName,
                                        "f1Score": score}])
                categoricalTable = pd.concat([categoricalTable, newRow], ignore_index=True)
            # compute RMSE for numerical targets
            elif getVariableType(targetVariableName, model.variables) == "numerical":
                score = self.computeSinglePredictionMetric(dataframe[targetVariableName],
                                                           predictedDataframe[targetVariableName],
                                                           root_mean_squared_error, predictingIndices)
                newRow = pd.DataFrame([{"model": modelName,
                                        "dataset": datasetName,
                                        "ruleNumber": i,
                                        "targetVariable": targetVariableName,
                                        "RMSE": score}])
                numericalTable = pd.concat([numericalTable, newRow], ignore_index=True)
            else:
                raise RuntimeError("Invalid variable type")
    
        return categoricalTable, numericalTable

def getVariableType(variableName, variablesMessage):
    for variable in variablesMessage:
        if variable.name == variableName:
            return variable.which()
