import random
from .addnoise import addSwapNoiseToMessage
import time
import pandas as pd
from .decisiontree import DecisionTreeAdapter
import joblib

class ModelGenerator:

    def __init__(self, interface_capnp, processDataUpdatesCpp, processDataUpdatesBaselinesDict) -> None:
        self.interface_capnp = interface_capnp
        self.processDataUpdatesCpp = processDataUpdatesCpp
        self.processDataUpdatesBaselinesDict = processDataUpdatesBaselinesDict

    def removeCyclicRules(self, model):
        nonCyclicRules = []
        # add all discovered rules step by step
        for i in range(len(model.rules)):
            # add rule to the result
            nonCyclicRules.append(model.rules[i])
            # check whether current model is cyclic
            modelToCheck = self.interface_capnp.RuleModel.new_message(
                rules=nonCyclicRules, variables=list(model.variables))
            isCyclic = self.processDataUpdatesCpp.isModelCyclic(modelToCheck).wait().cyclic
            # discard rule if model is cyclic
            if isCyclic:
                del nonCyclicRules[-1]
        
        # create model of initial rules without cycles
        return self.interface_capnp.RuleModel.new_message(rules=nonCyclicRules,
            variables=list(model.variables))

    def generateVariableDictionary(self, catVariablesCount, numVariablesCount, LETTERS):
        # pick as many variable names as necessary
        variableNames = LETTERS[:(catVariablesCount + numVariablesCount)]
        # randomly assign variable types to variable names and generate the desired number
        variableTypes = (["categorical"] * catVariablesCount) + (["numerical"] * numVariablesCount)
        random.shuffle(variableTypes)
        # create dictionary
        return {variableNames[i] : variableTypes[i] for i in range(len(variableNames))}

    def generateVariableList(self, candidateModel, variableDictionary, valueNames):
        # initialize list
        candidateModel.init("variables", len(variableDictionary))
        # generate a variable in the list for each entry in the variableDictionary
        for j, (variableName, variableType) in enumerate(variableDictionary.items()):
            # set the variable name, type and (optionally) domain
            candidateModel.variables[j].name = variableName
            if variableType == "categorical":
                candidateModel.variables[j].init("categorical")
                candidateModel.variables[j].categorical.init("domain", len(valueNames))
                for k in range(len(valueNames)):
                    candidateModel.variables[j].categorical.domain[k] = valueNames[k]
            elif variableType == "numerical":
                candidateModel.variables[j].numerical = None
            else:
                raise RuntimeError("Neither categorical nor numerical variable type")

    def generateValue(self, variableType, valueNames, conditionOperator = None,
                      minNumber=0.0, maxNumber=100.0, digits=1):
        if variableType == "categorical":
            # randomly choose categorical value
            return random.sample(valueNames, 1)[0]
        elif variableType == "numerical":
            center = (maxNumber + minNumber) / 2.0
            offsetFactor = 0.2
            offset = (maxNumber - minNumber) * offsetFactor

            # generate number that is sufficiently far away from the center of the interval
            if conditionOperator is None:
                largerThanCenter = bool(random.getrandbits(1))
                if largerThanCenter:
                    lowerBound = center + offset
                    upperBound = maxNumber
                else:
                    lowerBound = minNumber
                    upperBound = center - offset
            elif conditionOperator == "<=":
                lowerBound = center + offset / 2
                upperBound = maxNumber - offset / 2
            elif conditionOperator == ">=":
                lowerBound = minNumber + offset / 2
                upperBound = center - offset / 2
            else:
                raise RuntimeError("conditionOperator is unknown")
            
            
            return round(random.uniform(lowerBound, upperBound), digits)

    def generateRuleBuildingBlocks(self, variableDictionary, valueNames, targetType=None):
        # randomly choose variables (their names respectively)
        variableNames = list(variableDictionary.keys())
        conditionVariable = random.sample(variableNames, 1)[0]
        if targetType is None:
            updateVariable = random.sample(variableNames, 1)[0]
        else:
            possibleTargets = [name for name, varType in variableDictionary.items()
                               if varType == targetType]
            updateVariable = random.sample(possibleTargets, 1)[0]

        # randomly choose operator for numerical condition
        conditionOperator = "="
        if variableDictionary[conditionVariable] == "numerical":
            conditionOperator = random.sample(["<=", ">="], 1)[0]

        # randomly choose values for the variables
        conditionValue = self.generateValue(variableDictionary[conditionVariable], valueNames, conditionOperator)
        updateValue = self.generateValue(variableDictionary[updateVariable], valueNames)
        
        return (conditionVariable, conditionValue, conditionOperator), (updateVariable, updateValue)

    def generateCondition(self, conditionMessage, buildingBlocks):
        (conditionVariable, conditionValue, conditionOperator), _ = buildingBlocks
        # initialize one-element atomic condition and set its variable
        conditionMessage.init("elements", 1)
        conditionMessage.elements[0].init("atomicCondition")
        conditionMessage.elements[0].atomicCondition.variable = conditionVariable
        
        # generate atomic condition based on operator (and implied variable type)
        if conditionOperator == "=": # SSD++ has equality only for categorical attributes
            conditionMessage.elements[0].atomicCondition.init("categoricalValueCondition")
            conditionMessage.elements[0].atomicCondition.categoricalValueCondition.value = conditionValue
        elif conditionOperator == "<=":
            conditionMessage.elements[0].atomicCondition.init("lowerBoundCondition")
            conditionMessage.elements[0].atomicCondition.lowerBoundCondition.value = conditionValue
        elif conditionOperator == ">=":
            conditionMessage.elements[0].atomicCondition.init("upperBoundCondition")
            conditionMessage.elements[0].atomicCondition.upperBoundCondition.value = conditionValue
        else:
            raise RuntimeError("Invalid operator for generating condition message")

    def generateUpdateRule(self, updateRuleMessage, buildingBlocks, variableType):
        _, (updateVariable, updateValue) = buildingBlocks
        updateRuleMessage.variable = updateVariable
        # generate appropriate update rule that predicts one value based on the variable type
        if variableType == "categorical":
            updateRuleMessage.init("valueSetRule")
            updateRuleMessage.valueSetRule.init("set", 1)
            updateRuleMessage.valueSetRule.set[0] = updateValue
        elif variableType == "numerical":
            updateRuleMessage.init("valueRule")
            updateRuleMessage.valueRule.constant = updateValue
        else:
            raise RuntimeError("Neither numerical nor categorical variable")

    def generateTrueModels(self, targetType=None, numberOfModels=10, numberOfRules=5,
                           catVariablesCount=5, numVariablesCount=0, valuesPerCatVariable=4):
        # generate categorical value names
        LETTERS = "abcdefghijklmnopqrstuvwxyz"
        valueNames = LETTERS.upper()[:valuesPerCatVariable]
        # generate desired number of models
        trueModels = []
        for i in range(numberOfModels):
            # generate variable names and types
            variableDictionary = self.generateVariableDictionary(catVariablesCount,
                                                                 numVariablesCount, LETTERS)
            cyclicModelGenerated = True
            while cyclicModelGenerated:
                # generate variable list
                candidateModel = self.interface_capnp.RuleModel.new_message()
                self.generateVariableList(candidateModel, variableDictionary, valueNames)
                # generate rules
                previousRules = set()
                candidateModel.init("rules", numberOfRules)
                for j in range(numberOfRules):
                    # randomly generate variables, values and operator for rule
                    ruleAlreadyExists = True
                    while ruleAlreadyExists:
                        buildingBlocks = self.generateRuleBuildingBlocks(variableDictionary,
                                                                         valueNames, targetType)
                        ruleAlreadyExists = buildingBlocks in previousRules
                    
                    _, (updateVariable, _) = buildingBlocks
                    # create condition
                    self.generateCondition(candidateModel.rules[j].condition, buildingBlocks)
                    # create update rule
                    self.generateUpdateRule(candidateModel.rules[j].updateRule, buildingBlocks,
                                            variableDictionary[updateVariable])

                    # add rule to the set of all previously generated rules
                    previousRules.add(buildingBlocks)

                # check cyclicity of model
                cyclicModelGenerated = self.processDataUpdatesCpp.isModelCyclic(candidateModel).wait().cyclic
                if not cyclicModelGenerated:
                    trueModels.append(candidateModel)
        return trueModels

    def saveEmptyModels(self, experimentName, trueModels):
        for i, model in enumerate(trueModels):
            emptyModel = self.interface_capnp.RuleModel(rules=[], variables=list(model.variables))
            self.saveModel(experimentName, "", "", i, "", "empty", emptyModel)

    def saveTrueModels(self, experimentName, trueModels):
        for i, model in enumerate(trueModels):
            self.saveModel(experimentName, "", "", i, "", "true", model)

    def saveModel(self, experimentName, proportion, logLength, modelId, datasetId, method, model):
        fileName = experimentName + "/" + str(proportion) + "-" + str(logLength) + "-" + str(modelId)\
            + "-" + str(datasetId) + "-" + method + "-model.bin"
        if method == "tree":
            joblib.dump(model, fileName)
        else:
            with open(fileName, "wb") as file:
                model.write(file)
        return fileName

    def findTargetVariables(self, model):
        # get target variable type (categorical or numerical)
        oneTargetVariableName = model.rules[0].updateRule.variable
        targetVariableType = ""
        for variable in model.variables:
            if variable.name == oneTargetVariableName:
                targetVariableType = variable.which()

        # get all variables of the target variable type
        result = []
        for variable in model.variables:
            if variable.which() == targetVariableType:
                result.append(variable.name)
        return result

    def generateModelEstimates(self, experimentName, trueModels, noiseProportions, logLengths,
                                numberOfDatasets=1, runtimesFilename = "runtimes.csv"):
        runtimeDataFrame = pd.DataFrame({"filename": [], "runtime": []})

        for proportion in noiseProportions: # iterate over noise proportions
            for logLength in logLengths: # iterate over events/samples in the log
                for i, trueModel in enumerate(trueModels): # iterate over the generating models
                    traceLength = 10
                    numberOfTraces = logLength // traceLength
                    for j in range(numberOfDatasets): # iterate over the event logs per generating model
                        # generate event log
                        log = self.processDataUpdatesCpp.generateEventLog(
                            trueModel, numberOfTraces, traceLength).wait().log
                        # swap values to add noise in the log
                        noisyLog = addSwapNoiseToMessage(log, proportion, disableProgressBar=True)

                        # restrict target variables
                        targetVariables = self.findTargetVariables(trueModel)

                        # get MOODY model
                        startTime = time.time()
                        myModel = self.processDataUpdatesCpp.findModel(noisyLog, targetVariables).wait().model.as_builder()
                        duration = time.time() - startTime
                        filename = self.saveModel(experimentName, proportion, logLength, i, j, "my", myModel)
                        newRow = pd.DataFrame([{"filename": filename, "runtime": duration}])
                        runtimeDataFrame = pd.concat([runtimeDataFrame, newRow], ignore_index=True)

                        # get baseline models
                        for baselineName, interface in self.processDataUpdatesBaselinesDict.items():
                            startTime = time.time()
                            baselineModel = self.getBaselineModel(noisyLog, targetVariables, baselineName, interface)
                            duration = time.time() - startTime
                            # remove cycles from TURS models as they are unordered like MOODY
                            if baselineName == "turs":
                                baselineModel = self.removeCyclicRules(baselineModel)
                            filename = self.saveModel(experimentName, proportion, logLength, i, j, baselineName, baselineModel)
                            newRow = pd.DataFrame([{"filename": filename, "runtime": duration}])
                            runtimeDataFrame = pd.concat([runtimeDataFrame, newRow], ignore_index=True)
        
        runtimeDataFrame.to_csv(runtimesFilename, index=False)
    
    def getBaselineModel(self, eventLog, targetVariables, baselineName, interface):
        if baselineName == "tree":
            treeAdapter = DecisionTreeAdapter()
            return treeAdapter.findDecisionTreeModel(eventLog, targetVariables)
        else:
            return interface.findModel(eventLog, targetVariables).wait().model.as_builder()

    def generateHyperparameterModels(self, experimentName, trueModels, noiseProportions,
                                     conditionCounts, numberOfDatasets=1,
                                     runtimesFilename = "runtimes.csv"):
        runtimeDataFrame = pd.DataFrame({"filename": [], "runtime": []})

        for proportion in noiseProportions: # iterate over noise proportions
            for conditionCount in conditionCounts: # iterate over number of conditions to search
                for i, trueModel in enumerate(trueModels): # iterate over the generating models
                    traceLength = 10
                    logLength = 2000
                    numberOfTraces = logLength // traceLength
                    for j in range(numberOfDatasets): # iterate over the event logs per generating model
                        # generate event log
                        log = self.processDataUpdatesCpp.generateEventLog(
                            trueModel, numberOfTraces, traceLength).wait().log
                        # swap values to add noise in the log
                        noisyLog = addSwapNoiseToMessage(log, proportion, disableProgressBar=True)

                        # restrict target variables
                        targetVariables = self.findTargetVariables(trueModel)

                        # estimate model from event log
                        startTime = time.time()
                        myModel = self.processDataUpdatesCpp.findModel(
                            noisyLog, targetVariables, conditionCount=conditionCount).wait().model.as_builder()
                        duration = time.time() - startTime
                    
                        # save models for later evaluation
                        filename = self.saveModel(experimentName, proportion, conditionCount, i, j, "my", myModel)

                        # keep track of runtimes
                        newRow = pd.DataFrame([{"filename": filename, "runtime": duration}])
                        runtimeDataFrame = pd.concat([runtimeDataFrame, newRow], ignore_index=True)
        
        runtimeDataFrame.to_csv(runtimesFilename, index=False)
